static char rcsid[] = "@(#)$Id: cs_utf.c,v 1.3 2000/07/30 15:12:10 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "s_me.h"
#include "cs_imp.h"

/* 'Map' for UTF-8 -------------------------------------------------------- */

static struct utfelem { 
    uint16  start;           
    uint16  end;
    uint16  unicodeval;          /* Unicode value correspond with 'start' */
} utf8elems[] = {
    /* Identity map excluding: 
       0x0000 -- 0x007F    US-ASCII   -- by default identity on type=utf-8
       0xD800 -- 0xDFFF    Surrogates Area
       0xE000 -- 0xF8FF    Private Use Area
    */

    { 0x0080, 0xD7FF,  0x0080 },
    { 0xF900, 0xFFFF,  0xFFFF },    
};

/* Mappping to unicode */
static uint16 map_utf P_((struct map_info *map,
			  unsigned int ch));
static uint16 map_utf(map,ch)
     struct map_info *map;
     unsigned int ch;
{
    if (ch < 0x0080)
	return ch;               /* Identity */

    if (map) {
	int i;
	if (!map->map_initialized)
            map->map_init_it(map);

	for (i = 0; i < map->b.utfmap.elem_count; i++) {
	    if (ch >= map->b.utfmap.elems[i].start &&
		ch <= map->b.utfmap.elems[i].end) {
		return ( ch - map->b.utfmap.elems[i].start ) +
		    map->b.utfmap.elems[i].unicodeval;
	    }
	}
    }
    return MAPPING_NONE;
}

/* Mappping from unicode */
static uint16 map_utf_rev P_((struct map_info *map,
			      unsigned int val, int *found)); 
static uint16 map_utf_rev(map,val,found) 
     struct map_info *map;
     unsigned int val;
     int * found;
{
    if (val < 0x0080) {
	*found = 1;
	return val;               /* Identity */
    }
    
    if (map) {
	int i;
	if (!map->map_initialized)
            map->map_init_it(map);
	
	for (i = 0; i < map->b.utfmap.elem_count; i++) {
	    int n = map->b.utfmap.elems[i].end -
		map->b.utfmap.elems[i].start;
	    if (val >= map->b.utfmap.elems[i].unicodeval &&
		val <= map->b.utfmap.elems[i].unicodeval+n) {
		*found = 1;
		return ( val - map->b.utfmap.elems[i].unicodeval ) +
		    map->b.utfmap.elems[i].start;	    
	    }
	}
    }
    *found = 0;
    return 0x003F; /* '?' */
}

static void map_init_utf8 P_((struct  map_info *map));
static void map_init_utf8(map)
     struct  map_info *map;
{
    map_utf8.b.utfmap.elem_count = sizeof (utf8elems) / sizeof (utf8elems[0]);
    map_utf8.b.utfmap.elems = utf8elems;
    map_utf8.map_initialized = 1;

    dprint(5,(debugfile,"Map %s initialized\n",map->map_name));
}

struct  map_info map_utf8 = { &cs_utf8, "UNICODE", 0, map_init_utf8, 0 };

static int read_elemlist P_((const char *name,
			     struct utfelem ** elems,
			     int *elemcount));
static int read_elemlist(name,elems,elemcount)
     CONST char *name;
     struct utfelem ** elems;
     int *elemcount;
{
    static struct elemlist {
	char           * name;
	struct utfelem * elems;
	int              elemcount;
    } * ELIST = NULL;
    static int ECOUNT = 0;
    FILE *F;
    int i;
    char buffer[STRING];
    int l;
    char *fn = NULL;
    
    for (i = 0; i < ECOUNT; i++) {
	if (0 == strcmp(ELIST[i].name,name)) {
	    *elems      = ELIST[i].elems;
	    *elemcount  = ELIST[i].elemcount;
	    return 1;
	}
    }

    F = open_mapname(name,&fn);
    if (!F) {
	if (fn)
	    free(fn);
	return 0;
    }

    ELIST = safe_realloc(ELIST,(ECOUNT+1)*sizeof (struct elemlist));
    ELIST[ECOUNT].name      = safe_strdup(name);
    ELIST[ECOUNT].elems     = NULL;
    ELIST[ECOUNT].elemcount = 0;

    while (0 < (l = mail_gets(buffer,sizeof buffer,F))) {
	long s1, e1, s2, e2;
	char *p;
	if (buffer[l-1] != '\n') {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeMapTooLongLine,
			      "Map %s: %s: Too long line: %s"),
		      name,fn,buffer);
	}
	if (buffer[0] == '#')
	    continue;
	
	s1 = strtol(buffer,&p,16);	
	e1 = s1;
	
	if (*p == '-') 
	    e1 = strtol(p+1,&p,16);
	
	if (*p == ' ' || *p == '\t') {
	    while (*p == ' ' || *p == '\t')
		p++;
	    s2 = strtol(p,&p,16);
	    e2 = s2;
	    if (*p == '-') 
		e2 = strtol(p+1,&p,16);       	    	
	} else {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeMapBadValue,
			      "Map %s: %s: Bad value: %s"),
		      name,fn,buffer);
	    continue;
	}

	if (*p != '\n' || e1 - s1 < 0 || e1 - s1 != e2 - s2 ||
	    s1 < 0x0000 || e1 > 0xFFFF || s2 < 0x0000 || e1 > 0xFFFF) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeMapBadValue,
			      "Map %s: %s: Bad value: %s"),
		      name,fn,buffer);
	    continue;
	}
	
	ELIST[ECOUNT].elems = safe_realloc(ELIST[ECOUNT].elems,
					   (ELIST[ECOUNT].elemcount+1) *
					   sizeof (struct utfelem));
	ELIST[ECOUNT].elems[ELIST[ECOUNT].elemcount].start       = s1;
	ELIST[ECOUNT].elems[ELIST[ECOUNT].elemcount].end         = e1;
	ELIST[ECOUNT].elems[ELIST[ECOUNT].elemcount].unicodeval  = s2;
	ELIST[ECOUNT].elemcount++;
    }
    
    *elems      = ELIST[ECOUNT].elems;
    *elemcount  = ELIST[ECOUNT].elemcount;
    ECOUNT++;

    if (fn)
	free(fn);
    return 1;
}

static void map_init_bad P_((struct  map_info *map));
static void map_init_bad(map)
     struct  map_info *map;
{
    panic("STRING PANIC",__FILE__,__LINE__,"map_init_bad",
	  "map_init_bad called",0);
}

static struct  map_info * open_utf8_map P_((const char * map_name));
static struct  map_info * open_utf8_map(map_name)
     CONST char * map_name;
{
    struct map_info *ret = NULL;
    int err = 0;
    struct utfelem * elems;
    int elemcount;
    
    if (read_elemlist(map_name,&elems,&elemcount)) {	
	ret = safe_malloc(sizeof (struct map_info));
	ret -> map_type            = &cs_utf8;
	ret -> map_name            = safe_strdup(map_name);
	ret -> map_initialized     = 1;
	ret -> map_init_it         = map_init_bad;
	ret -> b.utfmap.elem_count = elemcount;
	ret -> b.utfmap.elems      = elems;
    }
    
    return ret;
}

/* ------------------------------------------------------------------------ */
static void cs_init_utf P_((struct string *str));
static void cs_init_utf(str)
     struct string *str;
{
    str->p->len = 0;
    str->p->a.words = 0;
}

static void cs_free_utf P_((struct string *str));
static void cs_free_utf(str)
     struct string *str;
{
    if (str->p->a.words) {
	free(str->p->a.words);
	str->p->a.words = NULL;
    }
    str->p->len = 0;
}

static int cs_add_streambyte_to_utf8 P_((struct string *str,int ch));
static int cs_add_streambyte_to_s_utf8 P_((struct charset_state *str, int ch));
static void cs_soft_reset_s_utf8 P_((struct charset_state *str));

static int cs_add_streambyte_to_utf8(str,ch)
     struct string *str;
     int ch;
{
    if (!str->p->state)
	str->p->state = new_state_1(str->string_type);

    if (!cs_add_streambyte_to_s_utf8(str->p->state,ch))
	return 0;

    if (str->p->state->p->ready) {
	if (str->p->state->p->a.utf8.value > 0xFFFF)
	    return 0;

	/* NOTE:  str->p->a.words is not NUL terminated        */
	str->p->a.words = safe_realloc(str->p->a.words,
				       (str->p->len+1)* sizeof (uint16));
	
	str->p->a.words[str->p->len++] = 
	    str->p->state->p->a.utf8.value;
	cs_soft_reset_s_utf8(str->p->state);
    }

    return 1;
}

static void cs_add_intdata_to_utf P_((struct string *str,
				      const struct string *data));
static void cs_add_intdata_to_utf(str,data)
     struct string *str;
     CONST struct string *data;
{
    if (data->p->len) {
	int i;
	/* realloc with size 0 is equivalent of free and may 
	   corrupt memory ...
	*/

	/* NOTE:  str->p->a.words is not NUL terminated        */
	str->p->a.words = safe_realloc(str->p->a.words,
				       (str->p->len+data->p->len)* 
				       sizeof (uint16));

	for (i = 0; i < data->p->len; i++) {
	    str->p->a.words[str->p->len++] = data->p->a.words[i];
	}
    }
}

static uint16 cs_give_unicode_from_utf8 P_((const struct string *str,
					    int pos, int *found));
static uint16 cs_give_unicode_from_utf8(str,pos,found)
     CONST struct string *str;
     int pos; 
     int *found;
{
    uint16 res;
    if (pos < 0 || pos >= str->p->len)
	panic("STRING PANIC",__FILE__,__LINE__,"cs_give_unicode_from_utf8",
	      "Index out of array",0);
    
    res = map_utf(str->string_type->map_info,str->p->a.words[pos]);

    if (res != MAPPING_NONE) {
	*found = 1;
	return res;
    }
    *found = 0;
    return 0x003F;  /* '?' */
}

static void cs_add_unicodedata_to_utf P_((struct string *str,
					  int len, uint16 *data));
static void cs_add_unicodedata_to_utf(str,len,data)
     struct string *str;
     int len; 
     uint16 *data;
{
    if (len > 0) {
	int i;
	/* realloc with size 0 is equivalent of free and may 
	   corrupt memory ...
	*/
	
	/* NOTE:  str->p->a.words is not NUL terminated        */
	str->p->a.words = safe_realloc(str->p->a.words,
				       (str->p->len+len)* sizeof (uint16));

	for (i = 0; i < len; i++) {
	    int found;	    
	    str->p->a.words[str->p->len++] = 
		map_utf_rev(str->string_type->map_info,data[i],&found);
	}
    }
}

static int cs_cmp_utf P_((struct string *str1,struct string *str2));
static int cs_cmp_utf(str1,str2)
     struct string *str1;
     struct string *str2;
{
    int i;

    if (str1->string_type->charset_type !=
	str2->string_type->charset_type) 
	panic("STRING PANIC",__FILE__,__LINE__,"cs_cmp_utf",
	      "String type mismatch",0);

    for (i = 0; i < str1->p->len && i < str2->p->len; i++) {
	if (str1->p->a.words[i] < str2->p->a.words[i])
	    return -1;
	if (str1->p->a.words[i] > str2->p->a.words[i])
	    return 1;
    }

    if (i < str1->p->len)
	return 1;
    if (i < str2->p->len)
	return -1;
    return 0;
}

typedef unsigned char utf_buffer[4];

static int gen_utf_char P_((utf_buffer buffer, unsigned int val));
static int gen_utf_char(buffer,val)
     utf_buffer buffer;
     unsigned int val;
{
    if (val < 0x0080) {
	buffer[0] = val;
	return 1;
    } else {
	int    l,x;
	uint16 bits;
	if (val < 0x7FF) {
	    bits = 192;
	    l = x = 2;
	} else { 
	    bits = 224;
	    l = x = 3;
	}

	while (x > 0) {
	    buffer[--x] = 128 | (val & 63);
	    val >>= 6;
	}
	if (val) {
	    panic("STRING PANIC",__FILE__,__LINE__,"gen_utf_char",
		  "Value overflow?",0);	
	}

	buffer[0] |= bits;
	return l;
    }
}


static unsigned char *cs_stream_from_utf8 P_((const struct string *str, 
				     int printable,
				     screen_info_p terminal));
static unsigned char *cs_stream_from_utf8(str,printable,terminal)
     CONST struct string *str;
     int printable;
     screen_info_p terminal;
{
    unsigned char * res = safe_malloc(str->p->len*3+1);
    int i;
    int x = 0;
    
    for (i = 0; i < str->p->len; i++) {
	uint16 val = str->p->a.words[i];
	if (printable) {
	    uint16 unicode = map_utf(str->string_type->map_info,val);
	    if (MAPPING_NONE == unicode ||
		!unicode_ch(unicode,UOP_printable))
		val = 0x003F;  /* '?' */
	}
	x += gen_utf_char(res+x,val);
    }
    
    res[x] = '\0';
    return res;
}

static int cs_can_ascii_utf P_((const struct string *str));
static int cs_can_ascii_utf(str)
     CONST struct string *str;
{
    int i;
    for (i = 0; i < str->p->len; i++) {
	/* str->p->a.words is unsigned */
	if (str->p->a.words[i] > 127)
	    return 0;
    }
    return 1;
}

static unsigned char *cs_streamclip_from_utf8 P_((const struct string *str,
					 int *pos, int len,
					 screen_info_p terminal));
static unsigned char *cs_streamclip_from_utf8(str,pos,len,terminal)
     CONST struct string *str;
     int *pos; 
     int len;
     screen_info_p terminal;
{
    unsigned char * ret;
    int l = 0,i;

    if (*pos < 0)
	panic("STRING PANIC",__FILE__,__LINE__,"cs_streamclip_from_utf8",
	      "Index out of array",0);

    if (len < 0)
	panic("STRING PANIC",__FILE__,__LINE__,"cs_streamclip_from_utf8",
	      "Negative size",0);
    
    ret = safe_malloc(len*3+1);
    for (i = 0; i < len && *pos < str->p->len; i++, (*pos)++) {
	uint16 val = str->p->a.words[i];
	uint16 unicode = map_utf(str->string_type->map_info,val);
	if (MAPPING_NONE == unicode ||
	    !unicode_ch(unicode,UOP_printable))
	    val = 0x003F;  /* '?' */
	l += gen_utf_char(ret+l,val);
    }
    ret[l] = '\0';
    return ret;
}

static void cs_clip_from_utf P_((struct string *ret,
				 const struct string *str,
				 int *pos, int len));
static void cs_clip_from_utf(ret,str,pos,len)
     struct string *ret;
     CONST struct string *str;
     int *pos; 
     int len;
{
    int i;

    if (ret->string_type->charset_type !=
	str->string_type->charset_type) 
	panic("STRING PANIC",__FILE__,__LINE__,"cs_clip_from_utf",
	      "String type mismatch",0);

    if (*pos < 0)
	panic("STRING PANIC",__FILE__,__LINE__,"cs_clip_from_utf",
	      "Index out of array",0);

    if (len < 0)
	panic("STRING PANIC",__FILE__,__LINE__,"cs_clip_from_utf",
	      "Negative size",0);

    if (len > 0) {
	/* realloc with len == 0 is equivalent of freeing and
	   may result corruption of memory ...
	*/

	/* NOTE:  str->p->a.words is not NUL terminated        */
	ret->p->a.words = safe_realloc(ret->p->a.words,
				       len* sizeof (uint16));
	ret->p->len = 0;
	
	for (i = 0; i < len && *pos < str->p->len; i++, (*pos)++) {   
	    ret->p->a.words[ret->p->len++] = str->p->a.words[*pos];
	}
    } else {  
	/* NULL (empty) string */

	if (ret->p->a.words)
	    free(ret->p->a.words);
	ret->p->a.words = NULL;
	ret->p->len = 0;
    }
}

static int cs_find_pattern_from_utf P_((const struct string *str,
					const struct string *pattern,
					int ignore_case));
static int cs_find_pattern_from_utf(str,pattern,ignore_case)
     CONST struct string *str;
     CONST struct string *pattern;
     int ignore_case;
{
    int ret = 0;
    int i, j;

    if (pattern->string_type->charset_type !=
	str->string_type->charset_type) 
	panic("STRING PANIC",__FILE__,__LINE__,"cs_find_pattern_from_utf",
	      "String type mismatch",0);

    if (ignore_case) {
	dprint(83,(debugfile,
		   "cs_find_pattern_from_utf=-1\n"));	
	return -1;   /* Use UNICODE comparision on upper level instead */
    }

    for (i = 0; i < str->p->len; ) {
	CONST int s = i + 1;
	
	for (j = 0; j < pattern->p->len && i < str->p->len; j++,i++) {
	    uint16 c2 = pattern->p->a.words[j];
	    uint16 c1 = str->p->a.words[i];
		
	    if (c1 != c2)
		break;
	}
	if (j >= pattern->p->len) {
	    dprint(83,(debugfile,
		       "cs_find_pattern_from_utf=1 MATCH\n"));
	    ret = 1;
	    break;
	}
	i = s;
    }
    if (!ret) {
	dprint(83,(debugfile,
		   "cs_find_pattern_from_utf=0 NO MATCH\n"));	
    }
    return ret;
}

static int cs_add_streambytes_to_utf8 P_((struct string *str, 
					  int count, 
					  const unsigned char *data));
static int cs_add_streambytes_to_utf8(str,count,data)
     struct string *str;
     int count; 
     CONST unsigned char *data;
{
 
    if (count > 0) {
	int i;
	/* realloc with size 0 is equivalent of free and may 
	   corrupt memory ...
	*/
	
	if (!str->p->state)
	    str->p->state = new_state_1(str->string_type);


	/* NOTE:  str->p->a.words is not NUL terminated        */
	str->p->a.words = safe_realloc(str->p->a.words,
				       (str->p->len+count)* sizeof (uint16));


	for (i = 0; i < count; i++) {
	    if (!cs_add_streambyte_to_s_utf8(str->p->state,data[i]))
		break;

	    if (str->p->state->p->ready) {
		if (str->p->state->p->a.utf8.value > 0xFFFF)
		    break;
		str->p->a.words[str->p->len++] = 
		    str->p->state->p->a.utf8.value;
		cs_soft_reset_s_utf8(str->p->state);
	    }
	}
	return i;	
    }
    return 0;
}

static struct  map_info * cs_find_utf8 P_((char * map_name));
static struct  map_info * cs_find_utf8(map_name)
     char * map_name;
{
    int i;
    struct  map_info *ret;

    static struct  map_info ** dyn_maps;
    static int dyn_map_count = 0;

    if (0 == istrcmp(map_name,map_utf8.map_name))
	return &map_utf8;

    for (i =0; i < dyn_map_count; i++)
	if (0 == strcmp(dyn_maps[i]->map_name,map_name))
	    return dyn_maps[i];

    ret = open_utf8_map(map_name);
    if (ret) {
	dyn_maps = safe_realloc(dyn_maps,
				(dyn_map_count + 1) * 
				sizeof (struct  map_info *));
	dyn_maps[dyn_map_count++] = ret;
    }
    return ret;
}

static void cs_remove_control_utf8 P_((const struct string *str));
static void cs_remove_control_utf8(str)
     CONST struct string *str;
{
    int i;
    
    for (i = 0; i < str->p->len; i++) {
	if (str->p->a.words[i] < 32 ||
	    str->p->a.words[i] == 127)
	    str->p->a.words[i] = 32;
	else {
	    uint16 unicode = map_utf(str->string_type->map_info,
				     str->p->a.words[i]);
	    if (MAPPING_NONE == unicode ||
		!unicode_ch(unicode,UOP_noctrl))
		str->p->a.words[i] = 32;	    
	}
    }
}

static void cs_add_state_to_utf P_((struct string *str, 
				    struct charset_state *ch));  
static void cs_add_state_to_utf(str,ch)
     struct string *str;
     struct charset_state *ch;
{
    if (ch->p->a.utf8.value > 0xFFFF)
	return;
    
    /* NOTE:  str->p->a.words is not NUL terminated        */
    str->p->a.words = safe_realloc(str->p->a.words,
				   (str->p->len+1)* sizeof (uint16));
    
    str->p->a.words[str->p->len++] = ch->p->a.utf8.value;
}  

static void cs_init_s_utf8 P_((struct charset_state *str));
static void cs_init_s_utf8(str)
     struct charset_state *str;
{
    str->p->a.utf8.bytes   = 0;
    str->p->a.utf8.idx     = 0;
    str->p->a.utf8.value   = 0;
    str->p->ready          = 0;
}

static void cs_free_s_utf8 P_((struct charset_state *str));
static void cs_free_s_utf8(str)
     struct charset_state *str;
{
    str->p->a.utf8.bytes   = 0;
    str->p->a.utf8.idx     = 0;
    str->p->a.utf8.value   = 0;
    str->p->ready          = 0;
}

static int split_utf8_byte P_((int c, unsigned char *count, 
			       unsigned char *val));
static int split_utf8_byte(c,count,val)
     int c; 
     unsigned char *count; 
     unsigned char *val;
{
    if (c < 0 || c > 0xFF)
	panic("STRING PANIC",__FILE__,__LINE__,"split_utf8_byte",
	      "Value not in range 0-255",0);

    if (c < 0x80) {
	*count = 0;
	*val   = c;
	return 1;
    } else {
	int counter = 0;
	int mask    = 0x80;

	for (mask=0x80, counter = 0;
	     (mask & c) != 0 && mask != 0;
	     mask >>= 1, counter++) {
	    c &= ~mask;
	}
	*count = counter;
	*val   = c;
	return (mask != 0);
    }
}

static int cs_add_streambyte_to_s_utf8(str,ch)
     struct charset_state *str; 
     int ch;
{
    unsigned char count;
    unsigned char val;

    if (!split_utf8_byte(ch,&count,&val)) {
    failure:
	str->p->a.utf8.bytes   = 0;
	str->p->a.utf8.idx     = 0;
	str->p->a.utf8.value   = 0;
	str->p->ready          = 0;
	return 0;
    }
    if (str->p->a.utf8.idx == 0) {
	if (count == 1)
	    goto failure;
	str->p->a.utf8.bytes  = count;
	str->p->a.utf8.idx    = 1;
	str->p->a.utf8.value  = val;
	if (0 == count)           /* Ascii character? */
	    str->p->ready = 1;
    } else if (count != 1) {
	goto failure;
    } else {
	str->p->a.utf8.idx++;
	str->p->a.utf8.value <<= 6;
	str->p->a.utf8.value  |= val;
    }

    if (str->p->a.utf8.idx == str->p->a.utf8.bytes)
	str->p->ready = 1;
    return 1;
}

static void cs_soft_reset_s_utf8(str)
     struct charset_state *str;
{
    str->p->a.utf8.bytes   = 0;
    str->p->a.utf8.idx     = 0;
    str->p->a.utf8.value   = 0;
    str->p->ready          = 0;
}

static uint16 cs_give_unicode_from_s_utf8 P_((struct charset_state *st,
					      int *found));
static uint16 cs_give_unicode_from_s_utf8(st,found)
     struct charset_state *st;
     int *found;
{
    uint16 val;

    *found = 0;
    val = map_utf(st->charset->map_info,st->p->a.utf8.value);
	
    if (val == MAPPING_NONE)
	val = 0x003F;  /* '?' */
    else
	*found = 1;
    return val;
}

static int cs_s_utf8_same_char P_((struct charset_state *A,
				   struct charset_state *B,
				   int ignore_case));
static int cs_s_utf8_same_char(A,B,ignore_case)
     struct charset_state *A;
     struct charset_state *B;
     int ignore_case;
{
    
    if (A->p->a.utf8.value == B->p->a.utf8.value)
	return 1;

    if (ignore_case) 
	return -1;  /* Use UNICODE values for comparision */
    
    return 0;
}

static int cs_s_utf8_printable P_((struct charset_state *st));
static int cs_s_utf8_printable(st)
     struct charset_state *st;
{
    uint16 val;
    
    val = map_utf(st->charset->map_info,st->p->a.utf8.value);

    return unicode_ch(val,UOP_printable) != 0;
}

/* If character corresponds one byte on stream, returns it.
 * Otherwise returns 0. This is used implement ReadCh().
 * It is assumed that returned character corresponds to
 * code characterter set (and perhaps also US-ASCII)
 */
static int cs_s_utf8_is_onebyte P_((struct charset_state *st));
static int cs_s_utf8_is_onebyte(st)
     struct charset_state *st;
{
    if (st->p->a.utf8.bytes == 0)
	return st->p->a.utf8.value;
    return 0;
}

struct charset_type cs_utf8    =  { "utf-8",
				     cs_init_utf,
				     cs_free_utf,
				     cs_add_streambyte_to_utf8,
				     cs_add_intdata_to_utf,
				     cs_give_unicode_from_utf8,
				     cs_add_unicodedata_to_utf,
				     cs_cmp_utf,
				     cs_stream_from_utf8,
				     cs_can_ascii_utf,
				     cs_streamclip_from_utf8,
				     cs_clip_from_utf,
				     cs_find_pattern_from_utf,
				     cs_add_streambytes_to_utf8,
				     cs_find_utf8,
				     cs_remove_control_utf8, 
				     cs_add_state_to_utf,
				     cs_init_s_utf8,
				     cs_free_s_utf8,
				     cs_add_streambyte_to_s_utf8,
				     cs_soft_reset_s_utf8,
				     cs_give_unicode_from_s_utf8,
				     cs_s_utf8_same_char,
				     cs_s_utf8_printable,
				     cs_s_utf8_is_onebyte,
				     NULL
};


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
