static char rcsid[] = "@(#)$Id: imap.c,v 1.3 2000/07/30 15:12:10 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "mbx_imp.h"
#include "s_me.h"
#include "s_elm.h"

#ifdef REMOTE_MBX

#include <errno.h>
extern int errno;


static void add_token P_((struct Imap_Token_Buffer *X, struct imap_token T));
static void add_token(X,T)
     struct Imap_Token_Buffer *X; 
     struct imap_token T;
{
    X->tokens = safe_realloc(X->tokens,  
			     (X->token_count+1) * sizeof (struct imap_token));
    X->tokens[X->token_count++] = T;
}

static void free_tokens P_((struct Imap_Token_Buffer *X));
static void free_tokens(X)
     struct Imap_Token_Buffer *X;
{

    if (X -> tokens) {
	int i;
	
	for (i = 0; i < X -> token_count; i++) {
	    if (X -> tokens[i].str) {
		free (X -> tokens[i].str);
		X -> tokens[i].str = NULL;
	    }
	    X -> tokens[i].imap_token = imap_zero;
	}
	
	free(X -> tokens);
	X -> tokens = NULL;
	X -> token_count = 0;
    }
}

static struct kwd {
    char         *   kw;
    enum token_type  t;
} keywords[] = {
    { "OK",       imap_status_keyword },
    { "NO",       imap_status_keyword },
    { "BAD",      imap_status_keyword },
    { "PREAUTH",  imap_status_keyword },
    { "BYE",      imap_status_keyword }
};

#define CAPA_IMAP4rev1           1

static struct capab {
    char         *    capa;
    long              mask;
} capabilities[] = {
    { "IMAP4rev1",   CAPA_IMAP4rev1 }
};

#define  IMAP_Seen               1
#define  IMAP_Answered           2
#define  IMAP_Flagged            4
#define  IMAP_Deleted            8
#define  IMAP_Draft             16
#define  IMAP_Recent            32

static struct imap_flag {
    char         *   flag;
    int              mask_elm;
    long             mask_imap;
} IMAP_flags[] = {
    { "\\Seen",        0,           IMAP_Seen     },
    { "\\Answered",    REPLIED,     IMAP_Answered },
    { "\\Flagged",     TAGGED,      IMAP_Flagged  },
    { "\\Deleted",     DELETED,     IMAP_Deleted  },
    { "\\Draft",       0,           IMAP_Draft  },
    { "\\Recent",      NEW,         IMAP_Recent  },
};

static void add_to_write_buffer P_((struct IMAP_MBX * M,
				    char **str));
static void add_to_write_buffer(M,str)
     struct IMAP_MBX * M;
     char **str;
{
    int l = strlen(*str);
    if (!M->write_len) {
	if (M->write_buffer) 
	    free(M->write_buffer);
	M->write_buffer = *str;
	M->write_len    = l;
    } else if (l > 0) {
	M->write_buffer = safe_realloc(M->write_buffer,
				       M->write_len+l);
	memmove(M->write_buffer + M->write_len,*str,l);
	M->write_len += l;
	free(*str);
	*str = NULL;
    }
}

static int imap_gen_writable_data P_((struct IMAP_MBX * M));
static int imap_gen_writable_data(M)
     struct IMAP_MBX * M;
{
    char * t;

    if (M->write_len > 0) {
	/* Have incomplete data ... */
	return 1;
    }

    if (M->write_buffer) {
	free(M->write_buffer);
	M->write_buffer = NULL;
    }

    if (M->imap_state != IMAP_command)
	return 0;    /* No data available */

    if (M->token_index >= M->write_tokens.token_count)
	return 0;    /* No data available */

    switch(M->token_index) {
    case -2 :        /* NOOP command */
	M->write_buffer = elm_message (FRM("%s NOOP\r\n"),
				       M->current_NOOP_tag);
	M->write_len    = strlen(M->write_buffer);
	return 1;

    case -1:         /* Normal command */
	M->write_buffer = elm_message(FRM("%s "),
				      M->current_tag);
	M->write_len    = strlen(M->write_buffer);
	M->literal_mode = no_literal;
	M->token_index++;
    }

    while (M->token_index < M->write_tokens.token_count) {
	char *s1 = NULL;
	char * space = "";
	if (M->token_index > 0 &&
	    M->write_tokens.tokens[M->token_index-1].imap_token != 
	    imap_list_begin)
	    space = " ";
	
	switch(M->write_tokens.tokens[M->token_index].imap_token) {
	case imap_literal:
	    switch(M->literal_mode) {
		char *s;
	    case continue_literal:
		if (M->write_buffer != NULL)
		    panic("MBX PANIC",__FILE__,__LINE__,
			  "imap_gen_writable_data",
			  "write_buffer not NULL",0);
		M->write_buffer = M->write_tokens.tokens[M->token_index].str;
		M->write_len    = M->write_tokens.tokens[M->token_index].
		    len_or_value;

		/* So that it is not free()ed twice */
		M->write_tokens.tokens[M->token_index].str = NULL; 

		M->literal_mode = no_literal;
		M->token_index++;
		continue;       /* Write literal */            
	    case waiting_literal_continue:
		dprint(13, (debugfile,
			    "imap_gen_writable_data=0:  Waiting literal continue\n"));
		return 0;
	    case no_literal:
		s = elm_message(FRM("%s{%d}\r\n"),
				space,
				M->write_tokens.tokens[M->token_index].
				len_or_value);
		add_to_write_buffer(M,&s);  /* May free() s */
		M->literal_mode = waiting_literal_continue;
		return 1;
	    }
	    panic("MBX PANIC",__FILE__,__LINE__,
		  "imap_gen_writable_data",
		  "Bad literal mode",0);
	case imap_number:
	    s1 = elm_message(FRM("%s%d"),space,
			     M->write_tokens.tokens[M->token_index].
			     len_or_value);
	    M->token_index++;
	    break;
	case imap_list_begin:
	    s1 = elm_message(FRM("%s("),space);
	    M->token_index++;
	    break;
	case imap_list_end:
	    s1 = elm_message(FRM(")"));
	    M->token_index++;
	    break;
	case imap_atom:
	    s1 = elm_message(FRM("%s%s"),space,
			     M->write_tokens.tokens[M->token_index].str);
	    M->token_index++;
	    break;
	case imap_string:
	    if (strspn(M->write_tokens.tokens[M->token_index].str,
		       " 1234567890+-.!*,()[]#_<>qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM") < 
		M->write_tokens.tokens[M->token_index].len_or_value) {
		dprint(13, (debugfile,
			    "imap_gen_writable_data: %d string \"%s\" converting to literal...\n",
			    M->token_index,
			    M->write_tokens.tokens [M->token_index].str));
		M->write_tokens.tokens[M->token_index].imap_token =
		    imap_literal;
		break;   /* RETRY */
	    }
	    s1 = elm_message(FRM("%s\"%s\""),space,
			     M->write_tokens.tokens[M->token_index].str);
	    M->token_index++;
	    break;
	default:
	    panic("MBX PANIC",__FILE__,__LINE__,
		  "imap_gen_writable_data",
		  "Bad token type",0);
	}

	if (s1)
	    add_to_write_buffer(M,&s1);  /* May free() s1 */
    }

    /* Write end of command (CRLF) */
    t = safe_strdup("\r\n");
    add_to_write_buffer(M,&t);  /* May free() t */

    return 1;  /* have writable data */
}

/* -1 == failure
   0  == continue   (read more data)
   1  == ready
*/
static int tokenize P_((struct IMAP_MBX * M));
static int tokenize(M)
     struct IMAP_MBX * M;
{
    int idx;
    int len;
    int ret;

    if (M->wanna_literal) {
	struct imap_token T;

    try_literal:
	if (M->tokenizer_state != state_list)
	    panic("MBX PANIC",__FILE__,__LINE__,"tokenize",
		  "Bad tokenizer_state (IMAP)",0);

	if (M->read_buffer.read_len < M->literal_len) {
	    dprint(13, (debugfile,"tokenize=0 (incomplete literal)\n"));
	    return 0;
	}

	T.imap_token = imap_literal;
	T.str = safe_malloc(M->literal_len+1);
	memcpy(T.str,M->read_buffer.read_buffer,M->literal_len);
	T.str[M->literal_len] = '\0';
	T.len_or_value = M->literal_len;
	add_token(&(M->read_tokens),T);
	cut_line(&(M->read_buffer),M->literal_len);
	M->wanna_literal = 0;
    }

    idx = 0;
    ret = 1;
    len = find_crlf(&(M->read_buffer),1);
    if (!len) {
	dprint(13, (debugfile,"tokenize=0 (no data)\n"));
	return 0;
    }

    while (idx < len) {	
	struct imap_token T;
	char * p = M->read_buffer.read_buffer +idx;
	int left = len - idx;
	
	if (!*p) {
	    dprint(50, (debugfile,"tokenize: idx=%d   END\n",idx));
	    break;
	}

	T.len_or_value = 0;
	T.str          = NULL;
	T.imap_token   = imap_zero;

	switch (M->tokenizer_state) {
	    char  *n;
	    int i;
	case state_tag:
	    n = memchr(p,' ',left);
	    
	    if (NULL == n || n == p) {
	    fail:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnparseableResponse,
				  "Unparseable response from IMAP server"));

		free_Read_Buffer(&(M->read_buffer));		
		dprint(15, (debugfile,"tokenize=-1 (failure)\n"));
		return -1;
	    }
	    *n = '\0';

	    dprint(50, (debugfile,
			"tokenize: idx=%d  TAG   '%.*s'\n",
			idx,n-p,p));

	    M->tokenizer_state = state_response;	    
	    if (n-p == 1 && p[0] == '+') {
		T.imap_token = imap_continue;
		M->tokenizer_state = state_text;
	    } else if (n-p == 1 && p[0] == '*')
		T.imap_token = imap_notag;
	    else {
		T.imap_token = imap_atom;
		T.str = safe_malloc(n-p+1);
		memcpy(T.str,p,n-p+1);
		T.len_or_value = n-p; 
	    }
	    add_token(&(M->read_tokens),T);

	    idx = (n - M->read_buffer.read_buffer) +1;
	    break;  
	case state_response:
	    if (p[0] >= '0' && p[0] <= '9') {
		char *c;
		T.len_or_value = strtol(p,&c,10);
		if (*c == ' ') {

		    dprint(50, (debugfile,
				"tokenize: idx=%d  RESPONSE (number) '%.*s'\n",
				idx,c-p,p));
		
		    T.imap_token = imap_number;
		    add_token(&(M->read_tokens),T);
		    idx = (c - M->read_buffer.read_buffer) +1;
		    break;
		}
	    }
	    n = memchr(p,' ',left);
	    if (n == p) 
		goto fail;
	    
	    if (NULL == n)
		n = p + left;
	    else 
		*n = '\0';
		
	    dprint(50, (debugfile,
			"tokenize: idx=%d  RESPONSE (keyword) '%.*s'\n",
			idx,n-p,p));

	    for (i = 0; i < sizeof keywords / sizeof (struct kwd); i++)
		if (0 == istrcmp(keywords[i].kw,p))
		    break;
	    
	    if (i < sizeof keywords / sizeof (struct kwd))
		T.imap_token = keywords[i].t;
	    else
		T.imap_token = imap_other_keyword;
	    
	    if (imap_status_keyword == T.imap_token)
		M->tokenizer_state = state_text;
	    else if (imap_other_keyword == T.imap_token)
		M->tokenizer_state = state_list;

	    T.str = safe_malloc(n-p+1);
	    memcpy(T.str,p,n-p+1);
	    T.len_or_value = n-p; 
	    add_token(&(M->read_tokens),T);

	    idx = (n - M->read_buffer.read_buffer) +1;
	    break;
	case state_text:
	    if ('[' == p[0]) {

		dprint(50, (debugfile,
			    "tokenize: idx=%d  TEXT   [\n",
			    idx));
		
		T.imap_token = imap_code_begin;
		M->tokenizer_state = state_text_code;
		idx++;
		add_token(&(M->read_tokens),T);
		break;
	    }	    

	    dprint(50, (debugfile,
			"tokenize: idx=%d  TEXT   '%.*s'\n",
			idx,left,p));

	    T.imap_token = imap_string;
	    T.str = safe_malloc(left);
	    memcpy(T.str,p,left);
	    T.len_or_value = left-2;   /* Not find_crlf() */

	    add_token(&(M->read_tokens),T);

	    idx += left;
	    break;
	case state_text_code:
	    if (']' == p[0]) {

		dprint(50, (debugfile,
			    "tokenize: idx=%d  CODE   ]\n",
			    idx));

		T.imap_token = imap_code_end;
		M->tokenizer_state = state_text;
		idx++;
		if (' ' == p[1])   /* Skip separated space between code */
		    idx++;         /* argument and textual code         */
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    /* FALLTHRU */
	case state_list:
	    if ('(' == p[0]) {

		dprint(50, (debugfile,
			    "tokenize: idx=%d  LIST   (\n",
			    idx));
		
		T.imap_token = imap_list_begin;
		idx++;
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    if (')' == p[0]) {

		dprint(50, (debugfile,
			    "tokenize: idx=%d  LIST   )\n",
			    idx));

		T.imap_token = imap_list_end;
		idx++;
		if (' ' == p[1])   /* Skip separated space between list */
		    idx++;         /* argument and next argument        */
		add_token(&(M->read_tokens),T);
		break;
	    }	    
	    if (p[0] >= '0' && p[0] <= '9') {
		char *c;
		T.len_or_value = strtol(p,&c,10);
		if (*c == ' ' || *c == ')' || *c == ']') {
		    
		    dprint(50, (debugfile,
				"tokenize: idx=%d  LIST (number)   '%.*s'\n",
				idx,c-p,p));
		    
		    T.imap_token = imap_number;
		    add_token(&(M->read_tokens),T);
		    if (*c == ' ')
			idx = (c - M->read_buffer.read_buffer) +1;
		    else
			idx = (c - M->read_buffer.read_buffer);
		    break;
		}
	    }

	    if (M->tokenizer_state != state_text_code) {
		if ('"' == p[0]) {
		    char * x, *y;
		    T.imap_token = imap_string;
		    T.str = safe_malloc(left);
		    y = T.str;

		    for (x = p+1; x < p + left && '\0' != *x && '"' != *x; 
			 x++) {
			if ('\\' == *x)
			    x++;
			if (y < T.str + left -1)
			    *y++ = *x;
		    }
		    if (*x != '"')
			goto fail;
		    x++;
		    *y = '\0';

		    dprint(50, (debugfile,
				"tokenize: idx=%d  LIST (string)   '%.*s'\n",
				idx,y-T.str,T.str));

		    T.len_or_value = y-T.str; 
		    add_token(&(M->read_tokens),T);
		    
		    if (' ' == *x) /* Skip separated space between string */
			x++;       /* argument and next argument          */
		    idx = x - M->read_buffer.read_buffer;		
		    break;
		}
		if (p[0] == '{' && left > 4 && p[left-3] == '}' &&
		    M->tokenizer_state == state_list) {
		    char *c;
		    M->literal_len = strtol(p+1,&c,10);
		    if (c == &(p[left-3])) {   /* Start parsing literal */

			dprint(50, (debugfile,
				    "tokenize: idx=%d  LIST (literal)   len=%d\n",
				    idx,M->literal_len));
			
			M->wanna_literal = 1;
			idx += left;
			ret = 0;               /* Continue reading ... */
			break;
		    }
		}
	    }
	    T.imap_token = imap_atom;
	    if (M->tokenizer_state == state_text_code)
		i = strcspn(p," ()]");
	    else
		i = strcspn(p,"(){ *%\"");
	    if (i == 0) {
		dprint(13, (debugfile,
			    "tokenize: Possible internal error:'%s\n",
			    p));
		goto fail;
	    }

	    dprint(50, (debugfile,
			"tokenize: idx=%d  LIST (atom)   '%.*s'\n",
			idx,i,p));
	    
	    T.imap_token = imap_atom;
	    T.str = safe_malloc(i+1);
	    memcpy(T.str,p,i);
	    T.str[i] = '\0';
	    idx += i;
	    if (' ' == M->read_buffer.read_buffer[idx])
		idx++;
	    add_token(&(M->read_tokens),T);
	    break;
	default:
	    panic("MBX PANIC",__FILE__,__LINE__,"tokenize",
		  "Bad tokenizer_state (IMAP)",0);
	}
    }
    
    if (!M->wanna_literal) {
	/* Again beginnging of line */
	M->tokenizer_state = state_tag;
    }
    cut_line(&(M->read_buffer),len);

    if (M->wanna_literal && M->read_buffer.read_len)
	goto try_literal;

    dprint(13, (debugfile,"tokenize=%d (token count %d)\n",
		ret,M->read_tokens.token_count));

    return ret;   /* Response parsed? */
}

static void pick_text P_((struct IMAP_MBX * M, struct string **Text,
			  char **code));
static void pick_text(M,Text,code)
     struct IMAP_MBX * M;
     struct string **Text;
     char **code;
{
    int i;
    int q = 0;
    
    *Text = NULL;
    *code = NULL;

    for (i = 2; i < M->read_tokens.token_count; i++) {
	if (imap_code_begin == M->read_tokens.tokens[i].imap_token)
	    q++;
	else if (imap_code_end == M->read_tokens.tokens[i].imap_token)
	    q--;
	else if (q && imap_atom == M->read_tokens.tokens[i].imap_token &&
		 !*code)
	    *code = M->read_tokens.tokens[i].str;
	else if (!q && imap_string == M->read_tokens.tokens[i].imap_token &&
		 !*Text)
	    /* User readable code may use MIME encoded words on message */
	    *Text = hdr_to_string(HDR_TEXT,
				  M->read_tokens.tokens[i].str,
				  system_charset,1);	    
    }
}

static int scan_list P_((struct imap_token ** ptr,
			 int                * left,
			 struct imap_token ** found_list,
			 int                * list_len));
static int scan_list(ptr,left,found_list,list_len)
     struct imap_token ** ptr;
     int                * left;
     struct imap_token ** found_list;
     int                * list_len;
{
    struct imap_token * p = *ptr;
    int                 l = *left;
    int i,q=0;

    *found_list = NULL;
    *list_len   = 0;

    if (l < 2 || imap_list_begin != p[0].imap_token)
	return 0;

    *found_list = p+1;
    
    for (i = 0; i < l; i++) {
	if (imap_list_begin == p[i].imap_token)
	    q++;
	else if (imap_list_end == p[i].imap_token) {
	    q--;
	    if (0 == q) {
		*list_len = i-1;
		*ptr      = p+  (i +1);
		*left     = l - (i +1);
		return 1;
	    }
	}
    }
    return 0;
}

static void free_reference P_((struct imap_reference *item));
static void free_reference(item)
     struct imap_reference *item;
{
    item->uid_number        = -1;
    item->imap_flags        = 0;
    item->rfc822_size       = 0;
    
    if (item->internaldate) {
	free(item->internaldate);
	item->internaldate  = NULL;
    }

    if (item->header) {
	free(item->header);
	item->header      = NULL;
	item->header_len = 0;	    
    }
    if (item->body) {
	free(item->body);
	item->body      = NULL;
	item->body_len  = 0;    
    }
}

static void zero_reference P_((struct imap_reference *item));
static void zero_reference(item)
     struct imap_reference *item;
{
    item->uid_number        = -1;
    item->imap_flags        = 0;
    item->rfc822_size       = 0;
    item->internaldate      = NULL;
    item->header_len        = 0;
    item->header            = 0;
    item->body_len          = 0;
    item->body              = 00;
}


static int got_header P_((struct imap_reference * item,
			  struct imap_token * lst));
static int got_header(item,lst)
     struct imap_reference * item;
     struct imap_token * lst;
{
    /* Originally malloced by tokenize() */
    if (item->header) {
	free(item->header);
	item->header = NULL;
    }
    switch(lst->imap_token) {
    case imap_atom:
	if (0 == istrcmp(lst->str,"NIL")) {
	    item->header     = NULL;
	    item->header_len = 0;
	    return 1;       /* OK */
	}
	return 0;           /* Bad */
    case imap_string:
    case imap_literal:
	/* Move malloced data to struct imap_reference */
	item->header     = lst->str;
	item->header_len = lst->len_or_value;
	lst->str         = NULL;  /* So that free_tokens() does not try free 
				   * it  
				   */
	return 1;           /* OK */
    }
    return 0;   /* Bad */
}

static int got_body P_((struct imap_reference * item,
			  struct imap_token * lst));
static int got_body(item,lst)
     struct imap_reference * item;
     struct imap_token * lst;
{
    /* Originally malloced by tokenize() */
    if (item->body) {
	free(item->body);
	item->body = NULL;
    }
    switch(lst->imap_token) {
    case imap_atom:
	if (0 == istrcmp(lst->str,"NIL")) {
	    item->body     = NULL;
	    item->body_len = 0;
	    return 1;       /* OK */
	}
	return 0;           /* Bad */
    case imap_string:
    case imap_literal:
	/* Move malloced data to struct imap_reference */
	item->body     = lst->str;
	item->body_len = lst->len_or_value;
	lst->str       = NULL;  /* So that free_tokens() does not try free 
				 * it  
				 */
	return 1;           /* OK */
    }
    return 0;   /* Bad */
}

static void process_fetch_items P_((struct imap_reference * item,
				    struct imap_token * lst,
				    int                 lst_len));
static void process_fetch_items(item,lst,lst_len)
     struct imap_reference * item;
     struct imap_token     * lst;
     int                     lst_len;
{			
    dprint(14, (debugfile,
	       "process_fetch_items: len=%d",
		lst_len));
    while (lst_len > 1) {
	struct imap_token * kwd = lst;
	lst_len--;
	lst++;
	
	if (imap_atom == kwd->imap_token) {
	    if (0 == istrcmp(kwd->str,"UID")) {
		if (imap_number == lst->imap_token) {
		    item->uid_number = lst->len_or_value;

		    dprint(14, (debugfile," UID %d",item->uid_number));

		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," UID ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after UID number expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"INTERNALDATE")) {
		if (imap_string == lst->imap_token) {
		    dprint(14, (debugfile," INTERNALDATE %s (len %d)",
				lst->str,lst->len_or_value));
		    if (lst->len_or_value != 26) {
			dprint(14, (debugfile,"\n"));
			dprint(1, (debugfile,
				   "process_fetch_items: Fetch response  (..) after INTERNALDATE 26 char string expected, got %d chars\n",
				   lst->len_or_value));
		    } else {
			if (item->internaldate)
			    free(item->internaldate);
			item->internaldate = lst->str;
			lst->str          = NULL;   /* Avoid double free() */
		    }
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," INTERNALDATE ????\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after INTERNALDATE string expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.SIZE")) {
		if (imap_number == lst->imap_token) {
		    item->rfc822_size = lst->len_or_value;

		    dprint(14, (debugfile," RFC822.SIZE %d",
				item->rfc822_size));
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," RFC822.SIZE ????\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after RFC822.SIZE number expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.HEADER")) {
		if (got_header(item,lst)) {		    
		    dprint(14, (debugfile," RFC822.HEADER ..."));
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," RFC822.HEADER ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after RFC822.HEADER nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"RFC822.TEXT")) {
		if (got_body(item,lst)) {		    
		    dprint(14, (debugfile," RFC822.TEXT ..."));
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," RFC822.TEXT ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after RFC822.BODY nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"BODY[HEADER]")) {
		if (got_header(item,lst)) {		    
		    dprint(14, (debugfile," BODY[HEADER] ..."));
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," BODY[HEADER] ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after BODY[HEADER] nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"BODY[TEXT]")) {
		if (got_body(item,lst)) {		    
		    dprint(14, (debugfile," BODY[TEXT] ..."));
		    lst_len--;
		    lst++;
		} else {
		    dprint(14, (debugfile," BODY[TEXT] ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after BODY[TEXT] nstring expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else if (0 == istrcmp(kwd->str,"FLAGS")) {
		struct imap_token * flg;
		int                 flg_len;
		
		item->imap_flags = 0;
		
		dprint(14, (debugfile," FLAGS "));
		if (scan_list(&lst,&lst_len,&flg,&flg_len)) {
		    int i, j;
		    
		    dprint(14, (debugfile,"(listlen=%d)",flg_len));
		    for (i = 0; i < flg_len; i++) {
			if (imap_atom == flg[i].imap_token) {
			    int found = 0;
			    for (j = 0; 
				 j < sizeof IMAP_flags / 
				     sizeof (struct imap_flag);
				 j++) {
				if (0 == istrcmp(flg[i].str,
						 IMAP_flags[j].flag)) {
				    item->imap_flags |= 
					IMAP_flags[j].mask_imap;
				    found++;
				    break;
				}
			    }
			    if (!found) {
				dprint(14,(debugfile,
					   " [Unknown flag %s] ",
					   flg[i].str));
			    }
			}			
		    }
		    dprint(15, (debugfile,
			    " [known flags:"));
		    for (j = 0;
			 j < sizeof IMAP_flags / sizeof (struct imap_flag);
			 j++) {
			if ((item->imap_flags & IMAP_flags[j].mask_imap)
			    == IMAP_flags[j].mask_imap)
			    dprint(15, (debugfile," %s",IMAP_flags[j].flag));
		    }
		    dprint(15, (debugfile,"]"));
		} else {
		    dprint(14, (debugfile," ???\n"));
		    dprint(1, (debugfile,
			       "process_fetch_items: Fetch response  (..) after FLAGS (list) expected, got type %d\n",
			       lst->imap_token));
		    break; /* Quit processing */
		}
	    } else {
		dprint(14, (debugfile," ???\n"));
		dprint(1, (debugfile,
			   "process_fetch_items: Fetch response  (..) list item %s unsupported\n",
			   kwd->str));
		break; /* Quit processing */
	    }				    
	} else {
	    dprint(14, (debugfile," ???\n"));
	    dprint(1, (debugfile,
		       "process_fetch_items: Fetch response  (..) list item have token type %d\n",
		       kwd->imap_token));
	    break;	/* Quit processing */	
	}
    }    
    dprint(14, (debugfile,"\n"));
}

static int imap_read_socket P_((int fd,void * data));

static void imap_clear_command P_((struct folder_info *folder));
static void imap_clear_command(folder)
     struct folder_info *folder;
{
    struct IMAP_MBX * CONST M = &(folder->p->a.imap_mbx);

    dprint(12, (debugfile,
		"imap_clear_command: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (IMAP_idle == M->imap_state) {
	dprint(12, (debugfile,
		    "imap_clear_command: Already clear!\n"));
    } else if (IMAP_closing == M->imap_state) {
	dprint(12, (debugfile,
		    "imap_clear_command: On closing state!\n"));
    } else if (IMAP_command_ready_NO  != M->imap_state &&
	       IMAP_command_ready_BAD != M->imap_state &&
	       IMAP_command_ready_OK  != M->imap_state) {
	dprint(12, (debugfile,
		    "imap_clear_command: WRONG STATE!\n"));
	M->imap_state = IMAP_error;
    } else
	M->imap_state = IMAP_idle;

    if (M->current_NOOP_tag &&
	M->imap_state != IMAP_error) {
	dprint(12, (debugfile,
		    "imap_clear_command: Waiting for idle NOOP complete...\n"));
	while(M->current_NOOP_tag && 
	      M->imap_state != IMAP_error) {
	    wait_for_action(imap_read_socket);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}

	dprint(12, (debugfile,
		    "imap_clear_command: idle NOOP wait completed\n"));
    }
    
    /* Reset possible incomplete send literal */
    free_tokens(& (M->write_tokens));
    M->literal_mode = no_literal;
    M->token_index  = -1;

    if (M->write_buffer) {
	free(M->write_buffer);
	M->write_buffer = NULL;
    }

    M->write_len = 0;
    
    if (M->current_tag) {
	free(M->current_tag);
	M->current_tag = NULL;
    }

    if (M->command_result_code) {
	free(M->command_result_code);
	M->command_result_code = NULL;
    }
}    

#define IMAP_TIMEOUT  25

static struct imap_reference  *find_reference P_((struct IMAP_MBX *M, 
						  struct imap_uid_ref *UID));
static struct imap_reference  *find_reference(M,UID)
     struct IMAP_MBX *M;
     struct imap_uid_ref *UID;
{
    int i;
    if (UID->last_mbx_index > 0 && UID->last_mbx_index < M->reference_count &&
	M->references[UID->last_mbx_index].uid_number ==
	UID->uid_number) {

	/* Found via cache */
	return &(M->references[UID->last_mbx_index]);
    }

    for (i = 1; i < M->reference_count; i++)
	if (M->references[i].uid_number == UID->uid_number) {
	    UID->last_mbx_index = i;  /* Update cache */
	    return &(M->references[UID->last_mbx_index]);
	}

    dprint(1, (debugfile,"find_reference: No mail with IMAP uid %d\n",
	       UID->uid_number));

    UID->last_mbx_index = -1;  /* Invalidate cache */
    return NULL;
}


/* Special return value 2 indivates that caller should re-enable 
 * imap_write_socket()
 */
static int parse_response P_((struct IMAP_MBX * M));
static int parse_response(M)
     struct IMAP_MBX * M;
{
    if (M->read_tokens.token_count < 2) {
	dprint(1, (debugfile,"parse_response: token_count=%d\n",
		   M->read_tokens.token_count));
	lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			  "Bad response from IMAP server"));
	return 0;
    }

    switch (M->read_tokens.tokens[0].imap_token) {
    case imap_notag:
	switch (M->read_tokens.tokens[1].imap_token) {
	    struct string *Text;
	    char *code;   /* Do not free() this -- this is just pointer
			     to read_tokens area
			  */
	case imap_status_keyword:
	    Text = NULL;
	    code = NULL;
	    pick_text(M,&Text,&code);
	    if (0 == istrcmp(M->read_tokens.tokens[1].str,"BYE")) {
		M->imap_state = IMAP_closing;
		if (Text)
		    lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapBye,
					  "IMAP server: %S"),
				  Text);		
	    } else if (code && Text && 0 == istrcmp(code,"ALERT")) {
		lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapAlert,
				      "IMAP alert: %S"),
			      Text);		
	    }
	    if (Text)
		free_string(&Text);
	    break;
	case imap_other_keyword:

	    if (0 == istrcmp(M->read_tokens.tokens[1].str,"CAPABILITY")) {
		int i,j;
		M->capability_bits = 0;

		for (i = 2; i < M->read_tokens.token_count; i++) {
		    if (imap_atom == M->read_tokens.tokens[i].imap_token) {
			for (j = 0; 
			     j < sizeof capabilities / sizeof (struct capab);
			     j++) {
			    if (0 == istrcmp(M->read_tokens.tokens[i].str,
					     capabilities[j].capa)) 
				M->capability_bits |= capabilities[j].mask;
			}
		    }
		}
		dprint(14, (debugfile,
			   "parse_response: (known) Capabilities:"));
		for (j = 0;
		     j < sizeof capabilities / sizeof (struct capab);
		     j++) {
		    if ((M->capability_bits & capabilities[j].mask)
			== capabilities[j].mask)
			dprint(14, (debugfile," %s",capabilities[j].capa));
		}
		dprint(14, (debugfile,"\n"));

	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"LIST")) {
		/* TODO: Handle LIST */
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"LSUB")) {
		/* TODO: Handle LSUB */
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"STATUS")) {
		/* TODO: Handle STATUS */
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"SEARCH")) {
		/* TODO: Handle STATUS */
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"FLAGS")) {
		int i,j;
		M->flag_bits = 0;

		for (i = 2; i < M->read_tokens.token_count; i++) {
		    int found = 0;
		    if (imap_atom == M->read_tokens.tokens[i].imap_token) {
			for (j = 0; 
			     j < sizeof IMAP_flags / sizeof (struct imap_flag);
			     j++) {
			    if (0 == istrcmp(M->read_tokens.tokens[i].str,
					     IMAP_flags[j].flag)) {
				M->flag_bits |= IMAP_flags[j].mask_imap;
				found++;
				break;
			    }			
			}
			if (!found) {
			    dprint(14,(debugfile,
				       "parse_response: Unknown flag %s\n",
				       M->read_tokens.tokens[i].str));
			}
		    }
		}
		dprint(14, (debugfile,
			    "parse_response: (known) Flags:"));
		for (j = 0;
		     j < sizeof IMAP_flags / sizeof (struct imap_flag);
		     j++) {
		    if ((M->flag_bits & IMAP_flags[j].mask_imap)
			== IMAP_flags[j].mask_imap)
			dprint(14, (debugfile," %s",IMAP_flags[j].flag));
		}
		dprint(14, (debugfile,"\n"));
	    } 
	    break;
	case imap_number:
	    if (imap_other_keyword == M->read_tokens.tokens[2].imap_token) {

		if (0 == istrcmp(M->read_tokens.tokens[2].str,"RECENT")) {
		    M->num_recent = M->read_tokens.tokens[1].len_or_value;

		} else  if (0 == istrcmp(M->read_tokens.tokens[2].str,
					 "EXISTS")) {
		    int val = M->read_tokens.tokens[1].len_or_value ;

		    /*
		      M->reference_count == number of messages + 1
		      M->references [0]     is not used
		    */

		    if (val+1 < M->reference_count) {
			dprint(1, (debugfile,
				   "parse_response: Bad EXISTS value: %d\n",
				   val));
		    } else if (val+1 > M->reference_count) {
			dprint(13, (debugfile,"parse_response: %d mails\n",
				    val));
			M->references = safe_realloc(M->references,
						     (val +1) *
						     sizeof (struct 
							     imap_reference));
			
			while (M->reference_count < val+1) {
			    struct imap_reference *item = 
				&(M->references[M->reference_count]);
			    zero_reference (item);
			    M->reference_count++;
			}
		    }

		} else  if (0 == istrcmp(M->read_tokens.tokens[2].str,
					 "EXPUNGE")) {
		    int val = M->read_tokens.tokens[1].len_or_value ;

		    /* M->references[0] is not used ...
		       M->references[ 1 .. num of messages ] is on use

		       M->reference_count == num of messages +1
		    */

		    if (val < 0 || val >= M->reference_count) {
			dprint(1, (debugfile,"parse_response: Bad EXPUNGE value: %d\n",
				   val));
		    } else {
			int i;
			free_reference (& (M->references[val]));

			for (i = val+1; i < M->reference_count; i++) {
			    M->references[i-1] = M->references[i];
			}
			M->reference_count--;			
		    }

		} else  if (0 == istrcmp(M->read_tokens.tokens[2].str,
					 "FETCH")) {
		    int val = M->read_tokens.tokens[1].len_or_value ;

		    /* M->references[0] is not used ...
		       M->references[ 1 .. num of messages ] is on use
		       
		       M->reference_count == num of messages +1
		    */

		    if (val < 0 || val >= M->reference_count) {
			dprint(1, (debugfile,"parse_response: Bad FETCH value: %d\n",
				   val));
		    } else {
			struct imap_token * ptr  = M->read_tokens.tokens+3;
			int                left = M->read_tokens.token_count-3;
			struct imap_token * lst;
			int                 lst_len;
			
			if (!scan_list(&ptr,&left,&lst,&lst_len) ||
			    left > 0) {
			    dprint(1, (debugfile,
				       "parse_response: scan_list failed to parse FECT response list\n"));
			    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
					      "Bad response from IMAP server"));
			    return 0;
			} else {
			    process_fetch_items(M->references+val,
						lst,lst_len);
			}
		    }
		} else {
		    dprint(1, (debugfile,
			       "parse_response: Unsupported data %s\n",
			       M->read_tokens.tokens[2].str));
		}
	    } else {
		dprint(1, (debugfile,"parse_response: token[2] type =%d\n",
			   M->read_tokens.tokens[2].imap_token));
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
				  "Bad response from IMAP server"));
		return 0;
	    }
	    break;
	default:
 	    dprint(1, (debugfile,"parse_response: token[1] type =%d\n",
		       M->read_tokens.tokens[1].imap_token));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	break;
    case imap_atom:
	if (M->current_NOOP_tag &&
	    0 == strcmp(M->current_NOOP_tag,
			M->read_tokens.tokens[0].str)) {
	    free(M->current_NOOP_tag);
	    M->current_NOOP_tag = NULL;

 	    dprint(14, (debugfile,
			"parse_response: Response for idle timer NOOP command -- ignored.\n"));
	    return 1;
	}
	if (! M->current_tag  ||
	    0 != strcmp(M->current_tag,
			M->read_tokens.tokens[0].str)) {
	    dprint(1, (debugfile,"parse_response: bad tag %s\n",
		       M->read_tokens.tokens[0].str));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}

	free(M->current_tag);
	M->current_tag = NULL;

	if (M->command_result_code) {
	    free(M->command_result_code);
	    M->command_result_code = NULL;
	}


	switch (M->read_tokens.tokens[1].imap_token) {
	    struct string *Text;
	    char *code;   /* Do not free() this -- this is just pointer
			     to read_tokens area
			  */	    
	case imap_status_keyword:
	    Text = NULL;
	    code = NULL;
	    pick_text(M,&Text,&code);

	    if (code)
		M-> command_result_code = safe_strdup(code);

	    if (code && Text && 0 == istrcmp(code,"ALERT")) {
		lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapAlert,
				      "IMAP alert: %S"),
			      Text);		
	    } else if (code && 0 == istrcmp(code,"READ-ONLY")) {
		M->folder_status &= ~IMAP_writable;
		dprint(13, (debugfile,
			    "parse_response: Folder is read-only\n"));
	    } else if (code && 0 == istrcmp(code,"READ-WRITE")) {
		M->folder_status |= IMAP_writable;
		dprint(13, (debugfile,
			    "parse_response: Folder is read-write\n"));
	    }

	    if (0 == istrcmp(M->read_tokens.tokens[1].str,"NO")) {
		M->imap_state = IMAP_command_ready_NO;
		if (Text)
		    lib_transient(CATGETS(elm_msg_cat, MeSet,MeImapError,
				      "IMAP server: %S"),
			      Text);		
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"BAD")) {
		M->imap_state = IMAP_command_ready_BAD;	    
		if (Text)
		    lib_error(CATGETS(elm_msg_cat, MeSet,MeImapError,
				      "IMAP server: %S"),
			      Text);
	    } else if (0 == istrcmp(M->read_tokens.tokens[1].str,"OK")) {
		/* Don't forget closing state ... */
		if (IMAP_closing != M->imap_state)
		    M->imap_state = IMAP_command_ready_OK;	    
	    } else {
		dprint(1, (debugfile,"parse_response: status=%s\n",
			   M->read_tokens.tokens[1].str));
		lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
				  "Bad response from IMAP server"));
		if (Text)
		    free_string(&Text);
		return 0;
	    }
	    if (Text)
		free_string(&Text);
	    break;
	default:
	    dprint(1, (debugfile,"parse_response: token[1] type =%d\n",
		       M->read_tokens.tokens[1].imap_token));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	break;

    case imap_continue:
	/* TODO: Handle AUTHENTICATE */

	if (M->literal_mode != waiting_literal_continue) {
	    dprint(1, (debugfile,
		       "parse_response: Unexpected continuation request\n"));
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			      "Bad response from IMAP server"));
	    return 0;
	}
	M->literal_mode = continue_literal;

	if (!imap_gen_writable_data(M)) {
	    panic("MBX PANIC",__FILE__,__LINE__,"parse_response",
		  "Not literal data for writing (IMAP)",0);
	}

	dprint(14, (debugfile,
		    "parse_response=2 (continue literal sending)\n")); 
	return 2;
    default:
	dprint(1, (debugfile,"parse_response: token[0] type =%d\n",
		   M->read_tokens.tokens[0].imap_token));
	lib_error(CATGETS(elm_msg_cat, MeSet,MeUnexpextedResponseImap,
			  "Bad response from IMAP server"));
	return 0;
    }

    return 1;
}

static int imap_write_socket P_((int fd,void * data));
static int imap_idle_timer P_((int fd,void * data));

static int imap_read_socket (fd,data)
     int fd; 
     void * data;
{
    struct folder_info *folder = data;
    int n, err = 0, i;
    int close_it = 0;
    int tok_ret;

    if (!data || 
	folder->folder_type != &imap_mbx ||
	folder -> p->a.imap_mbx.C.socket != fd) {
	
	panic("MBX PANIC",__FILE__,__LINE__,"imap_read_socket",
	      "Bad data argument !!!",0);
    }

    dprint(13, (debugfile,
		"imap_read_socket: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (IMAP_error == folder->p->a.imap_mbx.imap_state) {
	dprint(13, (debugfile,
		    "imap_read_socket=0 (socket need closing)\n"));
	return 0;
    }

    if (folder -> p->a.imap_mbx.wanna_literal)
	n = ReadFromSocket(fd, &(folder -> p->a.imap_mbx.read_buffer),
			   folder -> p->a.imap_mbx.literal_len -
			   folder -> p->a.imap_mbx.read_buffer.read_len);
    else
	n = ReadFromSocket(fd, &(folder -> p->a.imap_mbx.read_buffer),-1);

    if (n < 0) {
	err = errno;
	dprint(13, (debugfile,
		    "imap_read_socket: read error %s (errno %d)\n",
		    error_description(err),err));
    }

    dprint(50, (debugfile,"Buffer content="));
    for (i = 0; i < folder -> p->a.imap_mbx.read_buffer.read_len; i++) {
	dprint(50, (debugfile," %d",
		    folder -> p->a.imap_mbx.read_buffer.read_buffer[i]));
    }
    dprint(50, (debugfile,"\n (message?)="));
    for (i = 0; i < folder -> p->a.imap_mbx.read_buffer.read_len; i++) {
	dprint(50, (debugfile," %c",
		    isascii(folder -> p->a.imap_mbx.read_buffer.
			    read_buffer[i]) &&
		    isprint(folder -> p->a.imap_mbx.read_buffer.
			    read_buffer[i]) ? 
		    folder -> p->a.imap_mbx.read_buffer.read_buffer[i] :
		    '.'));
    }
    dprint(50, (debugfile,"\n"));
    
    /* Parse response ... */
    do {
	tok_ret = tokenize(& (folder -> p->a.imap_mbx));
	if (tok_ret < 0) {
	    dprint(13, (debugfile,
			"imap_read_socket: Failed to tokenize response ...\n"));
	    close_it = 1;
	} else if (tok_ret > 0) {  /* Tokenized */
	    int X;
	    if (!(X = parse_response(& (folder -> p->a.imap_mbx)))) {
		dprint(13, (debugfile,
			    "imap_read_socket: Failed to parse response ...\n"));
		close_it = 1;
	    } if (2 == X) {		
		/* Restart writing of literal ... */

		change_action(folder -> p->a.imap_mbx.C.socket,IMAP_TIMEOUT,
			      imap_read_socket,imap_write_socket,
			      imap_idle_timer,folder);
		dprint(13, (debugfile,
			    "imap_read_socket: Continue literal sending...\n"));
	    }
	    
	    free_tokens(& (folder -> p->a.imap_mbx.read_tokens));
	}	
    } while (0  < tok_ret && !close_it &&
	     folder -> p->a.imap_mbx.read_buffer.read_len > 0);
    
    if (n == 0) {
	if (folder->p->a.imap_mbx.imap_state != IMAP_closing)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectionClosedImap,
			      "Connection closed to IMAP server"));
	close_it = 1;
    }
    if (n < 0 && err != EINTR && err != EAGAIN && err != EWOULDBLOCK) {
	lib_error(CATGETS(elm_msg_cat, MeSet,MeErrorReadingImap,
			  "Error reading from IMAP server: %s"),
		  error_description(err));
	close_it = 1;
    }
    if (close_it) {
	folder->p->a.imap_mbx.imap_state = IMAP_error;
	dprint(13, (debugfile,
		    "imap_read_socket=0 (socket need closing)\n"));
	return 0;
    }
    dprint(13, (debugfile,
		"imap_read_socket=1\n"));
    return 1;
}


static int imap_write_socket (fd,data)
     int fd; 
     void * data;
{
    struct folder_info *folder = data;
    int n,err=0;

    if (!data || 
	folder->folder_type != &imap_mbx ||
	folder -> p->a.imap_mbx.C.socket != fd) {

	panic("MBX PANIC",__FILE__,__LINE__,"imap_write_socket",
	      "Bad data argument !!!",0);
    }

    dprint(13, (debugfile,
		"imap_write_socket: folder=%X (%s)\n",
		folder,folder->cur_folder));


    if (!imap_gen_writable_data(&(folder->p->a.imap_mbx))) {
	dprint(13, (debugfile,
		    "imap_write_socket: NO DATA TO WRITE!\n"));
    }

    if (IMAP_error == folder->p->a.imap_mbx.imap_state) {
	dprint(13, (debugfile,
		    "imap_write_socket=0 (socket need closing)\n"));
	return 0;
    }

    n = write(fd,folder -> p->a.imap_mbx.write_buffer,
	      folder -> p->a.imap_mbx.write_len);

    if (n > 0) {
	char * p = folder -> p->a.imap_mbx.write_buffer;
	folder -> p->a.imap_mbx.write_len -= n;
	if (folder -> p->a.imap_mbx.write_len)
	    memmove(p, p+n, folder->p->a.imap_mbx.write_len);
	dprint(13, (debugfile,
                    "imap_write_socket: Written %d bytes (%d left)\n",
		    n,folder -> p->a.imap_mbx.write_len));
    } else if (n < 0) {
	err = errno;
	dprint(13, (debugfile,
		    "imap_write_socket: write error %s (errno %d)\n",
		    error_description(err),err));
    }

    if (!folder -> p->a.imap_mbx.write_len) {
	if (!imap_gen_writable_data(& (folder -> p->a.imap_mbx))) {
	    dprint(13, (debugfile,
			"imap_write_socket=0 -- data written\n"));
	    return 0;
	}
    }

    if (n < 0 && err != EINTR && err != EAGAIN && err != EWOULDBLOCK) {
	lib_error(CATGETS(elm_msg_cat, MeSet,MeErrorWritingImap,
			  "Error writing to IMAP server: %s"),
		  error_description(err));
	folder->p->a.imap_mbx.imap_state = IMAP_error;
        dprint(13, (debugfile,
                    "imap_write_socket=0 (socket need closing)\n"));
        return 0;	
    }

    dprint(13, (debugfile,"imap_write_socket=1\n"));
    return 1;
}

#define MAX_TAG 99999

static char * tag_generator P_((void));
static char * tag_generator()
{
    /* NOTE: There is no need to have per imap connection tag 
       generating sequence
    */ 
    static char tag_letter = 'A';
    static int  tag_number = 0;

    tag_number++;
    if (tag_number > MAX_TAG) {
	tag_number = 1;
	tag_letter++;
	if ('Z' == tag_letter) {
	    tag_letter = 'A';
	    dprint(13, (debugfile,
			"tag_generator: Tag wrapping around\n"));
	}
    }

    return elm_message(FRM("%c%05d"),tag_letter,tag_number);
}

static int imap_idle_timer (fd,data)
     int fd; 
     void * data;
{
    struct folder_info *folder = data;
    int status;

    if (!data || 
	folder->folder_type != &imap_mbx ||
	folder -> p->a.imap_mbx.C.socket != fd) {

	panic("MBX PANIC",__FILE__,__LINE__,"imap_idle_timer",
	      "Bad data argument !!!",0);
    }

    dprint(13, (debugfile,
		"imap_idle_timer: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (IMAP_error == folder->p->a.imap_mbx.imap_state) {
	dprint(13, (debugfile,
		    "imap_idle_timer=0 (socket need closing)\n"));
	return 0;
    }

    if (folder->p->a.imap_mbx.current_NOOP_tag) {
	dprint(13, (debugfile,
		    "imap_idle_timer=1 (previous NOOP command is not finished)\n"));
	return 1;
    }


    if (folder->p->a.imap_mbx.token_index >= 0 || 
	folder->p->a.imap_mbx.imap_state != IMAP_idle) {
	dprint(13, (debugfile,
		    "imap_idle_timer=1 (NOT idle)\n"));
	return 1;	
    }
    folder->p->a.imap_mbx.current_NOOP_tag = tag_generator();
    folder->p->a.imap_mbx.token_index      = -2;    /* Produce NOOP command */
    folder->p->a.imap_mbx.imap_state = IMAP_command;   /* For imap_gen_writable_data */

    if (!imap_gen_writable_data(&(folder->p->a.imap_mbx))) {
	panic("MBX PANIC",__FILE__,__LINE__,"imap_idle_timer",
	      "No NOOP command for writing (IMAP)",0);
    }

    change_action(folder -> p->a.imap_mbx.C.socket,IMAP_TIMEOUT,
		  imap_read_socket,imap_write_socket,
		  imap_idle_timer,folder);

    /* Mark it again IDLE ... */
    folder->p->a.imap_mbx.imap_state = IMAP_idle;

    return 1;
}

static void imap_wait P_((struct folder_info *folder));
static void imap_wait(folder)
     struct folder_info *folder;
{
    while (IMAP_command_ready_OK != folder->p->a.imap_mbx.imap_state &&
	   IMAP_command_ready_NO != folder->p->a.imap_mbx.imap_state &&
	   IMAP_command_ready_BAD != folder->p->a.imap_mbx.imap_state &&
	   IMAP_closing != folder->p->a.imap_mbx.imap_state &&
	   IMAP_error != folder->p->a.imap_mbx.imap_state &&
           -1         != folder -> p->a.imap_mbx.C.socket) {


	wait_for_action(imap_read_socket);
	/* TODO:    Should user to have some way to interrupt progress ? */
    }
}


static void imap_idle_wait P_((struct folder_info *folder));
static void imap_idle_wait(folder)
     struct folder_info *folder;
{
    struct IMAP_MBX * M = & (folder->p->a.imap_mbx);


    if (M->current_NOOP_tag &&
	M->imap_state != IMAP_error) {
	dprint(12, (debugfile,
		    "imap_idle_wait: Waiting for idle NOOP complete...\n"));
	while(M->current_NOOP_tag && 
	      M->imap_state != IMAP_error) {
	    wait_for_action(imap_read_socket);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}

	dprint(12, (debugfile,
		    "imap_idle_wait: idle NOOP wait completed\n"));
    }
    if (IMAP_idle != M->imap_state) {
	dprint(12, (debugfile,
		    "imap_idle_wait: Waiting for command complete...\n"));

	imap_wait(folder);
	dprint(12, (debugfile,
		    "imap_idle_wait: command wait completed\n"));
    }

}


static int imap_command_ok P_((struct folder_info *folder,
			       imap_states *cmd_result,
			       char        **extra_code));
static int imap_command_ok(folder,cmd_result,extra_code) 
     struct folder_info *folder;
     imap_states *cmd_result;
     char **extra_code;
{
    int status = 0;
    dprint(12, (debugfile,
		"imap_command_ok: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (extra_code) 
	*extra_code = NULL;

    if (IMAP_idle == folder->p->a.imap_mbx.imap_state) {
	dprint(12, (debugfile,
		    "imap_command_ok: WRONG STATE\n"));
	status = 0;
	*cmd_result = IMAP_error;
	goto clean;
    }
    imap_wait(folder);

    status = IMAP_command_ready_OK == folder->p->a.imap_mbx.imap_state;
    *cmd_result = folder->p->a.imap_mbx.imap_state;

    if (IMAP_command_ready_OK == folder->p->a.imap_mbx.imap_state ||
	IMAP_command_ready_NO == folder->p->a.imap_mbx.imap_state ||
	IMAP_command_ready_BAD == folder->p->a.imap_mbx.imap_state) {
	
	if (extra_code) {
	    *extra_code = folder->p->a.imap_mbx.command_result_code;
	    folder->p->a.imap_mbx.command_result_code = NULL;	    
	}
    }

    if (IMAP_error == folder->p->a.imap_mbx.imap_state) {
	if (-1 != folder -> p->a.imap_mbx.C.socket ) {
	    dprint(12, (debugfile,
			"imap_command_ok: Closing socket\n"));
	    
	    clear_action(folder -> p->a.imap_mbx.C.socket);
	    close(folder -> p->a.imap_mbx.C.socket);
	    folder->p->a.imap_mbx.C.socket = -1;
	}	
	status = 0;
	goto clean;
    }

 clean:
    dprint(12, (debugfile,
		"imap_command_ok=%d\n",
		status));
    return status;
}

static void imap_command_push_atom P_((struct folder_info *folder,
				       const char *st));

static int start_imap_command P_((struct folder_info *folder,
				  char * verb));
static int start_imap_command(folder,verb)
     struct folder_info *folder;
     char * verb;
{
    int status = 0;

    dprint(12, (debugfile,
		"start_imap_command: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (IMAP_error == folder->p->a.imap_mbx.imap_state) {
    failure:
	if (-1 != folder -> p->a.imap_mbx.C.socket ) {
	    dprint(12, (debugfile,
			"start_imap_command: Closing socket\n"));
	    
	    clear_action(folder -> p->a.imap_mbx.C.socket);
	    close(folder -> p->a.imap_mbx.C.socket);
	    folder->p->a.imap_mbx.C.socket = -1;
	}	
	status = 0;
	goto clean;
    }

    if (IMAP_idle != folder->p->a.imap_mbx.imap_state) {
	dprint(12, (debugfile,
		    "start_imap_command: wrong state  ... waiting...\n"));
	
	imap_wait(folder);
	dprint(12, (debugfile,
		    "start_imap_command: ... wait ended\n"));
	if (POP_error == folder->p->a.imap_mbx.imap_state) 
	    goto failure;	
    }
    
    if (folder->p->a.imap_mbx.current_tag) {
	dprint(12, (debugfile,
		    "start_imap_command: Tag %s still active\n",
		    folder->p->a.imap_mbx.current_tag));
	status = 0;
	goto clean;	
    }

    folder->p->a.imap_mbx.imap_state  = IMAP_command;
    free_tokens(& (folder -> p->a.imap_mbx.write_tokens));
    folder->p->a.imap_mbx.token_index = -1;
    folder->p->a.imap_mbx.literal_mode = no_literal;

    if (-1 == folder -> p->a.imap_mbx.C.socket) {
	dprint(12, (debugfile,
		    "start_imap_command: socket not open\n"));
	status = 0;
	goto clean;
    }

    folder -> p->a.imap_mbx.current_tag = tag_generator();
    imap_command_push_atom(folder,verb);
    status          = 1; 

    dprint(15, (debugfile,
		"start_imap_command: Command %s ...\n",verb));

 clean:
    dprint(12, (debugfile,
		"start_imap_command=%d\n",
		status));
    return status;    
}

static void imap_command_push_atom(folder,st)
     struct folder_info *folder;
     CONST char *st;
{
    struct imap_token T;

    T.imap_token    = imap_atom;
    T.str           = safe_strdup(st);
    T.len_or_value  = strlen(st);
    add_token(&(folder -> p->a.imap_mbx.write_tokens),T);
}

static void imap_command_push_string P_((struct folder_info *folder,
					 const char *st));
static void imap_command_push_string(folder,st)
     struct folder_info *folder;
     CONST char *st;
{
    struct imap_token T;

    T.imap_token    = imap_string;
    T.str           = safe_strdup(st);
    T.len_or_value  = strlen(st);
    add_token(&(folder -> p->a.imap_mbx.write_tokens),T);
}

static void imap_command_push_astring P_((struct folder_info *folder,
					 const char *st));
static void imap_command_push_astring(folder,st)
     struct folder_info *folder;
     CONST char *st;
{
    if ('\0' == st[0] ||
	strspn(st,"ABCDEFGHIJKLMNOPQRSTUVXYZabcdefghijklmnopqrstuvxyz")
	!= strlen(st))
	imap_command_push_string(folder,st);
    else 
	imap_command_push_atom(folder,st);
}

static void imap_command_push_range P_((struct folder_info *folder,
					int A, int B));
static void imap_command_push_range(folder,A,B)
     struct folder_info *folder;
     int A,B;
{
    char * st = elm_message(FRM("%d:%d"),A,B);
    imap_command_push_atom(folder,st);
    free(st);
}

static void imap_command_push_number P_((struct folder_info *folder,
					int A));
static void imap_command_push_number(folder,A)
     struct folder_info *folder;
     int A;
{
    char * st = elm_message(FRM("%d"),A);
    imap_command_push_atom(folder,st);
    free(st);
}


static void imap_command_push_list P_((struct folder_info *folder,
				       const char **l));
static void imap_command_push_list(folder,l)
     struct folder_info *folder;
     CONST char **l;
{
    int i;
    struct imap_token T;

    T.imap_token    = imap_list_begin;
    T.str           = NULL;
    T.len_or_value  = 0;
    add_token(&(folder -> p->a.imap_mbx.write_tokens),T);

    for (i = 0; l[i]; l++)
	imap_command_push_atom(folder,l[i]);

    T.imap_token    = imap_list_end;
    T.str           = NULL;
    T.len_or_value  = 0;
    add_token(&(folder -> p->a.imap_mbx.write_tokens),T);
}



static void end_imap_command P_((struct folder_info *folder));
static void end_imap_command(folder)
     struct folder_info *folder;
{
    dprint(12, (debugfile,
		"end_imap_command: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (folder->p->a.imap_mbx.imap_state != IMAP_command)
	panic("MBX PANIC",__FILE__,__LINE__,
	      "end_imap_command","wrong state",0);

    if (!imap_gen_writable_data( &(folder -> p->a.imap_mbx))) {
	panic("MBX PANIC",__FILE__,__LINE__,"end_imap_command",
	      "No command for writing",0);
    }
    change_action(folder -> p->a.imap_mbx.C.socket,IMAP_TIMEOUT,
		  imap_read_socket,imap_write_socket,
		  imap_idle_timer,folder);
}

/* ----------------------------------------------------------------------- */

static void mbx_close_imap P_((struct folder_info *folder,
			      enum close_mode mode));
static void mbx_close_imap(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{
    dprint(11, (debugfile,
		"mbx_close_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (start_imap_command(folder,"LOGOUT")) {
	imap_states res;
	end_imap_command(folder);
	imap_command_ok(folder,&res,NULL);
	imap_clear_command(folder);
    }

    if (-1 != folder -> p->a.imap_mbx.C.socket ) {
	dprint(12, (debugfile,
		    "mbx_close_imap: Closing socket\n"));
	
	clear_action(folder -> p->a.imap_mbx.C.socket);
	close(folder -> p->a.imap_mbx.C.socket);
	folder->p->a.imap_mbx.C.socket = -1;
    }	

    switch (mode) {
    case CLOSE_NORMAL:
	
	/* falltru */
    case CLOSE_LEAVE_LOCKED:
	if (!folder -> p->fh_temp) {
	    dprint(1, (debugfile,
		       "mbx_close_imap: tempfile (%s) not open -- will not unlink\n",
		       folder -> cur_tempfolder));
	} else if (unlink(folder -> cur_tempfolder) != 0) {
	    if (errno != ENOENT) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantUnlinkTemp,
				  "Sorry, can't unlink the temp file %s [%s]!\n\r"),
			  folder -> cur_tempfolder, error_description(errno));
	    }
	} else {
	    dprint(1, (debugfile, 
		       "close_folder: Unlinking tempfile (%s).\n",
		       folder -> cur_tempfolder));
	}      
    }
    
    return;
}

static int  mbx_lock_imap  P_((int direction,struct folder_info *folder));
static int  mbx_lock_imap(direction,folder)
     int direction;
     struct folder_info *folder;
{
    int status;
    dprint(11, (debugfile,
		"mbx_lock_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    status = 1;

    dprint(11, (debugfile,
		"mbx_lock_imap=%d\n",
		status));
    return status;
}

static void mbx_init_imap P_((struct folder_info *folder));
static void mbx_init_imap(folder) 
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_init_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    folder -> cur_tempfolder[0] = '\0';
    /* remote_mbox.c sets temporary folder name ... */

    folder -> p = safe_malloc(sizeof (struct private_data));
    /* defined in hdrs/defs.h */
    bzero((void *)folder -> p,sizeof( struct private_data));   
    
    folder ->p->fh_temp     = NULL;
    folder ->p->fh_folder   = NULL;

    zero_remote_account(&(folder -> p->a.imap_mbx.C));

    folder -> p->a.imap_mbx.folder       = NULL; /* Set by remote_mbox.c */
    
    folder -> p->a.imap_mbx.folder_status   = 0;
    folder -> p->a.imap_mbx.capability_bits = 0;
    folder -> p->a.imap_mbx.flag_bits       = 0;
    folder -> p->a.imap_mbx.num_recent      = 0;
    folder -> p->a.imap_mbx.last_uid_number = 0;

    folder -> p->a.imap_mbx.references      = NULL;
    folder -> p->a.imap_mbx.reference_count = 0;

    zero_Read_Buffer(&(folder -> p->a.imap_mbx.read_buffer));
    folder -> p->a.imap_mbx.wanna_literal   = 0;
    folder -> p->a.imap_mbx.literal_len     = 0;

    folder -> p->a.imap_mbx.read_tokens.tokens          = NULL;
    folder -> p->a.imap_mbx.read_tokens.token_count     = 0;
    folder -> p->a.imap_mbx.tokenizer_state = state_tag;

    folder -> p->a.imap_mbx.current_tag          = NULL;
    folder -> p->a.imap_mbx.imap_state          = IMAP_not_logged;
    folder -> p->a.imap_mbx.command_result_code = NULL;

    folder -> p->a.imap_mbx.write_tokens.tokens      = NULL;
    folder -> p->a.imap_mbx.write_tokens.token_count = 0;
    folder -> p->a.imap_mbx.token_index  = -1;

    folder -> p->a.imap_mbx.write_buffer = NULL;
    folder -> p->a.imap_mbx.write_len    = 0;
    folder -> p->a.imap_mbx.literal_mode = no_literal;
    
    folder -> p->a.imap_mbx.current_NOOP_tag = NULL;
}


static void imap_clear_buffers P_((struct folder_info *folder));
static void imap_clear_buffers(folder)
     struct folder_info *folder;
{
    dprint(12, (debugfile,
		"imap_clear_buffers: folder=%X (%s)\n",
		folder,folder->cur_folder));    

    folder -> p->a.imap_mbx.folder_status   = 0;
    folder -> p->a.imap_mbx.capability_bits = 0;
    folder -> p->a.imap_mbx.flag_bits       = 0;
    folder -> p->a.imap_mbx.last_uid_number = 0;

    folder -> p->a.imap_mbx.num_recent      = 0;

    if (folder -> p->a.imap_mbx.references) {
	int i;
    
	for (i = 0; i < folder -> p->a.imap_mbx.reference_count; i++) {
	    free_reference(&(folder -> p->a.imap_mbx.references[i]));
	}
	free(folder -> p->a.imap_mbx.references);
	folder -> p->a.imap_mbx.references = NULL;
	folder -> p->a.imap_mbx.reference_count = 0;
    }

    free_Read_Buffer(&(folder -> p->a.imap_mbx.read_buffer));
    folder -> p->a.imap_mbx.wanna_literal = 0;
    folder -> p->a.imap_mbx.literal_len   = 0;
    
    free_tokens(& (folder -> p->a.imap_mbx.read_tokens));

    folder -> p->a.imap_mbx.tokenizer_state = state_tag;

    if (folder -> p->a.imap_mbx.current_tag) {
	free(folder -> p->a.imap_mbx.current_tag);
	folder -> p->a.imap_mbx.current_tag = NULL;
    }
    folder->p->a.imap_mbx.imap_state = IMAP_error;
    if (folder -> p->a.imap_mbx.command_result_code) {
	free(folder -> p->a.imap_mbx.command_result_code);
	folder -> p->a.imap_mbx.command_result_code = NULL;
    }

    free_tokens(& (folder -> p->a.imap_mbx.write_tokens));
    folder -> p->a.imap_mbx.token_index = -1;
    
    if (folder -> p->a.imap_mbx.write_buffer) {
	free(folder -> p->a.imap_mbx.write_buffer);
	folder -> p->a.imap_mbx.write_buffer = NULL;
	folder -> p->a.imap_mbx.write_len = 0;
    }
    folder -> p->a.imap_mbx.literal_mode = no_literal;

    if (folder->p->a.imap_mbx.current_NOOP_tag) {
	free(folder->p->a.imap_mbx.current_NOOP_tag);
	folder->p->a.imap_mbx.current_NOOP_tag = NULL;
    }
}

static int imap_open_connection P_((struct folder_info *folder));
static int imap_open_connection(folder)
     struct folder_info *folder;
{
    imap_states res;
    int status = 0;
    char * data1 = NULL;

    dprint(12, (debugfile,
		"imap_open_connection: folder=%X (%s)\n",
		folder,folder->cur_folder));

    /* Clear buffer from old data ... */
    imap_clear_buffers(folder);

    if (-1 == folder->p->a.imap_mbx.C.socket) {
	struct hostent *he = gethostbyname(folder->p->a.imap_mbx.C.host);
	PORTS IMAP_ports[] = { PORT_imap4, PORT_end }, got;

	if (!he) {
	    dprint(12, (debugfile,
			"imap_open_connection: %s: h_errno=%d\n",
			folder->p->a.imap_mbx.C.host,h_errno));
	    switch(h_errno) {
	    case HOST_NOT_FOUND:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoIMAPHost,
				  "No IMAP server host: %s"),
			  folder->p->a.imap_mbx.C.host);
		status = 0;
		break;
	    case NO_ADDRESS:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoAddress,
				  "Name %s have not IP-address"),
			  folder->p->a.imap_mbx.C.host);
		status = 0;
	    case TRY_AGAIN:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeTemporary,
				  "Address for %s is not yet found..."),
			  folder->p->a.imap_mbx.C.host);
		status = 0;
		break;			  
	    default:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNameError,
				  "Failed to get address for %s"),
			  folder->p->a.imap_mbx.C.host);
		status = 0;
		break;
	    }
	    goto clean;
	}
	
	/* canonify folder name .... */
	free(folder->cur_folder);
	folder->cur_folder = elm_message(FRM("%s@%s%s%s"),
					 folder->p->a.imap_mbx.C.username,
					 he->h_name,
					 folder->p->a.imap_mbx.folder[0] == '/'
					 ? "" : ":",
					 folder->p->a.imap_mbx.folder);
	free(folder->p->a.imap_mbx.C.host);
	folder->p->a.imap_mbx.C.host = safe_strdup(he->h_name);

	if (!connect_remote_account(&(folder->p->a.imap_mbx.C),
				    IMAP_ports,&got,he)) {
	    status = 0;
	    goto clean;
	}
    }

    /* We want server's initial response to connection ...
       .. maybe untagged BYE?
     */
    folder -> p->a.imap_mbx.imap_state = IMAP_idle;
    change_action(folder -> p->a.imap_mbx.C.socket,IMAP_TIMEOUT,
                  imap_read_socket,no_action_routine,imap_idle_timer,
                  folder);    
    wait_for_action(imap_read_socket);
    if (-1 ==  folder -> p->a.imap_mbx.C.socket ||
	IMAP_error   == folder -> p->a.imap_mbx.imap_state ||
	IMAP_closing == folder -> p->a.imap_mbx.imap_state) {
	dprint(12, (debugfile,
		    "imap_open_connection: Server does not accept connection\n"));
	goto clean;
    }
    
    if (!start_imap_command(folder,"LOGIN")) 
	goto clean;

    imap_command_push_astring(folder,folder -> p->a.imap_mbx.C.username);
    lib_transient(CATGETS(elm_msg_cat, MeSet,MeIMAPLogin,
		      "IMAP login to %s as %s ..."),
		  folder -> p->a.imap_mbx.C.host,
		  folder -> p->a.imap_mbx.C.username);
    data1 = lib_prompt(1,CATGETS(elm_msg_cat, MeSet,MeLoginPasswd,
				 "Password for %s@%s:"),
		       folder -> p->a.imap_mbx.C.username,
		       folder -> p->a.imap_mbx.C.host);
    if (!data1) {
	status = 0;
	goto clean;	    
    }
    imap_command_push_astring(folder,data1);
    end_imap_command(folder);

    if (!imap_command_ok(folder,&res,NULL))
	goto clean;
    status = 1;
    
 clean:
    imap_clear_command(folder);

    if (data1) {
	free(data1);
	data1 = NULL;
    }

    dprint(12, (debugfile,
		"imap_open_connection=%d\n",
		status));
    return status;
}


static int mbx_sessionlock_imap P_((struct folder_info *folder,
				    enum sessionlock_mode mode));
static int mbx_sessionlock_imap(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    imap_states res;
    int status = 0;

    dprint(11, (debugfile,
		"mbx_sessionlock_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (-1 == folder->p->a.imap_mbx.C.socket ||
	folder -> p->a.imap_mbx.imap_state  == IMAP_not_logged) {
	if (!imap_open_connection(folder)) {
	    status = 0;
	    goto clean;
	}
	folder->p->a.imap_mbx.folder_status = 0;
    }

    if (0 == (folder->p->a.imap_mbx.folder_status & IMAP_folder_open)) {

	switch(mode) {
	case SESSIONLOCK_NONE:
	    if (IMAP_use_examine) {
		if (!start_imap_command(folder,"EXAMINE")) 
		    goto clean;
		break;	
	    }
	    /* FALLTHRU */
	default:
	    if (!start_imap_command(folder,"SELECT")) 
	    goto clean;
	    break;
	}
	
	imap_command_push_astring(folder,folder -> p->a.imap_mbx.folder);
	end_imap_command(folder);
	
	if (!imap_command_ok(folder,&res,NULL))
	    goto clean;
	folder->p->a.imap_mbx.folder_status |= IMAP_folder_open;
    }

    switch(mode) {
	int need_reopen;
    case SESSIONLOCK_NONE:
	status = 1;
	goto clean;
    case SESSIONLOCK_TRUNCATE:
	need_reopen = 1;
#ifdef FTRUNCATE  
	if (folder->p->fh_temp) {
	    rewind(folder->p->fh_temp);
	    if (0 == ftruncate(fileno(folder->p->fh_temp),0)) {
		need_reopen = 0;
		dprint(10, (debugfile,
			    "truncate_tempfolder: tempfile %s truncated (not recreated)\n",
			    folder->cur_tempfolder));
	    }
	    else 
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmSorryCantTruncateTemp,
				  "Sorry, can't truncate the temp file %s [%s]!"),
			  folder -> cur_tempfolder, error_description(errno));
	    /* IF FAIL REOPEN IT INSTEAD */
	}
#endif
	if (!need_reopen)
	    break;  /* DONE */
	if (folder->p->fh_temp) {
	    fclose (folder->p->fh_temp);
	    folder->p->fh_temp = NULL;
	}
	
	if (0 != unlink(folder -> cur_tempfolder)) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantTruncateTemp,
			      "Sorry, can't truncate the temp file %s [%s]!"),
		      folder -> cur_tempfolder, error_description(err));    
	    status = 0;
	    goto clean;
	}
	goto create_it;
    case   SESSIONLOCK_REOPEN:
	if (folder -> p->fh_temp) {
	    int temp_handle;
	    fclose(folder -> p->fh_temp);
	    folder -> p->fh_temp = NULL;
	    
	    if ( (temp_handle = open(folder-> cur_tempfolder, 
				     O_RDWR,
				     0600)) == -1 ||
		 NULL == (folder -> p->fh_temp = fdopen(temp_handle,"w+"))) {
		int err = errno;
		
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantReopenTemp,
				  "Sorry, can't reopen the temp file %s [%s]!"),
			  folder -> cur_tempfolder, error_description(err));  
		
		if (-1 != temp_handle)
		    close(temp_handle);
		
		status = 0;
		goto clean;
	    }
	    
	    break;
	}
	/* FALLTHRU */
    case SESSIONLOCK_NORMAL:
    case SESSIONLOCK_CHECK:
    create_it:
        if (!folder -> p->fh_temp) {
	    int temp_handle;
	    
	    /* If we are changing files and we are changing to a spool file,
	     * make sure there isn't a temp file for it, because if
	     * there is, someone else is using ELM to read the new file,
	     * and we don't want to be reading it at the same time.
	     */
	    
	    if ( (temp_handle = open(folder-> cur_tempfolder, 
				     O_RDWR|O_CREAT|O_EXCL,
				     0600)) == -1) {
		int err = errno;
		
		if (err == EEXIST) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning1,
				      "You seem to have ELM already reading this mail!"));
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning2,
				      "You may not have two copies of ELM running simultaneously.  -- Exiting --"));
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning3,
				      "If this is in error, then you'll need to remove the following file:"));
		    lib_error(FRM("%s"),folder -> cur_tempfolder);
		    
		    status = 0;
		    goto clean;	    
		}
	    }
	    if (-1 == temp_handle || 
		NULL == (folder -> p->fh_temp = fdopen(temp_handle,"w+"))) {
		int err = errno;
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedCreateTmpFolder,
				  "Failed to create %.50s: %.60s"),
			  folder -> cur_tempfolder,
			  error_description(err));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIGiveUp,
				  "Ahhhh... I give up."));
		
		if (-1 != temp_handle)
		    close(temp_handle);
		
		status = 0;
		goto clean;	    
	    }
	    
	    dprint(1, (debugfile, 
		       "Creating tempfile (%s).\n",
		       folder -> cur_tempfolder));
	    
	    elm_chown(folder->cur_tempfolder, userid, groupid);
	    chmod(folder -> cur_tempfolder, 0700);	
	    /* shut off file for other people! */	    
	}
	break;
    }
    status = 1;

 clean:
    imap_clear_command(folder);

    dprint(11, (debugfile,
		"mbx_sessionlock_imap=%d\n",
		status));
    return status;
}

static int mbx_unlock_imap P_((int interrupt, struct folder_info *folder));
static int mbx_unlock_imap(interrupt,folder) 
     int interrupt;
     struct folder_info *folder;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_unlock_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    status = 1;

    dprint(11, (debugfile,
		"mbx_unlock_imap=%d\n",
		status));
    return status;
}

static void mbx_flush_imap P_((struct folder_info *folder));
static void mbx_flush_imap(folder) 
     struct folder_info *folder;
{
    imap_states res;
    dprint(11, (debugfile,
		"mbx_flush_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    mbx_flush_temp(folder);
    
#if 0         
    /* It does not make sense on here to force IMAP server to flush
     *  mailbox -- that flushing is needed only in
     *  mbx_end_keep_imap()
     */
    if (!start_imap_command(folder,"CHECK")) {
	dprint(11, (debugfile,
		    "mbx_flush_imap: Setting IMAP_flusherr\n"));
	folder->p->a.imap_mbx.folder_status |= IMAP_flusherr;
	goto clean;
    }
    end_imap_command(folder);
    
    if (!imap_command_ok(folder,&res,NULL)) {
	folder->p->a.imap_mbx.folder_status |= IMAP_flusherr;
	dprint(11, (debugfile,
		    "mbx_flush_imap: Setting IMAP_flusherr\n"));
    }
    
 clean: 
#endif

    imap_clear_command(folder);
    
}

static int mbx_ferror_imap P_((struct folder_info *folder, int clean));
static int mbx_ferror_imap(folder,clean) 
     struct folder_info *folder; 
     int clean; 
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_ferror_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (0 != (folder->p->a.imap_mbx.folder_status & IMAP_flusherr)) {
	dprint(11, (debugfile,
		    "mbx_ferror_imap: folder_status ~ IMAP_flusherr is set\n"));
	status = 1;
    }
    if (clean)
	folder->p->a.imap_mbx.folder_status &= ~IMAP_flusherr;

    if (folder->p->fh_temp) {
	if (ferror(folder->p->fh_temp))
	    status = 1;
	if (clean)
	    clearerr(folder->p->fh_temp);
    }

    dprint(11, (debugfile,
		"mbx_ferror_imap=%d\n",status));
    return status;
}


static int mbx_prepare_read_imap P_((struct folder_info *folder,
				    enum prepare_mode mode,
				    READ_STATE read_state_ptr));
static int mbx_prepare_read_imap(folder,mode,read_state_ptr)
     struct folder_info *folder;
     enum prepare_mode mode;
     READ_STATE read_state_ptr;
{
    int status = 0, i,start;
    int add_new_only = PREPARE_NEW_ONLY == mode ||
                       PREPARE_NEW_ONLY_NOLOCK == mode;

    dprint(11, (debugfile,
		"mbx_prepare_read_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                    : read_state_ptr=%X\n",
		read_state_ptr));

    /* Wait that IDLE timer command is completetd so that there
     * is no EXPUNGE coming...
     */
    imap_idle_wait(folder);

    if (add_new_only) {
	/* start from unread messages */

	read_state_ptr->a.imap_mbx.current_message.uid_number =
	    folder -> p->a.imap_mbx.last_uid_number+1;

	start = 1;
	while (start < folder -> p->a.imap_mbx.reference_count &&
	       folder -> p->a.imap_mbx.references[start].uid_number != -1 &&
	       folder -> p->a.imap_mbx.references[start].uid_number <
	       read_state_ptr->a.imap_mbx.current_message.uid_number)
	    start++;

	if (!folder->p->fh_temp) {
	    dprint(11, (debugfile,
			"mbx_prepare_read_pop -- NO temp file %s\n",
			folder->cur_tempfolder));
	} else if (fseek(folder -> p->fh_temp, 
			 folder -> mailfile_size, SEEK_SET) == -1) {
	    int err = errno;
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntSeekTempEnd,
			      "\nCouldn't fseek to end of temp mbox.\n"));
	    lib_error(FRM("** %s. **\n"), error_description(err));
	    dprint(1, (debugfile, 
		       "Error: Couldn't fseek to end of reopened temp mbox.  errno %s (%s)\n",
		       error_description(err), "mbx_prepare_read_spool"));
	    status = 0;
	    goto clean;	    
	}
	read_state_ptr->fbytes = folder-> mailfile_size;  /* start correctly */
    } else {
	/* Start from beginning */
	read_state_ptr->a.imap_mbx.current_message.uid_number = 1;
	start = 1;
    }

    dprint(11, (debugfile,
		"                    : UID start=%d  start=%d refcount=%d\n",
		read_state_ptr->a.imap_mbx.current_message.uid_number,start,
		folder -> p->a.imap_mbx.reference_count));
	
    if (start <= folder -> p->a.imap_mbx.reference_count-1) {	    

	CONST char * L[] = { "RFC822.SIZE", "UID", NULL };
	imap_states res;

	if (!start_imap_command(folder,"FETCH")) {
	    status = 0;
	    goto clean;
	}
    
	imap_command_push_range(folder,
				start,
				folder -> p->a.imap_mbx.reference_count-1);


	imap_command_push_list(folder,L);
	end_imap_command(folder);

	if (!imap_command_ok(folder,&res,NULL))
	    goto clean;
    }

    /* Estimate new size for percent display */
    for ( i = start; i < folder -> p->a.imap_mbx.reference_count; i++) {
	folder -> mailfile_size += 
	    folder-> p->a.imap_mbx.references[i].rfc822_size;
	folder -> p->a.imap_mbx.last_uid_number = 
	    folder-> p->a.imap_mbx.references[i].uid_number;
    }
    read_state_ptr->a.imap_mbx.data_idx = 0;
    status = 1;

    dprint(11, (debugfile,
		"                    new last UID %d\n",
		folder -> p->a.imap_mbx.last_uid_number));
 clean:
    imap_clear_command(folder);
    dprint(11, (debugfile,
		"mbx_prepare_read_imap=%d\n",status));
    return status;
}

static int mbx_end_read_imap P_((struct folder_info *folder,
				READ_STATE read_state_ptr,
				int silent));
static int mbx_end_read_imap(folder,read_state_ptr,silent)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     int silent;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_end_read_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                : read_state_ptr=%X\n",
		read_state_ptr));

    if (folder -> p->a.imap_mbx.reference_count > 1 &&
	folder-> p->a.imap_mbx.references[folder -> 
					 p->a.imap_mbx.reference_count-1].
	uid_number > read_state_ptr->a.imap_mbx.current_message.uid_number) {
	if (!silent) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeMessagesNotReadedImap,
			      "Messages not readed; current UID=%d, last UID=%d"),
		      read_state_ptr->a.imap_mbx.current_message.uid_number,
		      folder-> p->a.imap_mbx.references[folder -> 
						       p->a.imap_mbx.
						       reference_count-1].
		      uid_number);
	}
	status = 0;
	goto clean;	
    }
    
    if (!folder->p->fh_temp) {
	dprint(11, (debugfile,
		    "mbx_end_read_imap=1 -- NO temp file %s\n",
		    folder->cur_tempfolder));
	return 1;
    }
    
    if ((ferror(folder->p->fh_temp)) || 
	(fflush(folder->p->fh_temp) == EOF)) {
	if (!silent) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFlushOnTempFailed,
			      "Flush on temp file %s failed: %s"), 
		      folder -> cur_tempfolder,
		      error_description(err));
	}
	dprint(1, (debugfile, "Can't flush on temp file %s!!\n",
		   folder -> cur_tempfolder));
	status = 0;
	goto clean;
    }
    
    /* Use tempfile size as folder size */
    folder-> mailfile_size = file_bytes(folder-> cur_tempfolder);
    
    rewind(folder->p->fh_temp);
    status = 1;
    
 clean:
    dprint(11, (debugfile,
		"mbx_end_read_imap=%d\n",status));
    return status;
}



static void info_zero_imap_mbx P_((struct mbx_hdr_info *info));
static void info_zero_imap_mbx(info)
     struct mbx_hdr_info *info;
{
    info->a.imap_mbx.last_mbx_index = -1;
    info->a.imap_mbx.uid_number = 0;
}

static void info_free_imap_mbx P_((struct mbx_hdr_info *info));
static void info_free_imap_mbx(info)
     struct mbx_hdr_info *info;
{
    info->a.imap_mbx.last_mbx_index = -1;
    info->a.imap_mbx.uid_number = 0;
}

static struct info_type imap_mbx_info = { info_zero_imap_mbx, 
					  info_free_imap_mbx };


static int mbx_copy_envelope_imap P_((struct folder_info *folder,
				      READ_STATE read_state_ptr,
				      struct header_rec *entry));
static int mbx_copy_envelope_imap(folder,read_state_ptr,entry)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     struct header_rec *entry;
{
    int status = 0;
    char * data = NULL;
    char * tmp;
    imap_states res;
    struct imap_reference  *ref = NULL;

    int l,j;

    CONST char * L[] = { "FLAGS", "INTERNALDATE", "BODY.PEEK[HEADER]", NULL };

    dprint(11, (debugfile,
		"mbx_copy_envelope_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                     : read_state_ptr=%X\n",
		read_state_ptr));

    /* Wait that IDLE timer command is completetd so that there
     * is no EXPUNGE coming...
     */
    imap_idle_wait(folder);

    do {
	ref = find_reference(& (folder -> p->a.imap_mbx),
			     & (read_state_ptr->a.imap_mbx.current_message));
	if (!ref) {
	    CONST int count = folder -> p->a.imap_mbx.reference_count;
	    if (count  <= 1 ||
		folder-> p->a.imap_mbx.references[count-1].uid_number < 
		read_state_ptr->a.imap_mbx.current_message.uid_number) {
		dprint(11, (debugfile,
			    "                     : End of messages\n"));
		status = 0;
		goto clean;
	    }
	    read_state_ptr->a.imap_mbx.current_message.uid_number++;
	}
    } while (!ref);

    read_state_ptr -> linecounter   = 0;      /* Linecounter of current
					         message */
    entry->offset = read_state_ptr -> fbytes; /* Offset of current message */
    entry->status = VISIBLE;

    if (!start_imap_command(folder,"FETCH")) 
	goto clean;

    imap_command_push_number(folder,
			     read_state_ptr->
			     a.imap_mbx.current_message.last_mbx_index);
    imap_command_push_list(folder,L);
    end_imap_command(folder);
    
    if (!imap_command_ok(folder,&res,NULL))
	goto clean;

    /* Check that message is not deleted meanwhile --
       Should not happen: EXPUNGE response is not allowed during of FETCH
     */
    ref = find_reference(& (folder -> p->a.imap_mbx),
			 & (read_state_ptr->a.imap_mbx.current_message));
    if (!ref) {
	dprint(1, (debugfile,
		   "mbx_copy_envelope_imap: EXPUNGE middle of FETCH ???\n"));
	goto clean;
    }
    if (!ref->header) {
	dprint(1, (debugfile,
		   "mbx_copy_envelope_imap: BODY[HEADER] not got\n"));
	goto clean;
    }
    if (!ref->internaldate) {
	dprint(1, (debugfile,
		   "mbx_copy_envelope_imap: INTERNADATE not got\n"));
	goto clean;
    }

    if (0 == (ref->imap_flags & IMAP_Seen))
	entry->status |= UNREAD;

    dprint(14, (debugfile,
		"mbx_copy_envelope_imap: Flags"));
    for (j = 0; j < sizeof IMAP_flags / sizeof (struct imap_flag); j++)
	if (0 != (ref->imap_flags & IMAP_flags[j].mask_imap)) {
	   entry->status |= IMAP_flags[j].mask_elm; 
	   dprint(14, (debugfile," %s", IMAP_flags[j].flag));
	}
    dprint(14, (debugfile,"\n"));
	   
    /* Very lazy method to get something to put "From " -separator */
    if (0 == memcmp("Return-Path: ",ref->header,13)) {
	int i;
	for (i=13; i < ref->header_len; i++) {
	    if ('\r' == ref->header[i] ||
		'\n' == ref->header[i])
		break;
	}
	data = safe_malloc(i-12);
	memcpy(data,ref->header+13,i-13);
	data[i-13]='\0';
    } else
	data = safe_strdup("nobody@localhost");

    /* We do not generate weekday -- real_from() does not care */
    tmp = elm_message(FRM("From %s XXX %3.3s %2.2s %8.8s %5.5s %4.4s\n"),data,
		      ref->internaldate+3,
		      ref->internaldate,
		      ref->internaldate+12,
		      ref->internaldate+21,
		      ref->internaldate+7);
    free(data);
    data = tmp;

    if (!real_from(tmp,entry)) {
	dprint(1, (debugfile,
		   "mbx_copy_envelope_imap: real_from() failed with generated separator line: %s\n",
		   tmp));
	goto clean;
    }
    entry->env_from[0] = '\0';         /* Env from is bogus ... */

    data = tmp;    
    l = strlen(data);
    /* notice that on mailbox separator line we don't use \r\n
     * that even when on 'binary' mail message lines are terminated with 
     * \r\n
     */
    if (!mbx_copy_line_to_temp(folder,data,l)) {
	status = 0;
	goto clean;
    }
    read_state_ptr -> fbytes += l;
    read_state_ptr -> linecounter++;

    read_state_ptr->a.imap_mbx.data_idx = 0;
    change_rec_mbx_info(entry,&imap_mbx_info);
    entry->mbx_info->a.imap_mbx = read_state_ptr->a.imap_mbx.current_message;

    status = 1;
 clean:
    if (data) 
	free(data);
    data = NULL;

    imap_clear_command(folder);
    dprint(11, (debugfile,
		"mbx_copy_envelope_imap=%d\n",
		status));
    return status;
}

static CONST char * mbx_is_forwarded_imap P_((struct folder_info *folder,
					      READ_STATE read_state_ptr));
static CONST char * mbx_is_forwarded_imap(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_is_forwarded_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                     : read_state_ptr=%X\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_is_forwarded_imap=NULL\n"));
    return NULL;
}

static int imap_got_header_line P_((struct folder_info *folder,
				    READ_STATE read_state_ptr,
				    char **buffer,
				    int *len,
				    struct imap_reference  *ref));
static int imap_got_header_line(folder,read_state_ptr,buffer,len,ref)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     struct imap_reference  *ref;
{
    int i;

    if (read_state_ptr->a.imap_mbx.data_idx >= ref->header_len) {
	dprint(95, (debugfile,
		    "imap_got_header_line=0\n"));
	return 0;
    }

    for (i = read_state_ptr->a.imap_mbx.data_idx;
	 i < ref->header_len-1; i++) {
	if ('\r' == ref->header[i] &&
	    '\n' == ref->header[i+1]) {
	    i += 2;
	    break;
	}
    }

    *len = i - read_state_ptr->a.imap_mbx.data_idx;
    *buffer = safe_malloc(*len+1);
    memcpy(*buffer,
	   ref->header + read_state_ptr->a.imap_mbx.data_idx,
	   *len);

    dprint(95, (debugfile,
		"imap_got_header_line=1: len=%d, buffer=%.*s",
		*len,*len,(*buffer) ? (*buffer) : ""));
    if (*len < 1 || !(*buffer) || (*buffer)[*len -1] != '\n') {
	dprint(95, (debugfile,
		    "\nimap_got_header_line: NO NEWLINE\n"));
    } 
    return 1;    
}


static int imap_got_body_line P_((struct folder_info *folder,
				    READ_STATE read_state_ptr,
				    char **buffer,
				    int *len,
				    struct imap_reference  *ref));
static int imap_got_body_line(folder,read_state_ptr,buffer,len,ref)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     struct imap_reference  *ref;
{
    int i;

    if (read_state_ptr->a.imap_mbx.data_idx >= ref->body_len) {
	dprint(95, (debugfile,
		    "imap_got_body_line=0\n"));
	return 0;
    }

    for (i = read_state_ptr->a.imap_mbx.data_idx;
	 i < ref->body_len-1; i++) {
	if ('\r' == ref->body[i] &&
	    '\n' == ref->body[i+1]) {
	    i += 2;
	    break;
	}
    }


    *len = i - read_state_ptr->a.imap_mbx.data_idx;
    *buffer = safe_malloc(*len+1);
    memcpy(*buffer,
	   ref->body + read_state_ptr->a.imap_mbx.data_idx,
	   *len);

    dprint(95, (debugfile,
		"imap_got_body_line=1: len=%d, buffer=%.*s",
		*len,*len,(*buffer) ? (*buffer) : ""));
    if (*len < 1 || !(*buffer) || (*buffer)[*len -1] != '\n') {
	dprint(95, (debugfile,
		    "\nimap_got_body_line: NO NEWLINE\n"));
    } 
    return 1;    
}


static void imap_accept_line P_((struct folder_info *folder,
				 READ_STATE read_state_ptr,
				 char **buffer, int *len));
static void imap_accept_line(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    if (!*buffer)
	return;

    if (*len < 1 || (*buffer)[*len -1] != '\n') {
    } else {
	read_state_ptr -> linecounter++;
    }
    read_state_ptr -> fbytes              += *len;
    read_state_ptr->a.imap_mbx.data_idx   += *len;

    free(*buffer);
    *buffer = NULL;
    *len    = 0;
}



static int mbx_copy_header_imap P_((struct folder_info *folder,
				   READ_STATE read_state_ptr,
				   char **buffer, int *len));
static int mbx_copy_header_imap(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    int status = 0;
    char *buffer1 = NULL;
    int len1,i;
    struct imap_reference  *ref = NULL;

    CONST char * L[] = { "BODY.PEEK[TEXT]", NULL };

    dprint(11, (debugfile,
		"mbx_copy_header_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                   : read_state_ptr=%X\n",
		read_state_ptr));

    /* Wait that IDLE timer command is completetd so that there
     * is no EXPUNGE coming...
     */
    imap_idle_wait(folder);

    /* Check that message is not deleted meanwhile --
       Should not happen: EXPUNGE response is not allowed during of FETCH
       */
    ref = find_reference(& (folder -> p->a.imap_mbx),
			 & (read_state_ptr->a.imap_mbx.current_message));
    if (!ref) {
	dprint(1, (debugfile,
		   "mbx_copy_header_imap=0: EXPUNGE middle of FETCH ???\n"));
	return 0;
    }

hide:
    if (!imap_got_header_line(folder,read_state_ptr,&buffer1,&len1,ref)) {
	status = 0;
	goto clean;
    }

    if ((len1 == 1 && 0 == memcmp(buffer1,"\n",len1)) ||
	(len1 == 2 && 0 == memcmp(buffer1,"\r\n",len1))) {
	imap_states res;

	/* End of headers */
	if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	    status = 0;
	    goto clean;
	}
	imap_accept_line(folder,read_state_ptr,
			&buffer1,&len1);
	status = 0;

	read_body:

	/* Free used data */
	free(ref->header);
	ref->header     = NULL;
	ref->header_len = 0;

	/* FETCH body */
	if (!start_imap_command(folder,"FETCH")) 
	    goto clean;

	imap_command_push_number(folder,
				 read_state_ptr->
				 a.imap_mbx.current_message.last_mbx_index);
	imap_command_push_list(folder,L);
	end_imap_command(folder);

	imap_command_ok(folder,&res,NULL);

	goto clean;
    }
    if (NULL == memchr(buffer1,':',len1)) {
	/* End of headers -- bad header */
	dprint(1, (debugfile,
		   "mbx_copy_header_imap: IMAP server is returning bad header...\n"));
	status = 0;
	goto read_body;
    }

    do {
	append_buffer(buffer,len,buffer1,len1);
	if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	    status = 0;
	    goto clean;
	}
	imap_accept_line(folder,read_state_ptr,
			 &buffer1,&len1);
	if (!imap_got_header_line(folder,read_state_ptr,
				  &buffer1,&len1,ref)) {
	    break;
	}
    } while (len1 > 0 && whitespace(buffer1[0]));
    
    if (*buffer && *len > 6 &&
	header_cmp(*buffer, "Status", NULL)) {
	dprint(11, (debugfile,
		    "                   : Hiding header %s",*buffer));
	free(*buffer);
	*buffer = NULL;
	*len = 0;
	goto hide;
    }
	
    status = 1;

clean:
    if (buffer1) {
	free(buffer1);
	buffer1 = NULL;
    }

    if (!status) {
	/* So that reading of body starts correctly */
	read_state_ptr->a.imap_mbx.data_idx = 0;
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_header_imap=%d | len=%d, buffer=%s",
                status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
        dprint(50, (debugfile,
                    "\nmbx_copy_header_imap <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_body_imap P_((struct folder_info *folder,
				     READ_STATE read_state_ptr,
				     char **buffer, int *len,
				     long *content_remaining));

static int mbx_copy_body_imap(folder,read_state_ptr,buffer,len,
			      content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     long *content_remaining;
{ 
    char *buffer1 = NULL;
    int len1;
    int status = 0;
    struct imap_reference  *ref = NULL;

    dprint(11, (debugfile,
		"mbx_copy_body_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                 : read_state_ptr=%X\n",
		read_state_ptr));

    /* Check that message is not deleted meanwhile --
       Should not happen: EXPUNGE response is not allowed during of FETCH
       */
    ref = find_reference(& (folder -> p->a.imap_mbx),
			 & (read_state_ptr->a.imap_mbx.current_message));
    if (!ref) {
	dprint(1, (debugfile,
		   "mbx_copy_body_imap=0: EXPUNGE middle of FETCH ???\n"));
	return 0;
    }

    if (!imap_got_body_line(folder,read_state_ptr,
			    &buffer1,&len1,ref)) {
	status = 0;
	goto clean;
    }
    
    append_buffer(buffer,len,buffer1,len1);
    if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	status = 0;
	goto clean;
    }
    imap_accept_line(folder,read_state_ptr,
		     &buffer1,&len1);
    status = 1;

 clean:
    if (buffer1) {
	free(buffer1);
	buffer1 = NULL;
    }

    if (!status) {
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_body_imap=%d | len=%d, buffer=%s",
		status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
	dprint(50, (debugfile,
		    "\nmbx_copy_body_imap <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_envelope_end_imap P_((struct folder_info *folder,
					     READ_STATE read_state_ptr));
static int mbx_copy_envelope_end_imap(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    int status = 0;
    struct imap_reference  *ref = NULL;

    dprint(11, (debugfile,
		"mbx_copy_envelope_end_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                         : read_state_ptr=%X\n",
		read_state_ptr));

    /* Check that message is not deleted meanwhile --
       Should not happen: EXPUNGE response is not allowed during of FETCH
       */
    ref = find_reference(& (folder -> p->a.imap_mbx),
			 & (read_state_ptr->a.imap_mbx.current_message));
    if (ref) {
	if (ref->header) {
	    /* Free used data */
	    free(ref->header);
	    ref->header     = NULL;
	    ref->header_len = 0;
	}
	if (ref->body) {
	    /* Free used data */
	    free(ref->body);
	    ref->body     = NULL;
	    ref->body_len = 0;
	}

	/* Got next UID */
	if (ref+1 < 
	    folder -> p->a.imap_mbx.references + 
	    folder -> p->a.imap_mbx.reference_count) {
	    read_state_ptr->a.imap_mbx.current_message.uid_number =
		(ref+1)->uid_number;
	    read_state_ptr->a.imap_mbx.current_message.last_mbx_index =
		(ref+1) - folder -> p->a.imap_mbx.references;
	} else {
	    dprint(11, (debugfile,
			"mbx_copy_envelope_end_imap: End of mailbox\n"));
	    read_state_ptr->a.imap_mbx.current_message.uid_number++;
	}

    } else {
	read_state_ptr->a.imap_mbx.current_message.uid_number++;

	dprint(1, (debugfile,
		   "mbx_copy_envelope_end_imap: EXPUNGE middle of FETCH ???\n"));
    } 

    /* On mailbox format there is empty line (\n\n)
     * on end of mail...
     */

    if (!mbx_copy_line_to_temp(folder,"\n",1)) {
	status = 0;
	goto clean;
    }
    read_state_ptr -> fbytes ++;    /* Increment position */
    status = 1;

clean:
    imap_clear_command(folder);

    dprint(11, (debugfile,
		"mbx_copy_envelope_end_imap=%d\n",
		status));
    return status; /* Always end of message */
}

static int mbx_copy_envelope_reset_body_imap P_((struct folder_info *folder,
						READ_STATE read_state_ptr));
static int mbx_copy_envelope_reset_body_imap(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                                : read_state_ptr=%X\n",
		read_state_ptr));
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_imap=0\n"));
    return 0;
}

static FILE * mbx_imap_to_fd P_((struct folder_info *folder,long offset));
static FILE * mbx_imap_to_fd(folder,offset)
     struct folder_info *folder;
     long offset;
{
    FILE * status = NULL;
    dprint(11, (debugfile,
		"mbx_imap_to_fd: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (offset != -1L &&
	fseek(folder->p->fh_temp, offset , 
	      SEEK_SET) == -1) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCouldntSeekBytesIntoFolder,
			  "\nCouldn't seek %ld bytes into folder.\n"),
		  offset);		
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile,
		   "Error: Couldn't seek folder %s: (offset %ld) Errno %s (%s)\n",
		   folder->cur_tempfolder, offset, 
		   error_description(err), "mbx_imap_to_fd"));
	status = NULL;
	goto clean;
    }
    status = folder->p->fh_temp;

 clean:
    dprint(11, (debugfile,
		"mbx_imap_to_fd=%s (%X)\n",
		status ? "NON-NULL" : "NULL",
		status));
    return status;
}

static int mbx_new_mail_on_imap P_((struct folder_info *folder,int *bytes));
static int mbx_new_mail_on_imap(folder,bytes)
     struct folder_info *folder;
     int *bytes;
{
    int status = 0;
    int i;
    int start = 1;
    imap_states res;
	
    
    dprint(11, (debugfile,
		"mbx_new_mail_on_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    *bytes = 0;

    if (!start_imap_command(folder,"NOOP")) 
	goto clean;
    end_imap_command(folder);
    if (!imap_command_ok(folder,&res,NULL))
	goto clean;

    imap_clear_command(folder);

    /* Messages which do not have UID have unknown ... */
    
    start = 1;
    while (start < folder -> p->a.imap_mbx.reference_count &&
	   folder -> p->a.imap_mbx.references[start].uid_number != -1) {
	start++;
    }

    dprint(11, (debugfile,
		"                    : start=%d refcount=%d last uid number=%d\n",
		start,folder -> p->a.imap_mbx.reference_count,
		folder -> p->a.imap_mbx.last_uid_number));
	
    if (start <= folder -> p->a.imap_mbx.reference_count-1) {	    
	
	CONST char * L[] = { "RFC822.SIZE", "UID", NULL };
	
	if (!start_imap_command(folder,"FETCH")) {
	    status = 0;
	    goto clean;
	}
	
	imap_command_push_range(folder,
				start,
				folder -> p->a.imap_mbx.reference_count-1);
	
	imap_command_push_list(folder,L);
	end_imap_command(folder);
	
	if (!imap_command_ok(folder,&res,NULL))
	    goto clean;
    }

    for (i = 1; i < folder -> p->a.imap_mbx.reference_count; i++) {
	if (folder -> p->a.imap_mbx.references[i].uid_number >
	    folder -> p->a.imap_mbx.last_uid_number) {
	    status = 1;
	    *bytes += folder-> p->a.imap_mbx.references[i].rfc822_size;
	}
    }
    
 clean:
    imap_clear_command(folder);

    dprint(11, (debugfile,
		"mbx_new_mail_on_imap=%d (*bytes=%d)\n",
		status,*bytes));
    return status;
}

static int mbx_consider_remove_imap P_((struct folder_info *folder));
static int mbx_consider_remove_imap(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_consider_remove_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_consider_remove_imap=0\n"));
    return 0;
}

static int mbx_prepare_keep_imap P_((struct folder_info *folder,
				    KEEP_STATE keep_state_ptr));
static int mbx_prepare_keep_imap(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    int status = 0; 
    dprint(11, (debugfile,
		"mbx_prepare_keep_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));
    status = 1;   
    dprint(11, (debugfile,
		"mbx_prepare_keep_imap=%d\n",
		status));
    return status;
}

static int mbx_end_keep_imap P_((struct folder_info *folder,
				KEEP_STATE keep_state_ptr));
static int mbx_end_keep_imap(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    char *fname;
    int status = 0;
    imap_states res;

    dprint(11, (debugfile,
		"mbx_end_keep_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));



    if (0 != (folder->p->a.imap_mbx.folder_status & IMAP_flusherr))
	goto clean;

    if (!start_imap_command(folder,"EXPUNGE"))
	goto clean;
    end_imap_command(folder);
    
    if (!imap_command_ok(folder,&res,NULL))
	goto clean;
    imap_clear_command(folder);

    if (!start_imap_command(folder,"CHECK"))
	goto clean;
    end_imap_command(folder);
    
    if (!imap_command_ok(folder,&res,NULL))
	goto clean;
        
    status = 1;

 clean:
    folder->p->a.imap_mbx.folder_status &= IMAP_flusherr;
    imap_clear_command(folder);

    dprint(11, (debugfile,
		"mbx_end_keep_imap=%d\n",
		status));
    return status;
}

static void mbx_mark_keep_imap P_((struct folder_info *folder,
				      KEEP_STATE keep_state_ptr,
				      struct header_rec *entry,
				      int keep));
static void mbx_mark_keep_imap(folder,keep_state_ptr,entry,keep)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
     struct header_rec *entry;
     int keep;
{
    dprint(11, (debugfile,
		"mbx_mark_keep_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    /* Wait that IDLE timer command is completetd so that there
     * is no EXPUNGE coming...
     */
    imap_idle_wait(folder);

    if(entry->mbx_info &&
       &imap_mbx_info == entry->mbx_info->type_code) {

	struct imap_reference  *ref = 
	    find_reference(& (folder -> p->a.imap_mbx),
			   & (entry->mbx_info->a.imap_mbx));
	    
	if (ref) {
	    int j;

	    CONST char * LIST[ sizeof IMAP_flags / 
			     sizeof (struct imap_flag) +1]; 
	    imap_states res;

	    int X;

	    if (0 == (entry->status & UNREAD))
		ref->imap_flags |= IMAP_Seen;
	    
	    for (j = 0; 
		 j < sizeof IMAP_flags / sizeof (struct imap_flag); 
		 j++) {

		if (0 != (IMAP_flags[j].mask_elm & entry->status))
		    ref->imap_flags |= IMAP_flags[j].mask_imap;		 
	    }
	    if (keep)
		ref->imap_flags &= ~IMAP_Deleted;
	    else
		ref->imap_flags |= IMAP_Deleted;
	    

	    if (!start_imap_command(folder,"STORE")) {
		folder->p->a.imap_mbx.folder_status |= IMAP_flusherr;
		goto clean;
	    }
	    imap_command_push_number(folder,
				     entry->mbx_info->
				     a.imap_mbx.last_mbx_index);

	    dprint(14, (debugfile,
			"mbx_mark_keep_imap: (mailbox index %d) Flags",
			entry->mbx_info->a.imap_mbx.last_mbx_index));

	    imap_command_push_atom(folder,"FLAGS");

	    X = 0;
	    for (j = 0; 
		 j < sizeof IMAP_flags / sizeof (struct imap_flag); 
		 j++) {
		if (0 != (ref->imap_flags & IMAP_flags[j].mask_imap) &&
		    0 == (IMAP_Recent     & IMAP_flags[j].mask_imap)) {
		    LIST[X++] = IMAP_flags[j].flag;
		    dprint(14, (debugfile," %s",IMAP_flags[j].flag));
		}
	    }	    
	    LIST[X] = NULL;

	    dprint(14, (debugfile,"\n"));

	    imap_command_push_list(folder,LIST);
	    end_imap_command(folder);

	    if (!imap_command_ok(folder,&res,NULL)) {
		folder->p->a.imap_mbx.folder_status |= IMAP_flusherr;
		goto clean;
	    }

	}
    }
clean:
    imap_clear_command(folder);
}

CONST char * mbx_imap_type P_((struct folder_info *folder));
CONST char * mbx_imap_type(folder)
     struct folder_info *folder;
{
    char *result = NULL;

    dprint(11, (debugfile,
		"mbx_imap_type: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (0 != (folder -> p->a.imap_mbx.folder_status & IMAP_writable)) {

	if (0 == istrcmp(folder -> p->a.imap_mbx.folder,
			 "INBOX")) {
	    static char *mailbox = NULL;

	    if (!mailbox)
		mailbox = catgets(elm_msg_cat, MeSet, MeImapMailbox, 
				  "IMAP mailbox");
	    result = mailbox;
	} else {
	    static char *mailbox_normal = NULL;

	    if (!mailbox_normal)
		mailbox_normal = catgets(elm_msg_cat, MeSet, MeImapFolder, 
				  "IMAP folder");
	    result = mailbox_normal;
	}
    } else {
	static char *ro_mailbox = NULL;
	
	if (!ro_mailbox)
	    ro_mailbox = catgets(elm_msg_cat, MeSet, MeROImapFolder, 
				 "Read-only IMAP folder");	
	result = ro_mailbox;
    }

    dprint(11, (debugfile,
		"mbx_imap_type=%s\n",
	        result));
    return result;
}


static int mbx_start_edit_imap P_((struct folder_info *folder, 
				  CONST char **buffer));
static int mbx_start_edit_imap(folder,buffer)
     struct folder_info *folder;
     CONST char **buffer;
{
    dprint(11, (debugfile,
		"mbx_start_edit_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    dprint(11, (debugfile,
		"mbx_start_edit_imap=0\n"));
    return 0;
}

static int mbx_end_edit_imap P_((struct folder_info *folder));
static int mbx_end_edit_imap(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_end_edit_imap: folder=%X (%s)\n",
		folder,folder->cur_folder));

    dprint(11, (debugfile,
		"mbx_end_edit_imap=0\n"));
    return 0;
}

static void mbx_free_imap P_((struct folder_info *folder));
static void mbx_free_imap(folder)
     struct folder_info *folder;
{
  dprint(11, (debugfile,
	      "mbx_free_imap: folder=%X (%s)\n",
	      folder,folder->cur_folder));    

    if (folder -> p->fh_temp) {
	fclose(folder -> p->fh_temp);
	folder -> p -> fh_temp = NULL;
    }
    if (folder -> p -> fh_folder) {
	fclose(folder -> p -> fh_folder);	
	folder -> p -> fh_folder = NULL;
    }

    free_remote_account(&(folder -> p->a.imap_mbx.C));
    if (folder -> p->a.imap_mbx.folder) {
	free(folder -> p->a.imap_mbx.folder);
	folder -> p->a.imap_mbx.folder = NULL;
    }

    imap_clear_buffers(folder);

    free(folder -> p);
    folder -> p = NULL;
}


static void mbx_zero_rs_fields_imap P_((struct read_folder_state *rs));
static void mbx_zero_rs_fields_imap(rs)
     struct read_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_zero_rs_fields_imap: rs=%X\n",
		rs));

    rs->a.imap_mbx.current_message.last_mbx_index = -1;
    rs->a.imap_mbx.current_message.uid_number     = 0;
    rs->a.imap_mbx.data_idx = 0;
}

static void mbx_free_rs_fields_imap P_((struct read_folder_state *rs));
static void mbx_free_rs_fields_imap(rs)
     struct read_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_free_rs_fields_imap: rs=%X\n",
		rs));

    rs->a.imap_mbx.current_message.last_mbx_index = -1;
    rs->a.imap_mbx.current_message.uid_number     = 0;
    rs->a.imap_mbx.data_idx = 0;    
}

static void mbx_zero_ks_fields_imap P_((struct keep_folder_state *rs));
static void mbx_zero_ks_fields_imap(rs)
     struct keep_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_zero_ks_fields_imap: rs=%X\n",
		rs));
}

static void mbx_free_ks_fields_imap P_((struct keep_folder_state *rs));
static void mbx_free_ks_fields_imap(rs)
     struct keep_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_free_ks_fields_imap: rs=%X\n",
		rs));
}

static int mbx_get_imap_mode P_((struct folder_info *folder));
static int mbx_get_imap_mode(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_get_imap_mode: folder=%X (%s)\n",
		folder,folder->cur_folder));

    if (0 != (folder -> p->a.imap_mbx.folder_status & IMAP_writable)) {
	if (0 == istrcmp(folder -> p->a.imap_mbx.folder,
			 "INBOX")) {	    
	    dprint(11, (debugfile,
			"mbx_get_imap_mode=FOLDER_MBOX\n"));
	    return FOLDER_MBOX;
	} else {
	    dprint(11, (debugfile,
			"mbx_get_imap_mode=0\n"));
	    return 0;
	}
    } else {
	dprint(11, (debugfile,
		    "mbx_get_imap_mode=FOLDER_RDONLY\n"));
	return FOLDER_RDONLY;
    }
}


struct folder_type imap_mbx = { mbx_close_imap,
				mbx_lock_imap,
				mbx_init_imap,
				mbx_sessionlock_imap,
				mbx_unlock_imap,
				mbx_flush_imap,
				mbx_ferror_imap,
				mbx_prepare_read_imap,  
				mbx_end_read_imap,
				mbx_copy_envelope_imap,
				mbx_is_forwarded_imap,
				mbx_copy_header_imap,
				mbx_copy_body_imap,
				mbx_copy_envelope_end_imap,
				mbx_copy_envelope_reset_body_imap,
				mbx_imap_to_fd,
				mbx_new_mail_on_imap,
				mbx_consider_remove_imap,
				mbx_prepare_keep_imap,
				mbx_end_keep_imap,
				mbx_mark_keep_imap,
				mbx_imap_type,
				mbx_start_edit_imap,
				mbx_end_edit_imap,
				mbx_free_imap,
				mbx_zero_rs_fields_imap,
				mbx_free_rs_fields_imap,
				mbx_zero_ks_fields_imap,
				mbx_free_ks_fields_imap,
				mbx_get_imap_mode };
#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


