static char rcsid[] = "@(#)$Id: mbox.c,v 1.16 2000/07/30 15:12:10 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.16 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************
 * Based on code ../src/newmbox.c and ../src/lock.c. 
 * That code was following copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#include "headers.h"
#include "mbx_imp.h"
#include "s_me.h"
#include "s_elm.h"

#include <sys/stat.h>

#ifdef I_UTIME
#  include <utime.h>
#endif
#ifdef I_SYSUTIME
#  include <sys/utime.h>
#endif

/* Locking primites ------------------------------------------------------- */

#ifdef SYSCALL_LOCKING
#  if (defined(BSD_TYPE) || !defined(apollo))
#    include <sys/file.h>
#  endif
#endif

#include <errno.h>

static int GrabRead_the_file P_((int flock_fd));

static int GrabRead_the_file(flock_fd)
     int flock_fd;
{

    int	retcode	= FLOCKING_OK;
#ifdef USE_FCNTL_LOCKING
    struct flock lock_info;
#endif

    errno = 0;

#ifdef   USE_FCNTL_LOCKING
    lock_info.l_type = F_RDLCK;
    lock_info.l_whence = 0;
    lock_info.l_start = 0;
    lock_info.l_len = 0;
    
    if (fcntl(flock_fd, F_SETLK, &lock_info) != 0) {
	int err = errno;
	dprint(25, (debugfile,
		    "GrabRead_the_file: fcntl: %s (errno %d)\n",
		    error_description(err),err));
	return ((err == EACCES) || (err == EAGAIN))
	    ? FLOCKING_RETRY
	    : FLOCKING_FAIL ;
    }
#endif
    
#ifdef	USE_FLOCK_LOCKING
    if (flock (flock_fd, LOCK_NB | LOCK_SH)) {
	int err = errno;
	dprint(25, (debugfile,
		    "GrabRead_the_file: flock: %s (errno %s)\n",
		    error_description(err),err));
	retcode = ((err == EWOULDBLOCK) || (err == EAGAIN))
	    ? FLOCKING_RETRY
	    : FLOCKING_FAIL ;
	
#ifdef USE_FCNTL_LOCKING
	lock_info.l_type = F_UNLCK;
	lock_info.l_whence = 0;
	lock_info.l_start = 0;
	lock_info.l_len = 0;
	
	/*
	 *  Just unlock it because we did not succeed with the
	 *  flock()-style locking. Never mind the return value.
	 *  It was our own lock anyway if we ever got this far.
	 */
	fcntl (flock_fd, F_SETLK, &lock_info);
#endif
	return retcode;
    }
#endif

    dprint(25, (debugfile,
		"GrabRead_the_file=FLOCKING_OK\n"));
    return FLOCKING_OK;
}


int Grab_the_file(flock_fd)
     int flock_fd;
{
    int	retcode	= FLOCKING_OK;
#ifdef USE_FCNTL_LOCKING
    struct flock lock_info;
#endif

    errno = 0;

#ifdef   USE_FCNTL_LOCKING
    lock_info.l_type = F_WRLCK;
    lock_info.l_whence = 0;
    lock_info.l_start = 0;
    lock_info.l_len = 0;
    
    if (fcntl(flock_fd, F_SETLK, &lock_info) != 0) {
	int err = errno;
	dprint(25, (debugfile,
		    "Grab_the_file: fcntl: %s (errno %d)\n",
		    error_description(err),err));
	return ((err == EACCES) || (err == EAGAIN))
	    ? FLOCKING_RETRY
	    : FLOCKING_FAIL ;
    }
#endif
    
#ifdef	USE_FLOCK_LOCKING
    if (flock (flock_fd, LOCK_NB | LOCK_EX)) {
	int err = errno;
	dprint(25, (debugfile,
		    "Grab_the_file: flock: %s (errno %d)\n",
		    error_description(err),err));
	retcode = ((err == EWOULDBLOCK) || (err == EAGAIN))
	    ? FLOCKING_RETRY
	    : FLOCKING_FAIL ;
	
#ifdef USE_FCNTL_LOCKING
	lock_info.l_type = F_UNLCK;
	lock_info.l_whence = 0;
	lock_info.l_start = 0;
	lock_info.l_len = 0;
	
	/*
	 *  Just unlock it because we did not succeed with the
	 *  flock()-style locking. Never mind the return value.
	 *  It was our own lock anyway if we ever got this far.
	 */
	fcntl (flock_fd, F_SETLK, &lock_info);
#endif
	return retcode;
    }
#endif

    dprint(25, (debugfile,
		"Grab_the_file=FLOCKING_OK\n"));
    return FLOCKING_OK;
}

int Release_the_file(flock_fd)
     int flock_fd;
{
#ifdef USE_FCNTL_LOCKING
    struct flock lock_info;
#endif

    int	fcntlret = 0,
	flockret = 0,
	fcntlerr = 0,
	flockerr = 0;
    
#ifdef	USE_FLOCK_LOCKING
    errno = 0;
    flockret = flock (flock_fd, LOCK_UN);
    flockerr = errno;
#endif

#ifdef	USE_FCNTL_LOCKING
    lock_info.l_type = F_UNLCK;
    lock_info.l_whence = 0;
    lock_info.l_start = 0;
    lock_info.l_len = 0;

    errno = 0;
    fcntlret = fcntl (flock_fd, F_SETLK, &lock_info);
    fcntlerr = errno;
#endif

    if (fcntlret) {
	errno = fcntlerr;
	dprint(25, (debugfile,
		    "Release_the_file=%d\n",fcntlret));
	return fcntlret;
    } else if (flockret) {
	errno = flockerr;
	dprint(25, (debugfile,
		    "Release_the_file=%d\n",flockret));
	return flockret;
    }
    dprint(25, (debugfile,
		"Release_the_file=%d\n",0));
    return 0;
}

/* Folder type dependent routines ------------------------------------------ */


static void mbx_close_no_name P_((struct folder_info *folder,
				  enum close_mode mode));
static void mbx_close_no_name(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{
    dprint(11, (debugfile,
		"mbx_close_no_name: folder=%X (%s)\n",
		folder,folder->cur_folder));

    return;
}

static void mbx_close_non_spool P_((struct folder_info *folder,
				    enum close_mode mode));
static void mbx_close_non_spool(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{
    dprint(11, (debugfile,
		"mbx_close_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));

    /* normal folder */

    if (folder -> p->fh_folder) {
	switch (mode) {
	case CLOSE_NORMAL:
	    if (lockfolders)
		Release_the_file(fileno(folder -> p->fh_folder));      
	    break;
	case CLOSE_LEAVE_LOCKED:
	    break;
	}
    }

    return;
}


static void mbx_close_spool P_((struct folder_info *folder,
				enum close_mode mode));
static void mbx_close_spool(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{    
    dprint(11, (debugfile,
		"mbx_close_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));


    switch (mode) {
    case CLOSE_NORMAL:

	if (folder -> p->a.spool.lock_state == ON)
	    unlock_folder(0,folder);
    
	/* falltru */
    case CLOSE_LEAVE_LOCKED:
	if (!folder -> p->fh_temp) {
	    dprint(1, (debugfile,
		       "mbx_close_spool: tempfile (%s) not open -- will not unlink\n",
		       folder -> cur_tempfolder));
	} else if (unlink(folder -> cur_tempfolder) != 0) {
	    if (errno != ENOENT) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantUnlinkTemp,
				  "Sorry, can't unlink the temp file %s [%s]!\n\r"),
			  folder -> cur_tempfolder, error_description(errno));
	    }
	} else {
	    dprint(1, (debugfile, 
		       "close_folder: Unlinking tempfile (%s).\n",
		       folder -> cur_tempfolder));
	}      
    }
}

static int  mbx_lock_no_name  P_((int direction,struct folder_info *folder));
static int  mbx_lock_no_name(direction,folder)
     int direction;
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_lock_no_name: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_lock_no_name=0\n"));
    return 0;
}

static int  mbx_lock_non_spool P_((int direction,
				    struct folder_info *folder));
static int  mbx_lock_non_spool(direction,folder)
     int direction;
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_lock_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_lock_non_spool=1 (dummy)\n"));
    return 1;
}


#ifdef	USE_DOTLOCK_LOCKING		

static int  mbx_dotlock_file P_((int direction,
				 struct folder_info *folder));
static int  mbx_dotlock_file(direction,folder)
     int direction;
     struct folder_info *folder;
{
    int status = 0;
    int create_iteration = 0;
    char pid_buffer[SHORT];

    struct stat    buf;		/* stat command  */
    int stat_code = -1;

    int err       = 0;
    int temp_fd = -1;
    int create_fd = -1;	/* file descriptor for creating lock file */

#ifdef SAVE_GROUP_MAILBOX_ID
    gid_t oldgid = getegid();
    setgid(mailgroupid); /* reset id so that we can get at lock file */
#endif

    if(folder->p->a.spool.lock_state == ON) {
	dprint(1, (debugfile,
		   "mbx_dotlock_file: Folder %s already locked.\n", 
		   folder->cur_folder));
	if (direction == INCOMING)
	    dprint(1, (debugfile,
		       "mbx_dotlock_file: ERROR: ... And direction INCOMING! Software error!\n"));
	else {
	    status = 1;
	    goto clean;
	}		   
    }
    
#ifdef PIDCHECK
    /** first, try to read the lock file, and if possible, check the pid.
	If we can validate that the pid is no longer active, then remove
	the lock file.
    **/

    if((temp_fd=open(folder->p->a.spool.lockfile,O_RDONLY)) != -1) {
	if (read(temp_fd, pid_buffer, SHORT) > 0) {
	    int pid_value = atoi(pid_buffer);
	    if (pid_value) {
		int kill_status = kill(pid_value, 0);
		if (kill_status != 0 && errno != EPERM) {
		    if (unlink(folder->p->a.spool.lockfile) != 0) {
			dprint(1, (debugfile,
				   "Error %s\n\ttrying to unlink file %s (%s)\n", 
				   error_description(errno), 
				   folder->p->a.spool.lockfile, "lock"));
			lib_error(CATGETS(elm_msg_cat, ElmSet, 
					  ElmLeaveCouldntRemoveCurLock,
					  "\nCouldn't remove the current lock file %s\n"), 
				  folder->p->a.spool.lockfile);
			lib_error(FRM("** %s **\n"), 
				  error_description(errno));
			close(temp_fd);
			status = 0;
			goto clean;
		    }
		}
	    }
	}
	close(temp_fd);
    }
#endif
	      
    /* try to assert create lock file MAX_ATTEMPTS times */
    do {
	err = 0;
	if((create_fd=open(folder->p->a.spool.lockfile,
				   O_WRONLY | O_CREAT | 
				   O_EXCL,0444)) != -1)
	    break;
	else {
	    err = errno;
	    stat_code = stat(folder->p->a.spool.lockfile,&buf);

	    /* Creation of lock failed NOT because it already exists!!! */
	    /* If /var/mail nfs mounted on Solaris 2.3 at least you can */
	    /* get EACCES.  Treat it like EEXIST.                       */

	    if(err != EEXIST && stat_code < 0) {		    
		    
		if (err != EACCES || !is_in_spool(folder->p->a.spool.lockfile)) {
		    dprint(1, (debugfile, 
			       "Error encountered attempting to create lock %s\n", 
			       folder->p->a.spool.lockfile));
		    dprint(1, (debugfile, "** %s **\n", error_description(err)));
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmLeaveErrorCreatingLock,
				      "\nError encountered while attempting to create lock file %s;\n"),
			      folder->p->a.spool.lockfile);
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmErrorAccess,
				      "%.45s: %.33s"),
			      folder->p->a.spool.lockfile,
			      error_description(err));
		} else {	
		    dprint(1, (debugfile,
			       "Can't create lock file: creat(%s) raises error %s (lock)\n", 
			       folder->p->a.spool.lockfile, 
			       error_description(errno)));
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmLeaveCantCreateLock,
				      "Can't create lock file! Need write permission in \"%s\".\n"),
			      mailhome);
		}
		status = 0;
		goto clean;
	    }
	}
	dprint(2, (debugfile,"File '%s' already exists!  Waiting...(lock)\n", 
		   folder->p->a.spool.lockfile));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveWaitingToRead,
			  "Waiting to read mailbox while mail is being received: attempt #%d"),
		  create_iteration);
#if POLL_METHOD	  
	wait_for_timeout(5);
#else
	sleep(5);
#endif
    } while (create_iteration++ < MAX_ATTEMPTS);

    if(create_fd == -1) {	
	struct stat    buf2;	/* stat command  */
	int stat2_code = stat(folder->cur_folder,&buf2);
	time_t now = time(NULL);
	int can_remove = 0;

	dprint(2, (debugfile,
		   "mbx_dotlock_file: now       = %d \n",
		   now)); 
	if (stat_code == 0) {
	    dprint(1, (debugfile,
		       "mbx_dotlock_file: lock time = %d (%s)\n",
		       buf.st_mtime,folder->p->a.spool.lockfile));
	    can_remove = buf.st_mtime < now - 10 * 60;
	    dprint(1, (debugfile,
		       "                  (modified %1.1f minutes ago)\n",
		       (now - buf.st_mtime) / 60.0));
	}
	if (stat2_code == 0) {
	    dprint(1, (debugfile,
		       "mbx_dotlock_file: read time = %d (%s)\n",
		       buf2.st_atime,folder->cur_folder));
	    if (buf2.st_atime > now - 10 * 60)
		can_remove = 0;
	    dprint(1, (debugfile,
		       "                  (readed %1.1f minutes ago)\n",
		       (now - buf2.st_atime) / 60.0));
	}
	dprint(2, (debugfile,
		   "mbx_dotlock_file: can remove = %d \n",
		   can_remove)); 

	/* we weren't able to create the lock file */
	
	if (can_remove) {
	    /** time to waste the lock file!  Must be there in error! **/
	    dprint(2, (debugfile, 
		       "Warning: I'm giving up waiting - removing lock file(lock)\n"));
	    if (direction == INCOMING)
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmLeaveTimedOutRemoving,
				  "\nTimed out - removing current lock file..."));
	    else
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveThrowingAwayLock,
				  "Throwing away the current lock file!"));

	    if (unlink(folder->p->a.spool.lockfile) != 0) {
		dprint(1, (debugfile,
			   "Error %s\n\ttrying to unlink file %s (%s)\n", 
			   error_description(errno), folder->p->a.spool.lockfile, "lock"));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveCouldntRemoveCurLock,
				  "\nCouldn't remove the current lock file %s\n"), 
			  folder->p->a.spool.lockfile);	     
		lib_error(FRM("** %s **\n"), 
			  error_description(errno));
		status = 0;
		goto clean;
	    }

	    /* we've removed the bad lock, let's try to assert lock once more */
	    if((create_fd=open(folder->p->a.spool.lockfile,
			       O_WRONLY | O_CREAT | O_EXCL,0444)) == -1){

		/* still can't lock it - just give up */
		dprint(1, (debugfile, 
			   "Error encountered attempting to create lock %s\n", 
			   folder->p->a.spool.lockfile));
		dprint(1, (debugfile, "** %s **\n", error_description(errno)));

		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorCreatingLock,
				  "\nError encountered while attempting to create lock file %s;\n"),
			  folder->p->a.spool.lockfile);	    
		lib_error(FRM("** %s. **\n\n"), 
			  error_description(errno));
		status = 0;
		goto clean;
	    }
	}
	else {
	    /* Okay...we die and leave, not updating the mailfile mbox or
	       any of those! */

	    if (direction == INCOMING) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveGivingUp,
				  "\n\nGiving up after %d iterations.\n"), 
			  create_iteration);		
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeavePleaseTryAgain,
				  "\nPlease try to read your mail again in a few minutes.\n\n"));
		dprint(1, (debugfile, 
			   "Warning: bailing out after %d iterations...(lock)\n",
			   create_iteration));
		status = 0;
		goto clean;
		
	    } else {
		dprint(1, (debugfile, 
			   "Warning: after %d iterations, timed out! (lock)\n",
			   create_iteration));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorTimedOutLock,
				  "Timed out on locking mailbox.  Leaving program.\n"));
		status = 0;
		goto clean;
	    }
	}
    }

    status = 1;
    /* If we're here we successfully created the lock file */
    dprint(5,
	   (debugfile, "Lock %s %s for file %s on.\n", 
	    folder->p->a.spool.lockfile,
	    (direction == INCOMING ? "incoming" : "outgoing"), 
	    folder->cur_folder));
    
    /* Place the pid of Elm into the lock file for SVR3.2 and its ilk */
    elm_sfprintf(pid_buffer, sizeof pid_buffer,
		 FRM("%d"), 
		 getpid());
    write(create_fd, pid_buffer, strlen(pid_buffer));
    
 clean:
    if (create_fd != -1)
	close(create_fd);

#ifdef SAVE_GROUP_MAILBOX_ID
    setgid(oldgid); 
#endif
    dprint(2, (debugfile,
	       "mbx_dotlock_file = %d \n",
	       status));

    return status;
}
#endif

#ifdef SYSCALL_LOCKING
static int  mbx_syscall_lock_file P_((int direction,
				      struct folder_info *folder));
static int  mbx_syscall_lock_file(direction,folder)
     int direction;
     struct folder_info *folder;
{
    int status = 0;
    int stat = FLOCKING_FAIL;
    int flock_iteration = 0;

    /* try to assert lock MAX_ATTEMPTS times */
    do {
	stat = Grab_the_file (fileno(folder->p->fh_folder));
	if (FLOCKING_OK == stat)
	    break;

	switch (stat) {	    
	case	FLOCKING_FAIL:
	    /*
	     *	Creation of lock failed
	     *	NOT because it already exists!!!
	     */
	    
	    dprint (1, (debugfile, 
			"Error encountered attempting to flock %s\n",
			folder->cur_folder));
	    dprint (1, (debugfile, "** %s **\n", 
			error_description(errno)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorFlockMailbox,
			      "\nError encountered while attempting to flock mailbox %s;\n"), 
		      folder->cur_folder);	    
	    lib_error(FRM("** %s. **\n\n"), 
		      error_description(errno));
	    
	    status = 0;
	    goto clean;

	case	FLOCKING_RETRY:
	default:
	    dprint (2, (debugfile,
			"Mailbox '%s' already locked!  Waiting...(lock)\n", 
			folder->cur_folder));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveWaitingToRead,
			      "Waiting to read mailbox while mail is being received: attempt #%d"),
		      flock_iteration);
#if POLL_METHOD	  
	    wait_for_timeout(5);
#else
	    sleep(5);
#endif
	}
      } while (flock_iteration++ < MAX_ATTEMPTS);

      lib_transient(FRM(""));

      if (stat != FLOCKING_OK) {
	  /* We couldn't lock the file. We die and leave not updating
	   * the mailfile mbox or any of those! */
	  
	  if (direction == INCOMING) {
	      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveGivingUp,
				"\n\nGiving up after %d iterations.\n"), 
			flock_iteration);	    
	      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeavePleaseTryAgain,
				"\nPlease try to read your mail again in a few minutes.\n\n"));
	      dprint(1, (debugfile, 
			 "Warning: bailing out after %d iterations...(lock)\n",
			 flock_iteration));
	  } else {
	      dprint(1, (debugfile, 
			 "Warning: after %d iterations, timed out! (lock)\n",
			 flock_iteration));
	  }
	  lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorTimedOutLock,
			    "Timed out on locking mailbox.  Leaving program.\n"));
	  status = 0;
	  goto clean;
      }

      /* We locked the file */
      dprint(5,
	     (debugfile, "Lock %s on file %s on.\n",
	      (direction == INCOMING ? "incoming" : "outgoing"), 
	      folder->cur_folder));
      status = 1;

 clean:
      dprint(2, (debugfile,
		 "mbx_syscall_lock_file = %d \n",
		 status));
      return status;
}
#endif

static int  mbx_lock_spool P_((int direction,
			       struct folder_info *folder));
static int  mbx_lock_spool(direction,folder)
     int direction;
     struct folder_info *folder;
{
     /** Create lock file to ensure that we don't get any mail 
	 while altering the folder contents!
	 If it already exists sit and spin until 
	 either the lock file is removed...indicating new mail
	 or
	 we have iterated MAX_ATTEMPTS times, in which case we
	 either fail or remove it and make our own (determined
	 by if REMOVE_AT_LAST is defined in header file
	 If direction == INCOMING then DON'T remove the lock file
	 on the way out!  (It'd mess up whatever created it!).
	 
	 But if that succeeds and if we are also locking by flock(),
	 follow a similar algorithm. Now if we can't lock by flock(),
	 we DO need to remove the lock file, since if we got this far,
	 we DID create it, not another process.
      **/
    
    int status = 0;

    dprint(11, (debugfile,
		"mbx_lock_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));

#ifdef	USE_DOTLOCK_LOCKING	
    if (!mbx_dotlock_file(direction,folder)) {
	status = 0;
	goto clean;
    }
#endif	

#ifdef SYSCALL_LOCKING
      /* Now we also need to lock the file with flock(2) */

    if(!folder->p->fh_folder) {
	dprint(11, (debugfile,
		    "mbx_lock_spool: non-existent -- no syscall locking.\n"));
    } else if (!mbx_syscall_lock_file(direction,folder)) {
#ifdef	USE_DOTLOCK_LOCKING
	(void)unlink(folder->p->a.spool.lockfile);
#endif /* USE_DOTLOCK_LOCKING */
	status = 0;
	goto clean;
    }
#endif

    dprint(5,
	   (debugfile, "Lock %s for file %s on successfully.\n",
	    (direction == INCOMING ? "incoming" : "outgoing"), 
	    folder->cur_folder));

    folder->p->a.spool.lock_state = ON;
    status = 1;
    
 clean:
    dprint(11, (debugfile,
		"mbx_lock_spool=%d\n", status));

    return status;
}

static int mbx_unlock_no_name P_((int interrupt, struct folder_info *folder));

static int mbx_unlock_no_name(interrupt,folder) 
     int interrupt;
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_unlock_no_name: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_unlock_no_name=0\n"));
    return 0;
}

static int mbx_unlock_non_spool P_((int interrupt, struct folder_info *folder));
static int mbx_unlock_non_spool(interrupt,folder) 
     int interrupt;
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_unlock_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_unlock_non_spool=1 (dummy)\n"));
    return 1;
}

#ifdef SYSCALL_LOCKING

static int mbx_syscall_unlock_file P_((int interrupt, 
				       struct folder_info *folder));

static int mbx_syscall_unlock_file(interrupt,folder)
     int interrupt; 
     struct folder_info *folder;
{
    int status = 0;
    int retcode;

    if (0 != (retcode = 
	      Release_the_file (fileno(folder->p->fh_folder)))) { 
	dprint(1, (debugfile,
		   "Error %s\n\ttrying to unlock file %s (%s)\n", 
		   error_description(errno), 
		   folder->cur_folder, "unlock"));
    } else
	status = 1;

    return status;
}
#endif

#ifdef USE_DOTLOCK_LOCKING

static int mbx_dotunlock_file P_((int interrupt, 
				       struct folder_info *folder));

static int mbx_dotunlock_file(interrupt,folder)
     int interrupt; 
     struct folder_info *folder;
{
    int status = 0;
    int retcode;

#ifdef SAVE_GROUP_MAILBOX_ID
    gid_t oldgid = getegid();
    setgid(mailgroupid); /* reset id so that we can get at lock file */
#endif

    if((retcode = unlink(folder->p->a.spool.lockfile)) == 0) {	/* remove lock file */
	status = 1;
    } else {
	dprint(1, (debugfile,
		   "Error %s\n\ttrying to unlink file %s (%s)\n", 
		   error_description(errno), 
		   folder->p->a.spool.lockfile,
		   "unlock"));
	if (!interrupt)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmLeaveCouldntRemoveOwnLock,
			      "Couldn't remove my own lock file %s!"), 
		      folder->p->a.spool.lockfile);
    }

#ifdef SAVE_GROUP_MAILBOX_ID
    setgid(oldgid); 
#endif
    return status;
}
#endif


static int mbx_unlock_spool P_((int interrupt, struct folder_info *folder));

static int mbx_unlock_spool(interrupt,folder)
     int interrupt; 
     struct folder_info *folder;
{
    /** Remove the lock file!    This must be part of the interrupt
	processing routine to ensure that the lock file is NEVER
	left sitting in the mailhome directory!
	
	If also using flock(), remove the file lock as well.
    **/

    int status = 1;

    dprint(11, (debugfile,
		"mbx_unlock_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));

#ifndef USE_DOTLOCK_LOCKING
    dprint(5,
	   (debugfile, "Lock (no .lock) for file %s %s off.\n",
	    folder->cur_folder, 
	    (folder-> p->a.spool.lock_state == ON ? "going" : "already")));
#else   /* USE_DOTLOCK_LOCKING */
    dprint(5,
	   (debugfile, "Lock %s for file %s %s off.\n",
 	    (folder->p->a.spool.lockfile ? folder->p->a.spool.lockfile : "none"),
	    folder->cur_folder,
	   (folder->p->a.spool.lock_state == ON ? "going" : "already")));
#endif  /* USE_DOTLOCK_LOCKING */

    if(folder->p->a.spool.lock_state == ON) {
#ifdef SYSCALL_LOCKING
	if(!folder->p->fh_folder) {
	    dprint(11, (debugfile,
			"mbx_unlock_spool: non-existent -- no syscall locking.\n"));
	} else  if (!mbx_syscall_unlock_file(interrupt,folder))
	    status = 0;
#endif

#ifdef USE_DOTLOCK_LOCKING
	if (!mbx_dotunlock_file(interrupt,folder))
	    status = 0;    
#endif
    
	if (status)
	    folder->p->a.spool.lock_state = OFF; /* indicate we don't have a lock on */
    
    }
    dprint(11, (debugfile,
		"mbx_unlock_non_spool=%d\n",
		status));
    return status;
}

static void mk_temp_mail_fn P_((char *tempfn, char *mbox,
				int tempfn_size));


static void mbx_init_null P_((struct folder_info *folder));
static void mbx_init_null(folder) 
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_init_null: folder=%X (%s)\n",
		folder,folder->cur_folder));

    folder -> cur_tempfolder[0] = '\0';
    folder -> p = NULL;
}

static void mbx_init_non_spool P_((struct folder_info *folder));
static void mbx_init_non_spool(folder) 
     struct folder_info *folder;
{	
    dprint(11, (debugfile,
		"mbx_init_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));

    folder -> p = safe_malloc(sizeof (struct private_data));
    /* defined in hdrs/defs.h */
    bzero((void *)folder -> p,sizeof( struct private_data));   
    
    folder ->p->fh_temp     = NULL;
    folder ->p->fh_folder   = NULL;
}

static void mbx_init_spool P_((struct folder_info *folder));
static void mbx_init_spool(folder) 
     struct folder_info *folder;
{	
    dprint(11, (debugfile,
		"mbx_init_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));
    mk_temp_mail_fn(folder-> cur_tempfolder, 
		    folder->cur_folder, 
		    sizeof folder -> cur_tempfolder);

    folder -> p = safe_malloc(sizeof (struct private_data));
    /* defined in hdrs/defs.h */
    bzero((void *)folder -> p,sizeof (struct private_data));   
    folder ->p->fh_temp     = NULL;
    folder ->p->fh_folder   = NULL;

    folder -> p -> a.spool.lock_state = OFF;
#ifdef  USE_DOTLOCK_LOCKING
    folder -> p -> a.spool.lockfile=
	safe_strdup(mk_lockname(folder->cur_folder));
#endif  /* USE_DOTLOCK_LOCKING */

}

static void mk_temp_mail_fn(tempfn, mbox, tempfn_size)
     char *tempfn, *mbox;
     int tempfn_size;
{
    /** create in tempfn the name of the temp file corresponding to
	mailfile mbox. 
    **/
  
    char *cp,*ptr;
    int in_spool = is_in_spool(mbox);
    
    dprint(8,(debugfile,"mk_temp_mail_fn: in_spool=%d, mbox=%s\n",
	      in_spool,mbox));
	       
    if (strlen(default_temp) + strlen(temp_mbox) > tempfn_size-1) {
	dprint(1,(debugfile,"mk_temp_mail_fn: Too long path!\n"));
	strfcpy(tempfn,"TEMP_MBOX", tempfn_size);
	return;
    }

    elm_sfprintf(tempfn, tempfn_size,
		 FRM("%s%s"), default_temp, temp_mbox);
  
    if((cp = rindex(mbox, '/')) != NULL) {    
	cp++;
	if (strcmp(cp, "mbox") == 0 || strcmp(cp, "mailbox") == 0 ||
	    strcmp(cp, "inbox") == 0 || *cp == '.') {
	    ptr = username;
	}
	else {
	    ptr = cp;
	}
    } else {
	ptr = mbox;
    }
    
    if (strlen(tempfn) + strlen(ptr) > tempfn_size-1) {
	dprint(1,(debugfile,
		  "mk_temp_mail_fn: Too long path or mailbox!\n"));
    } else
	strfcat(tempfn, ptr, tempfn_size);

    if (!in_spool) {
	/* Assume that this is user's incoming mail area */
	if (strlen(tempfn) + strlen(username) > tempfn_size-2) {
	    dprint(1,(debugfile,
		      "mk_temp_mail_fn: Too long path or username!\n"));
	} else {
	    strfcat(tempfn, "-", tempfn_size);
	    strfcat(tempfn, username, tempfn_size);
	}
    }
    dprint(8,(debugfile,"mk_temp_mail_fn: tempfname=%s\n",tempfn));
}
 
static void mbx_free_null P_((struct folder_info *folder));
static void mbx_free_null(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_free_null: folder=%X (%s)\n",
		folder,folder->cur_folder));    
}

static void mbx_free_non_spool P_((struct folder_info *folder));
static void mbx_free_non_spool(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_free_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));    
    
    if (folder -> p -> fh_temp) {
	fclose( folder -> p -> fh_temp);
	folder -> p -> fh_temp = NULL;
    }
    if (folder -> p -> fh_folder) {
	fclose(folder -> p -> fh_folder);	
	folder -> p -> fh_folder = NULL;
    }
    free(folder -> p);
    folder -> p = NULL;
}


static void mbx_free_spool P_((struct folder_info *folder));
static void mbx_free_spool(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_free_non_spool: folder=%X (%s)\n",
		folder,folder->cur_folder));    

    if (folder -> p->fh_temp) {
	fclose(folder -> p->fh_temp);
	folder -> p -> fh_temp = NULL;
    }
    if (folder -> p -> fh_folder) {
	fclose(folder -> p -> fh_folder);	
	folder -> p -> fh_folder = NULL;
    }
#ifdef	USE_DOTLOCK_LOCKING
    if (folder -> p -> a.spool.lockfile) {
	free(folder -> p -> a.spool.lockfile);
	folder -> p -> a.spool.lockfile = NULL;
    }
#endif
    free(folder -> p);
    folder -> p = NULL;
}


static int mbx_sessionlock_no_name P_((struct folder_info *folder,
				       enum sessionlock_mode mode));

static int mbx_sessionlock_no_name(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    dprint(11, (debugfile,
		"mbx_sessionlock_no_name: folder=%X (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_sessionlock_no_name=0\n"));
    return 0;
}

static int mbx_sessionlock_filehandle P_((struct folder_info *folder,
					  enum sessionlock_mode mode));

static int mbx_sessionlock_filehandle(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    int status = 0;

    if (mode == SESSIONLOCK_CHECK) {
	int code = 0;
	if ((code = can_open(folder->cur_folder, 
			      "r+")) != 0 &&
	    code != ENOENT) {
	    dprint(1, (debugfile,
		       "Error: given file %s as folder - unreadable (%s)!\n", 
		       folder->cur_folder, 
		       error_description(code)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCantOpenFolderRead,
			      "Can't open folder '%s' for reading!"), 
		      folder->cur_folder);
	    status = 0;
	    goto clean;
	}
	dprint(12, (debugfile,
		    "mbx_sessionlock_filehandle: %s SESSIONLOCK_CHECK OK %s\n",
		    folder->cur_folder,
		    code ? error_description(code) : ""));
    }

    if (mode == SESSIONLOCK_REOPEN && folder->p->fh_folder) {
	fclose(folder->p->fh_folder);
	folder->p->fh_folder = NULL;
    }

    status = 1;

 clean:
    return status;
}     

static int mbx_sessionlock_read_only P_((struct folder_info *folder,
				       enum sessionlock_mode mode));

static int mbx_sessionlock_read_only(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    int status = 0;

    dprint(11, (debugfile,
		"mbx_sessionlock_read_only: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (mode == SESSIONLOCK_CHECK) {
	if ((errno = can_open(folder->cur_folder, 
			      "r")) != 0) {
	    dprint(1, (debugfile,
		       "Error: given file %s as folder - unreadable (%s)!\n", 
		       folder->cur_folder, 
		       error_description(errno)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCantOpenFolderRead,
			      "Can't open folder '%s' for reading!"), 
		      folder->cur_folder);
	    status = 0;
	    goto clean;
	}
    }
    
    if (mode == SESSIONLOCK_REOPEN && folder->p->fh_folder) {
	fclose(folder->p->fh_folder);
	folder->p->fh_folder = NULL;
    }

    if (mode == SESSIONLOCK_TRUNCATE) {
	dprint(10, (debugfile, 
		   "mbx_sessionlock_read_only: mode == SESSIONLOCK_TRUNCATE (ignored) : %s\n",
		   folder->cur_folder));
    }

    if (!folder->p->fh_folder) {	
	folder->p->fh_folder = fopen(folder->cur_folder,"r");
	
	if (!folder->p->fh_folder) {
	    int err = errno;
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailOnOpenNewmbox,
			      "\nfail on open in newmbox, open %s failed!!\n"),
		      folder->cur_folder);	    
	    lib_error(FRM("** %s. **\n"), 
		      error_description(err));
	    dprint(1, (debugfile, "fail on open in mbx_sessionlock_non_spool, file %s!!\n",
		       folder->cur_folder));
	    status = 0;
	    goto clean;
	}
	dprint(10, (debugfile, "Mailfile (folder) opened: %s\n",
		    folder->cur_folder));
    } else
	rewind(folder->p->fh_folder);

    if (mode == SESSIONLOCK_NONE) {
	status = 1;
	goto clean;
    }

    if (lockfolders) {
	switch (GrabRead_the_file(fileno(folder -> p->fh_folder))) {
	case FLOCKING_OK:
	    dprint(1, (debugfile, 
		       "Folder locked (%s).\n",
		       folder -> cur_folder));
	    break;
	case FLOCKING_RETRY:
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRW,
			      "You seem to have ELM already reading (read-write) this mail!"));	
	    status = 0;
	    goto clean;	    

	case	FLOCKING_FAIL:
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorLockFolder,
			      "Error encountered while attempting to lock folder %s"), 
		      folder->cur_folder);	    
	    lib_error(FRM("** %s. **"), 
		      error_description(errno));

	    status = 0;
	    goto clean;
	}
    }

    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_sessionlock_read_only=%d\n",status));
    return status;
}

static int mbx_sessionlock_non_spool P_((struct folder_info *folder,
				       enum sessionlock_mode mode));

static int mbx_sessionlock_non_spool(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    int status = 0;

    dprint(11, (debugfile,
		"mbx_sessionlock_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!mbx_sessionlock_filehandle(folder,mode)) {
	status = 0;
	goto clean;
    }

    if (mode == SESSIONLOCK_TRUNCATE) {
	dprint(10, (debugfile, 
		   "mbx_sessionlock_non_spool: mode == SESSIONLOCK_TRUNCATE (ignored) : %s\n",
		   folder->cur_folder));
    }

    if (!folder->p->fh_folder) {	
	folder->p->fh_folder = fopen(folder->cur_folder,"r+");

	if (!folder->p->fh_folder) {
	    int err = errno;
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailOnOpenNewmbox,
			      "\nfail on open in newmbox, open %s failed!!\n"),
		      folder->cur_folder);	    
	    lib_error(FRM("** %s. **\n"), 
		      error_description(err));
	    dprint(1, (debugfile, "fail on open in mbx_sessionlock_non_spool, file %s!!\n",
		       folder->cur_folder));
	    status = 0;
	    goto clean;
	}
	dprint(10, (debugfile, "Mailfile (folder) opened: %s\n",
		    folder->cur_folder));
    } else
	rewind(folder->p->fh_folder);

    if (mode == SESSIONLOCK_NONE) {
	status = 1;
	goto clean;
    }

    if (lockfolders) {
	switch (Grab_the_file(fileno(folder -> p->fh_folder))) {
	case FLOCKING_OK:
	    dprint(1, (debugfile, 
		       "Folder locked (%s).\n",
		       folder -> cur_folder));
	    break;
	case FLOCKING_RETRY:
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning1,
			      "You seem to have ELM already reading this mail!"));	
	    status = 0;
	    goto clean;	    

	case	FLOCKING_FAIL:
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorLockFolder,
			      "Error encountered while attempting to lock folder %s"), 
		      folder->cur_folder);	    
	    lib_error(FRM("** %s. **"), 
		      error_description(errno));

	    status = 0;
	    goto clean;
	}
    }

    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_sessionlock_non_spool=%d\n",status));
    return status;
}


static int mbx_sessionlock_spool P_((struct folder_info *folder,
				       enum sessionlock_mode mode));

static int mbx_sessionlock_spool(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    int status = 0;

    dprint(11, (debugfile,
		"mbx_sessionlock_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!mbx_sessionlock_filehandle(folder,mode)) {
	status = 0;
	goto clean;
    }

    if (!folder->p->fh_folder) {
	folder->p->fh_folder = fopen(folder->cur_folder,"r+");
	if (!folder->p->fh_folder) {
	    if (errno != ENOENT) {
		int err = errno;
	    
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailOnOpenNewmbox,
				  "\nfail on open in newmbox, open %s failed!!\n"),
			  folder->cur_folder);		
		lib_error(FRM("** %s. **\n"), 
			  error_description(err));
		dprint(1, (debugfile, "fail on open in mbx_sessionlock_non_spool, file %s!!\n",
			   folder->cur_folder));
		status = 0;
		goto clean;
	    } else {
		folder->mailfile_size = 0;    
		/* must non-existant folder */
	    }
	}
	dprint(10, (debugfile, "Mailfile (folder) opened: %s %s\n",
		    folder->cur_folder,
		    folder->p->fh_folder ? "" : " (non existant)"));
    } else {
	rewind(folder->p->fh_folder);
    }

    switch(mode) {
	int need_reopen;
    case SESSIONLOCK_NONE:
	status = 1;
	goto clean;
    case SESSIONLOCK_TRUNCATE:
	need_reopen = 1;
#ifdef FTRUNCATE  
	if (folder->p->fh_temp) {
	    rewind(folder->p->fh_temp);
	    if (0 == ftruncate(fileno(folder->p->fh_temp),0)) {
		need_reopen = 0;
		dprint(10, (debugfile,
			    "truncate_tempfolder: tempfile %s truncated (not recreated)\n",
			    folder->cur_tempfolder));
	    }
	    else 
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmSorryCantTruncateTemp,
				  "Sorry, can't truncate the temp file %s [%s]!"),
			  folder -> cur_tempfolder, error_description(errno));
	    /* IF FAIL REOPEN IT INSTEAD */
	}
#endif
	if (!need_reopen)
	    break;  /* DONE */
	if (folder->p->fh_temp) {
	    fclose (folder->p->fh_temp);
	    folder->p->fh_temp = NULL;
	}

	if (0 != unlink(folder -> cur_tempfolder)) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantTruncateTemp,
			      "Sorry, can't truncate the temp file %s [%s]!"),
		      folder -> cur_tempfolder, error_description(err));    
	    status = 0;
	    goto clean;
	}
	goto create_it;
    case   SESSIONLOCK_REOPEN:
	if (folder -> p->fh_temp) {
	    int temp_handle;
	    fclose(folder -> p->fh_temp);
	    folder -> p->fh_temp = NULL;

	    if ( (temp_handle = open(folder-> cur_tempfolder, 
				     O_RDWR,
				     0600)) == -1 ||
		 NULL == (folder -> p->fh_temp = fdopen(temp_handle,"w+"))) {
		int err = errno;
		
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantReopenTemp,
				  "Sorry, can't reopen the temp file %s [%s]!"),
			  folder -> cur_tempfolder, error_description(err));  
		
		if (-1 != temp_handle)
		    close(temp_handle);
  
		status = 0;
		goto clean;
	    }

	    break;
	}
	/* FALLTHRU */
    case SESSIONLOCK_NORMAL:
    case SESSIONLOCK_CHECK:
    create_it:
	if (!folder -> p->fh_temp) {
	    int temp_handle;

	    /* If we are changing files and we are changing to a spool file,
	     * make sure there isn't a temp file for it, because if
	     * there is, someone else is using ELM to read the new file,
	     * and we don't want to be reading it at the same time.
	     */
    
	    if ( (temp_handle = open(folder-> cur_tempfolder, 
				     O_RDWR|O_CREAT|O_EXCL,
				     0600)) == -1) {
		int err = errno;
		
		if (err == EEXIST) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning1,
				      "You seem to have ELM already reading this mail!"));		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning2,
				      "You may not have two copies of ELM running simultaneously.  -- Exiting --"));		   
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAlreadyRunning3,
				      "If this is in error, then you'll need to remove the following file:"));		  
		    lib_error(FRM("%s"),folder -> cur_tempfolder);
		    
		    status = 0;
		    goto clean;	    
		}
	    }
	    if (-1 == temp_handle || 
		NULL == (folder -> p->fh_temp = fdopen(temp_handle,"w+"))) {
		int err = errno;
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedCreateTmpFolder,
				  "Failed to create %.50s: %.60s"),
			  folder -> cur_tempfolder,
			  error_description(err));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIGiveUp,
				  "Ahhhh... I give up."));
		
		if (-1 != temp_handle)
		    close(temp_handle);
		
		status = 0;
		goto clean;	    
	    }
	    
	    dprint(1, (debugfile, 
		       "Creating tempfile (%s).\n",
		       folder -> cur_tempfolder));
	    
	    elm_chown(folder->cur_tempfolder, userid, groupid);
	    chmod(folder -> cur_tempfolder, 0700);	
	    /* shut off file for other people! */	    
	}
	break;
    }
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_sessionlock_spool=%d\n",status));
    return status;
}

static void mbx_flush_null P_((struct folder_info *folder));
static void mbx_flush_null(folder) 
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_flush_null: folder=%p (%s)\n",
		folder,folder->cur_folder));
}

void mbx_flush_temp(folder) 
     struct folder_info *folder;
{
    dprint(12, (debugfile,
		"mbx_flush_temp: folder=%p (%s)\n",
		folder,folder->cur_folder));
    
    if (!folder->p->fh_temp)
	return;
    
#ifdef ultrix
    /** Ultrix returns an error if fflush is called on a file opened
	for read only, so we have to account for this.
    **/
    
    if (folder->p->fh_temp->_flag & (_IOREAD|_IORW) == _IOREAD)
	return;
#endif
    fflush(folder->p->fh_temp);
    return;    
}

static void mbx_flush_non_spool P_((struct folder_info *folder));
static void mbx_flush_non_spool(folder) 
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_flush_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!folder->p->fh_folder)
	return;

#ifdef ultrix
	/** Ultrix returns an error if fflush is called on a file opened
	    for read only, so we have to account for this.
	**/

	if (folder->p->fh_folder->_flag & (_IOREAD|_IORW) == _IOREAD)
	    return;
#endif
	fflush(folder->p->fh_folder);
	return;

}

static void mbx_flush_spool P_((struct folder_info *folder));
static void mbx_flush_spool(folder) 
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_flush_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    mbx_flush_temp(folder);
    mbx_flush_non_spool(folder);    
}

static int mbx_ferror_no_name P_((struct folder_info *folder, int clean));
static int mbx_ferror_no_name(folder,clean) 
     struct folder_info *folder; 
     int clean; 
{
    dprint(11, (debugfile,
		"mbx_ferror_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_ferror_no_name=1\n"));
    return 1;
}

static int mbx_ferror_non_spool P_((struct folder_info *folder, int clean));
static int mbx_ferror_non_spool(folder,clean) 
     struct folder_info *folder; 
     int clean; 
{
    int status;
    dprint(11, (debugfile,
		"mbx_ferror_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    status = ferror(folder->p->fh_folder);
    if (clean)
	clearerr(folder->p->fh_folder);

    dprint(11, (debugfile,
		"mbx_ferror_non_spool=%d\n",status));
    return status;
}


static int mbx_ferror_spool P_((struct folder_info *folder, int clean));
static int mbx_ferror_spool(folder,clean) 
     struct folder_info *folder; 
     int clean; 
{
    int status;
    dprint(11, (debugfile,
		"mbx_ferror_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if(!folder->p->fh_folder) {
	dprint(11, (debugfile,
		    "mbx_ferror_spool: non-existent -- no error.\n"));
	status = 0;
    } else {
	status = ferror(folder->p->fh_folder);
	if (clean)
	    clearerr(folder->p->fh_folder);
    }

    if (folder->p->fh_temp) {
	if (ferror(folder->p->fh_temp))
	    status = 1;
	if (clean)
	    clearerr(folder->p->fh_temp);
    }

    dprint(11, (debugfile,
		"mbx_ferror_spool=%d\n",status));
    return status;
}

static void malloc_read_folder_state P_((struct read_folder_state **ptr));
static void malloc_read_folder_state(ptr)
     struct read_folder_state **ptr;
{
    *ptr = safe_malloc(sizeof (struct read_folder_state));

    dprint(11, (debugfile,
		"malloc_read_folder_state: (*ptr)=%p\n",
		*ptr));

    (*ptr) -> fbytes        = 0;
    /* (*ptr) -> count         = 0; */
    (*ptr) -> linecounter   = 0;
}

static void mbx_zero_rs_fields_file P_((struct read_folder_state *rs));
static void mbx_zero_rs_fields_file(rs)
     struct read_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_zero_rs_fields_file: rs=%p\n",
		rs));

    rs -> a.file.next_line     = NULL;
    rs -> a.file.next_line_len = 0;    
}

static void free_read_folder_state P_((struct read_folder_state **ptr));
static void free_read_folder_state(ptr)
     struct read_folder_state **ptr;
{
    dprint(11, (debugfile,
		"free_read_folder_state: (*ptr)=%p\n",
		*ptr));

    free(*ptr);
    *ptr = NULL;
}

static void mbx_free_rs_fields_file P_((struct read_folder_state *rs));
static void mbx_free_rs_fields_file(rs)
     struct read_folder_state *rs;
{
    dprint(11, (debugfile,
		"mbx_free_rs_fields_file: rs=%p\n",
		rs));

    if (rs -> a.file.next_line) {
	free(rs -> a.file.next_line);
	rs -> a.file.next_line = NULL;
    }
    rs -> a.file.next_line_len = 0;        
}


static int mbx_prepare_read_no_name P_((struct folder_info *folder,
				       enum prepare_mode mode,
				       READ_STATE read_state_ptr));
static int mbx_prepare_read_no_name(folder,mode,read_state_ptr)
     struct folder_info *folder;
     enum prepare_mode mode;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_prepare_read_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                        : read_state_ptr=%p\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_prepare_read_no_name=0\n"));
    return 0;
}

static int mbx_prepare_seek_folder P_((struct folder_info *folder,
				       int add_new_only,
				       READ_STATE read_state_ptr));
static int mbx_prepare_seek_folder(folder,add_new_only,read_state_ptr)
     struct folder_info *folder;
     int add_new_only;
     READ_STATE read_state_ptr;
{
    int status = 0;
    long result;

    if (add_new_only) {
	if (fseek(folder -> p->fh_folder, 
		  folder -> mailfile_size, SEEK_SET) == -1) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntSeekEndFolder,
				"\nCouldn't seek to %ld (end of folder) in %s!\n"),
			folder-> mailfile_size, 
			folder-> cur_folder);		   
	    lib_error("** %s. **\n", error_description(err));
	    dprint(1, (debugfile,
		       "Error: Couldn't seek to end of folder %s: (offset %ld) Errno %s (%s)\n",
		       folder -> cur_folder, 
		       folder -> mailfile_size, 
		       error_description(err), "mbx_prepare_seek_folder"));
	    status = 0;
	    goto clean;
	}
	/* read_state_ptr->count  = message_count; */	  /* next available  */
	read_state_ptr->fbytes = folder-> mailfile_size;  /* start correctly */
    }

    result = file_bytes(folder->cur_folder);
    if (-1L == result) {
	status = 0;
	goto clean;
    }
    folder -> mailfile_size = result;

    status = 1;

 clean:
    return status;
}

static int mbx_prepare_read_non_spool P_((struct folder_info *folder,
					  enum prepare_mode mode,
					  READ_STATE read_state_ptr));
static int mbx_prepare_read_non_spool(folder,mode,read_state_ptr)
     struct folder_info *folder;
     enum prepare_mode mode;
     READ_STATE read_state_ptr;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_prepare_read_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                          : read_state_ptr=%p\n",
		read_state_ptr));

    if (!mbx_prepare_seek_folder(folder, 
				 PREPARE_NEW_ONLY        == mode ||
				 PREPARE_NEW_ONLY_NOLOCK == mode,
				 read_state_ptr)) {
	status = 0;
	goto clean;
    }
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_prepare_read_non_spool=%d\n",status));
    return status;
}

static int mbx_prepare_read_spool P_((struct folder_info *folder,
				      enum prepare_mode mode,
				      READ_STATE read_state_ptr));
static int mbx_prepare_read_spool(folder,mode,read_state_ptr)
     struct folder_info *folder;
     enum prepare_mode mode;
     READ_STATE read_state_ptr;
{
    int status = 0;
    int add_new_only =  PREPARE_NEW_ONLY == mode ||
	                PREPARE_NEW_ONLY_NOLOCK == mode;

    dprint(11, (debugfile,
		"mbx_prepare_read_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                      : read_state_ptr=%p\n",
		read_state_ptr));

    if (PREPARE_NOLOCK == mode || 
	PREPARE_NEW_ONLY_NOLOCK == mode) {
	dprint(11, (debugfile, 
		    "mbx_prepare_read_spool: mode == PREPARE_NOLOCK\n"));
    } else if (folder -> p->a.spool.lock_state == ON) {
	dprint(1, (debugfile, 
		   "mbx_prepare_read_spool: %s already locked.\n",
		   folder->cur_folder));
    } else {
	/* ensure no mail arrives while we do this! */
	if (!lock_folder(INCOMING, folder)) {
	    status = 0;
	    goto clean;
	}
    }

    if (add_new_only) {
	if (!folder->p->fh_temp) {
	    dprint(11, (debugfile,
			"mbx_prepare_read_spool -- NO temp file %s\n",
			folder->cur_tempfolder));
	} else if (fseek(folder -> p->fh_temp, 
		  folder -> mailfile_size, SEEK_SET) == -1) {
	    int err = errno;
	    
	    unlock_folder(0,folder);	/* remove lock file! */
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntSeekTempEnd,
			      "\nCouldn't fseek to end of temp mbox.\n"));
	    lib_error(FRM("** %s. **\n"), error_description(err));
	    dprint(1, (debugfile, 
		       "Error: Couldn't fseek to end of reopened temp mbox.  errno %s (%s)\n",
		       error_description(err), "mbx_prepare_read_spool"));
	    status = 0;
	    goto clean;	    
	}
    }

    if (!mbx_prepare_seek_folder(folder,add_new_only,read_state_ptr)) {
	status = 0;
	goto clean;
    }
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_prepare_read_spool=%d\n",status));
    return status;
}


static int mbx_end_read_null P_((struct folder_info *folder,
				 READ_STATE read_state_ptr,
				 int silent));

static int mbx_end_read_null(folder,read_state_ptr,silent)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     int silent;
{
    dprint(11, (debugfile,
		"mbx_end_read_null: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                 : read_state_ptr=%p\n",
		read_state_ptr));
    dprint(11, (debugfile,
		"mbx_end_read_null=1\n"));
    return 1;
}

static int mbx_end_read_non_spool P_((struct folder_info *folder,
				      READ_STATE read_state_ptr,
				      int silent));

static int mbx_end_read_non_spool(folder,read_state_ptr,silent)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     int silent;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_end_read_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                  : read_state_ptr=%p\n",
		read_state_ptr));

    rewind(folder->p->fh_folder);
    status = 1;

    dprint(11, (debugfile,
		"mbx_end_read_non_spool=%d\n",status));
    return status;
}


static int mbx_end_read_spool P_((struct folder_info *folder,
				  READ_STATE read_state_ptr,
				  int silent));

static int mbx_end_read_spool(folder,read_state_ptr,silent)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     int silent;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_end_read_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                  : read_state_ptr=%p\n",
		read_state_ptr));

    if (PREPARE_NOLOCK == read_state_ptr->mode ||
	PREPARE_NEW_ONLY_NOLOCK == read_state_ptr->mode) {
	dprint(11, (debugfile,
		    "mbx_end_read_spool: PREPARE_NOLOCK == mode\n"));
    } else 
	unlock_folder(0, folder);

    if (!folder->p->fh_temp) {
	dprint(11, (debugfile,
		    "mbx_end_read_spool=1 -- NO temp file %s\n",
		    folder->cur_tempfolder));
	return 1;
    }

    if ((ferror(folder->p->fh_temp)) || 
	(fflush(folder->p->fh_temp) == EOF)) {
	if (!silent) {
	    int err = errno;
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFlushOnTempFailed,
			      "Flush on temp file %s failed: %s"), 
		      folder -> cur_tempfolder,
		      error_description(err));
	}
	dprint(1, (debugfile, "Can't flush on temp file %s!!\n",
		   folder -> cur_tempfolder));
	status = 0;
	goto clean;
    }
	 
    /* sanity check on append - is resulting temp file longer??? */
    if ( file_bytes(folder-> cur_tempfolder) != 
	 folder-> mailfile_size) {
	if (!silent) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLengthNESpool,
			      "\nnewmbox - length of mbox. != spool mailbox length!!\n"));
	}
	dprint(1, (debugfile, "newmbox - mbox. != spool mail length"));
	
    }
	  
    rewind(folder->p->fh_temp);
    status = 1;
    
 clean:
    dprint(11, (debugfile,
		"mbx_end_read_spool=%d\n",status));
    return status;
}

static int mbx_copy_envelope_no_name P_((struct folder_info *folder,
					 READ_STATE read_state_ptr,
					 struct header_rec *entry));
static int mbx_copy_envelope_no_name(folder,read_state_ptr,entry)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     struct header_rec *entry;
{
    dprint(11, (debugfile,
		"mbx_copy_envelope_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                         : read_state_ptr=%p\n",
		read_state_ptr));

    entry->offset = 0;
    entry->status = VISIBLE;

    dprint(11, (debugfile,
		"mbx_copy_envelope_no_name=0\n"));
    return 0;
}

static int read_line P_((FILE *fh,char **buffer, int *len, int max_read));
static int read_line(fh,buffer,len,max_read)
     FILE *fh;
     char **buffer; 
     int *len; 
     int max_read;
{
    char buffer1[VERY_LONG_STRING];
    int limit = sizeof buffer1;
    int line_bytes;
    
    if (limit > max_read) {
	limit = max_read+1;
	dprint(95, (debugfile,
		    "read_line: limit = %d, max_read = %d\n",limit,max_read));
    }
    if (limit < 1)
	return 0;

    line_bytes = mail_gets(buffer1, limit, fh);
    if (0 == line_bytes) {
	dprint(95, (debugfile,
		    "read_line=0: line_bytes=0\n"));
	return 0;
    }
    *buffer = safe_malloc(line_bytes+1);
    
    /* bcopy is defined on hdrs/defs.h */
    bcopy(buffer1,(*buffer),line_bytes); 
    (*buffer)[line_bytes]='\0';
    *len = line_bytes;

    while (line_bytes > 0 && buffer1[line_bytes - 1] != '\n') {
	limit = sizeof buffer1;
	
	if (limit + *len > max_read)
	    limit = max_read - *len;
	
	if (limit < 1)
	    break;
	line_bytes = mail_gets(buffer1, sizeof buffer1, fh);
	if (0 == line_bytes)
	    break;
	*buffer = safe_realloc(*buffer, *len + line_bytes+1);
	
	/* bcopy is defined on hdrs/defs.h */
	bcopy(buffer1,(*buffer)+(*len),line_bytes); 
	(*buffer)[(*len)+line_bytes]='\0';
	*len += line_bytes;
    }
    dprint(95, (debugfile,
		"read_line=1: len=%d, buffer=%.*s",
		*len,*len,(*buffer) ? (*buffer) : ""));
    if (*len < 1 || !(*buffer) || (*buffer)[*len -1] != '\n') {
	dprint(95, (debugfile,
		    "\nread_line: NO NEWLINE\n"));
    } 
    return 1;
}

static int mbx_read_buffered_line P_((struct folder_info *folder,
				      READ_STATE read_state_ptr,
				      char **buffer, int *len));
static int mbx_read_buffered_line(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    if(read_state_ptr ->a.file.next_line) {
	*buffer = read_state_ptr -> a.file.next_line;
	*len    = read_state_ptr -> a.file.next_line_len;
    } else {
	
	if (!read_line(folder->p->fh_folder,buffer,len,
		       folder->mailfile_size-read_state_ptr->fbytes))
	    return 0;

	read_state_ptr -> a.file.next_line      = *buffer;
	read_state_ptr -> a.file.next_line_len  = *len;
	
    }
    
    dprint(90, (debugfile,
		"mbx_read_buffered_line: len=%d, buffer=%.*s",
		*len,*len,(*buffer) ? (*buffer) : ""));
    if (*len < 1 || !(*buffer) || (*buffer)[*len -1] != '\n') {
	dprint(90, (debugfile,
		    "\nmbx_read_buffered_line: NO NEWLINE\n"));
    } 

    return 1;
}

static void mbx_accept_buffered_line P_((struct folder_info *folder,
					 READ_STATE read_state_ptr,
					 char **buffer, int *len));
static void mbx_accept_buffered_line(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    if (*buffer != read_state_ptr -> a.file.next_line)
	abort();
    if (!*buffer)
	return;

    if (*len < 1 || (*buffer)[*len -1] != '\n') {
    } else {
	read_state_ptr -> linecounter++;
    }
    read_state_ptr -> a.file.next_line     = NULL;
    read_state_ptr -> fbytes       += read_state_ptr -> a.file.next_line_len;
    read_state_ptr -> a.file.next_line_len = 0;

    free(*buffer);
    *buffer = NULL;
    *len    = 0;
}

#ifdef MMDF
static int mbx_copy_envelope_check_MMDF P_((char *buffer,int len));
static int mbx_copy_envelope_check_MMDF(buffer,len)
     char *buffer;
     int len; 
{
    if (strcmp(buffer, MSG_SEPARATOR) == 0)
	return 1;
    return 0;
}
#endif

static int mbx_copy_envelope_non_spool P_((struct folder_info *folder,
					   READ_STATE read_state_ptr,
					   struct header_rec *entry));
static int mbx_copy_envelope_non_spool(folder,read_state_ptr,entry)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     struct header_rec *entry;
{
    char * buffer = NULL;
    int  len;

    /* NO COPY */
    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                           : read_state_ptr=%p\n",
		read_state_ptr));

    read_state_ptr -> linecounter   = 0;      /* Linecounter of current
					         message */
    entry->offset = read_state_ptr -> fbytes; /* Offset of current message */
    entry->status = VISIBLE | UNREAD;         /* Default status of message */

    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer,&len)) {
	status = 0;
	goto clean;
    }
    if (0 == len) {
	status = 0;
	goto clean;
    }

#ifdef MMDF    
    if (mbx_copy_envelope_check_MMDF(buffer,len)) {
	mbx_accept_buffered_line(folder,read_state_ptr,
				 &buffer,&len);
	if (!mbx_read_buffered_line(folder,read_state_ptr,
					 &buffer,&len)) {
	    status = 0;
	    goto clean;
	}	
	status = 1;
    } else {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFolderCorrupt,
			  "Folder is corrupt!!  I can't read it!!"));
	status = -1;
	goto clean;
    }
#endif

    if (real_from(buffer,entry)) {
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer,&len);
	status = 1;

	while (mbx_read_buffered_line(folder,read_state_ptr,
					   &buffer,&len) &&
	       first_word_nc(buffer,">From ")) {
	    forwarded(buffer+6, entry); /* return address */
	    mbx_accept_buffered_line(folder,read_state_ptr,
					  &buffer,&len);	    
	}

    } else if (0 == status) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFolderCorrupt,
			  "Folder is corrupt!!  I can't read it!!"));
	status = -1;
    }

 clean:

    dprint(11, (debugfile,
		"mbx_copy_envelope_non_spool=%d\n",
		status));
    return status;
}

int mbx_copy_line_to_temp(folder,buffer,len)
     struct folder_info *folder;
     char *buffer; 
     int len;
{
    if (!folder->p->fh_temp) {
	dprint(15, (debugfile,
		    "mbx_copy_line_to_temp=1 -- NO temp file %s\n",
		    folder->cur_tempfolder));
	return 1;
    }

    if (fwrite(buffer, 1, len, folder->p->fh_temp) != len) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmWriteToTempFailed,
			  "\nWrite to temp file %s failed!!\n"),
		  folder->cur_tempfolder);       
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile, "Can't write to temp file %s!!\n",
		   folder->cur_tempfolder));
	return 0;
    }
    dprint(15, (debugfile,
		"mbx_copy_line_to_temp=1 -> %.*s",
		len,buffer));
    if (len < 1 || buffer[len-1] != '\n') {
	dprint(15, (debugfile,
		    "\nmbx_copy_line_to_temp <-- NO NEWLINE\n"));
    }

    return 1;
}


static int mbx_copy_envelope_spool P_((struct folder_info *folder,
				       READ_STATE read_state_ptr,
				       struct header_rec *entry));
static int mbx_copy_envelope_spool(folder,read_state_ptr,entry)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     struct header_rec *entry;
{
    char * buffer = NULL;
    int  len;

    /* COPY */
    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                       : read_state_ptr=%p\n",
		read_state_ptr));

    read_state_ptr -> linecounter   = 0;   /* Linecounter of current message */
    entry->status  = VISIBLE | NEW | UNREAD; /* Default status of message    */

    
    if (!folder->p->fh_folder) {
	dprint(11, (debugfile,
		    "                      : folder not open...\n"));
	status = 0;
	goto clean;
    } 

    do {
	entry->offset = read_state_ptr -> fbytes; 
	/* Offset of current message */

	if (!mbx_read_buffered_line(folder,read_state_ptr,
					 &buffer,&len)) {
	    status = 0;
	    goto clean;
	}
	if (0 == len) {
	    status = 0;
	    goto clean;
	}
		
	if ((len == 1 && 0 == memcmp(buffer,"\n",len)) ||
	    (len == 2 && 0 == memcmp(buffer,"\r\n",len))) {
	    if (!mbx_copy_line_to_temp(folder,buffer,len)) {
		status = 0;
		goto clean;
	    }

	    mbx_accept_buffered_line(folder,read_state_ptr,
					  &buffer,&len);
	    continue;
	}
    } while (0); 

#ifdef MMDF
    if (mbx_copy_envelope_check_MMDF(buffer,len)) {
	if (!mbx_copy_line_to_temp(folder,buffer,len)) {
	    status = 0;
	    goto clean;
	}
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer,&len);
	if (!mbx_read_buffered_line(folder,read_state_ptr,
					 &buffer,&len)) {
	    status = 0;
	    goto clean;
	}	
	status = 1;
    } else {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFolderCorrupt,
			  "Folder is corrupt!!  I can't read it!!"));
	status = -1;
	goto clean;
    }
#endif

    if (real_from(buffer,entry)) {
	if (!mbx_copy_line_to_temp(folder,buffer,len)) {
	    status = 0;
	    goto clean;
	}

	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer,&len);
	status = 1;

	while (mbx_read_buffered_line(folder,read_state_ptr,
					   &buffer,&len) &&
	       first_word_nc(buffer,">From ")) {
	    forwarded(buffer+6, entry); /* return address */

	    if (!mbx_copy_line_to_temp(folder,buffer,len)) {
		status = 0;
		goto clean;
	    }
	    mbx_accept_buffered_line(folder,read_state_ptr,
					  &buffer,&len);	    
	}

    } else if (0 == status) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFolderCorrupt,
			  "Folder is corrupt!!  I can't read it!!"));
	status = -1;
    }


 clean:
    dprint(11, (debugfile,
		"mbx_copy_envelope_spool=%d\n",
		status));
    return status;
}


static CONST char * mbx_is_forwarded_null P_((struct folder_info *folder,
					      READ_STATE read_state_ptr));
static CONST char * mbx_is_forwarded_null(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_is_forwarded_null: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                     : read_state_ptr=%p\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_is_forwarded_null=NULL\n"));
    return NULL;
}

static CONST char * mbx_is_forwarded_spool P_((struct folder_info *folder,
					      READ_STATE read_state_ptr));
static CONST char * mbx_is_forwarded_spool(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    char * buffer = NULL;
    CONST char * status = NULL; 
    int  len;

    dprint(11, (debugfile,
		"mbx_is_forwarded_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                      : read_state_ptr=%p\n",
		read_state_ptr));
    

    if (!folder->p->fh_folder) {
	dprint(11, (debugfile,
		    "                      : folder not open...\n"));
    } else {
	if (!mbx_read_buffered_line(folder,read_state_ptr,
				    &buffer,&len)) {
	    status = NULL;
	    goto clean;
	}

	if (len > 11 && first_word_nc(buffer, "forward to "))
	    status = buffer + 11;
    }

 clean:

    dprint(11, (debugfile,
		"mbx_is_forwarded_spool=%s\n",
		status ? status : "NULL"));
    return status;
}

static int mbx_copy_header_no_name P_((struct folder_info *folder,
				       READ_STATE read_state_ptr,
				       char **buffer, int *len));
static int mbx_copy_header_no_name(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    dprint(11, (debugfile,
		"mbx_copy_header_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                       : read_state_ptr=%p\n",
		read_state_ptr));
    
    dprint(11, (debugfile,
		"mbx_copy_header_no_name=0\n"));
    return 0;
}

void append_buffer(buffer,len,buffer1,len1)
     char **buffer; 
     int *len;
     char *buffer1; 
     int len1;
{
    int res = *len + len1 + 1;

    *buffer = safe_realloc(*buffer,res);

    /* bcopy is defined on hdrs/defs.h */
    bcopy(buffer1,(*buffer)+(*len),len1); 
    (*buffer)[(*len) + len1] = '\0';
    (*len) += len1;
}

static int mbx_copy_header_non_spool P_((struct folder_info *folder,
				       READ_STATE read_state_ptr,
				       char **buffer, int *len));
static int mbx_copy_header_non_spool(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    char *buffer1 = NULL;
    int len1;
    int status = 0;

    dprint(11, (debugfile,
		"mbx_copy_header_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                         : read_state_ptr=%p\n",
		read_state_ptr));


    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 0;
	goto clean;
    }

    if ((len1 == 1 && 0 == memcmp(buffer1,"\n",len1)) ||
	(len1 == 2 && 0 == memcmp(buffer1,"\r\n",len1))) {
	/* End of headers */
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer1,&len1);
	status = 0;
	goto clean;
    }
    
    if (NULL == memchr(buffer1,':',len1) ||
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1) ||
#endif
	0 == memcmp(buffer1,"From ",5)) {
	/* End of headers -- bad header */
	status = 0;
	goto clean;
    }

    do {
	append_buffer(buffer,len,buffer1,len1);
	mbx_accept_buffered_line(folder,read_state_ptr,
				 &buffer1,&len1);
	if (!mbx_read_buffered_line(folder,read_state_ptr,
					 &buffer1,&len1)) {
	    break;
	}
    } while (len1 > 0 && whitespace(buffer1[0]));
    status = 1;

 clean:
    /* Not need to free() -- buffer1 is same than next_line -buffer */

    if (!status) {
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_header_non_spool=%d | len=%d, buffer=%s",
		status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
	dprint(50, (debugfile,
		    "\nmbx_copy_header_non_spool <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_header_spool P_((struct folder_info *folder,
				     READ_STATE read_state_ptr,
				     char **buffer, int *len));
static int mbx_copy_header_spool(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    char *buffer1 = NULL;
    int len1;

    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_header_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                     : read_state_ptr=%p\n",
		read_state_ptr));

    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 0;
	goto clean;
    }

    if ((len1 == 1 && 0 == memcmp(buffer1,"\n",len1)) ||
	(len1 == 2 && 0 == memcmp(buffer1,"\r\n",len1))) {
	/* End of headers */
	if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	    status = 0;
	    goto clean;
	}
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer1,&len1);
	status = 0;
	goto clean;
    }
    
    if (NULL == memchr(buffer1,':',len1) ||
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1) ||
#endif
	0 == memcmp(buffer1,"From ",5)) {
	/* End of headers -- bad header */
	status = 0;
	goto clean;
    }

    do {
	append_buffer(buffer,len,buffer1,len1);
	if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	    status = 0;
	    goto clean;
	}
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer1,&len1);
	if (!mbx_read_buffered_line(folder,read_state_ptr,
					 &buffer1,&len1)) {
	    break;
	}
    } while (len1 > 0 && whitespace(buffer1[0]));
    status = 1;

 clean:
    /* Not need to free() -- buffer1 is same than next_line -buffer */

    if (!status) {
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_header_spool=%d | len=%d, buffer=%s",
		status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
	dprint(50, (debugfile,
		    "\nmbx_copy_header_spool <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_body_no_name P_((struct folder_info *folder,
				     READ_STATE read_state_ptr,
				     char **buffer, int *len,
				     long *content_remaining));

static int mbx_copy_body_no_name(folder,read_state_ptr,buffer,len,
				 content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     long *content_remaining;

{ 
    dprint(11, (debugfile,
		"mbx_copy_body_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                     : read_state_ptr=%p\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_copy_body_no_name=0\n"));
    return 0;
}

static int mbx_copy_body_non_spool P_((struct folder_info *folder,
				       READ_STATE read_state_ptr,
				       char **buffer, int *len,
				       long *content_remaining));


static int mbx_copy_body_non_spool(folder,read_state_ptr,buffer,len,
				   content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     long *content_remaining;
{
    char *buffer1;
    int len1;

    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_body_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                       : read_state_ptr=%p\n",
		read_state_ptr));


    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 0;
	goto clean;
    }

    if (*content_remaining < 1 && 
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1)
#else
	0 == memcmp(buffer1,"From ",5) &&
	real_from(buffer1,NULL)
#endif
	) {
	status = 0;
	goto clean;
    }

    append_buffer(buffer,len,buffer1,len1);
    mbx_accept_buffered_line(folder,read_state_ptr,
				  &buffer1,&len1);
    
    if (*content_remaining > *len)
	*content_remaining -= *len;
    else if (*content_remaining >= 0)
	*content_remaining = 0;
    status = 1;
    
 clean:
    /* buffer1 is same then next_line -buffer -- not need free() */

    if (!status) {
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_body_non_spool=%d | len=%d, buffer=%s",
		status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
	dprint(50, (debugfile,
		    "\nmbx_copy_body_non_spool <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_body_spool P_((struct folder_info *folder,
				   READ_STATE read_state_ptr,
				   char **buffer, int *len,
				   long *content_remaining));

static int mbx_copy_body_spool(folder,read_state_ptr,buffer,len,
				   content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     long *content_remaining;
{
    char *buffer1;
    int len1;
 
    int status = 0;

    dprint(11, (debugfile,
		"mbx_copy_body_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                   : read_state_ptr=%p\n",
		read_state_ptr));

    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 0;
	goto clean;
    }

    if (*content_remaining < 1 && 
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1) 
#else
	len1 >= 5 && 0 == memcmp(buffer1,"From ",5) &&
	real_from(buffer1,NULL)
#endif
	) {
	status = 0;
	goto clean;
    }

    append_buffer(buffer,len,buffer1,len1);
    if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	status = 0;
	goto clean;
    }
    mbx_accept_buffered_line(folder,read_state_ptr,
				  &buffer1,&len1);
    
    if (*content_remaining > *len)
	*content_remaining -= *len;
    else if (*content_remaining >= 0)
	*content_remaining = 0;
    status = 1;

 clean:
    /* buffer1 is same then next_line -buffer -- not need free() */

    if (!status) {
	if (*buffer) {
	    free(*buffer);
	    *buffer = NULL;
	}
    }

    dprint(50, (debugfile,
		"mbx_copy_body_spool=%d | len=%d, buffer=%s",
		status,*len,
		*buffer ? *buffer : "NULL"));
    if (*len < 1 || !*buffer || (*buffer)[*len -1] != '\n') {
	dprint(50, (debugfile,
		    "\nmbx_copy_body_spool <- NO NEWLINE\n"));
    }
    return status;
}

static int mbx_copy_envelope_end_no_name P_((struct folder_info *folder,
					     READ_STATE read_state_ptr));

static int mbx_copy_envelope_end_no_name(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_copy_envelope_end_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                             : read_state_ptr=%p\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_copy_envelope_end_no_name=1\n"));
    return 1; /* Always end of message */
}

#ifndef MMDF
static int mbx_copy_envelope_end_from P_((struct folder_info *folder,
					  READ_STATE read_state_ptr));

static int mbx_copy_envelope_end_from(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    char *buffer1;
    int len1;
    int status = 0;

    /* Will be re-handled on copy_envelope_folder() */
    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 1;
	goto clean;
    }
    if (real_from(buffer1,NULL))
	status = 1;
    else
	status = 0;
 clean:
    dprint(12, (debugfile,
		"mbx_copy_envelope_end_from=%d\n",
		status));
    return status;
}
#endif


static int mbx_copy_envelope_end_non_spool P_((struct folder_info *folder,
					       READ_STATE read_state_ptr));

static int mbx_copy_envelope_end_non_spool(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    char *buffer1;
    int len1;

    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_end_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                               : read_state_ptr=%p\n",
		read_state_ptr));


    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 1;
	goto clean;
    }

    if (
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1) 
#else
	(len1 == 1 && 0 == memcmp(buffer1,"\n",len1)) ||
	(len1 == 2 && 0 == memcmp(buffer1,"\r\n",len1))
#endif
	) {
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer1,&len1);
	status = 1;
    }

#ifndef MMDF
    /* Will be re-handled on copy_envelope_folder() */
    if (mbx_copy_envelope_end_from(folder,read_state_ptr))
	status = 1;
    else
	status = 0;
#endif
 clean:
    dprint(11, (debugfile,
		"mbx_copy_envelope_end_non_spool=%d\n",
		status));
    return status; 
}

static int mbx_copy_envelope_end_spool P_((struct folder_info *folder,
					   READ_STATE read_state_ptr));

static int mbx_copy_envelope_end_spool(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    char *buffer1;
    int len1;

    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_end_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                           : read_state_ptr=%p\n",
		read_state_ptr));


    if (!mbx_read_buffered_line(folder,read_state_ptr,
				     &buffer1,&len1)) {
	status = 1;
	goto clean;
    }

    if (
#ifdef MMDF
	mbx_copy_envelope_check_MMDF(buffer1,len1) 
#else
	(len1 == 1 && 0 == memcmp(buffer1,"\n",len1)) ||
	(len1 == 2 && 0 == memcmp(buffer1,"\r\n",len1))
#endif
	) {
	if (!mbx_copy_line_to_temp(folder,buffer1,len1)) {
	    status = 0;
	    goto clean;
	}
	mbx_accept_buffered_line(folder,read_state_ptr,
				      &buffer1,&len1);
	status = 1;
    }

#ifndef MMDF
    /* Will be re-handled on copy_envelope_folder() */
    if (mbx_copy_envelope_end_from(folder,read_state_ptr))
	status = 1;
    else
	status = 0;
#endif
 clean:
    dprint(11, (debugfile,
		"mbx_copy_envelope_end_spool=%d\n",
		status));
    return status; 
}

static int mbx_copy_envelope_reset_body_no_name P_((struct folder_info *folder,
						    READ_STATE
						    read_state_ptr));
static int mbx_copy_envelope_reset_body_no_name(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                                    : read_state_ptr=%p\n",
		read_state_ptr));

    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_no_name=0\n"));
    return 0;
}

static int mbx_reset_folder P_((struct folder_info *folder,
				READ_STATE read_state_ptr));

static int mbx_reset_folder(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    int status = 0;
    if (fseek(folder->p->fh_folder, read_state_ptr -> fbytes_body , 
	      SEEK_SET) == -1) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCouldntSeekBytesIntoFolder,
			  "\nCouldn't seek %ld bytes into folder.\n"),
		  read_state_ptr -> fbytes_body);		
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile,
		   "Error: Couldn't seek folder %s: (offset %ld) Errno %s (%s)\n",
		   folder->cur_folder, read_state_ptr -> fbytes_body, 
		   error_description(err), "mbx_reset_folder"));
	status = 0;
	goto clean;
    }
    read_state_ptr -> fbytes = read_state_ptr -> fbytes_body;
    mbx_free_rs_fields_file(read_state_ptr);
    status = 1;
 clean:
    dprint(12, (debugfile,
		"mbx_reset_folder=%d\n",
		status));
    return status;
}

static int mbx_copy_envelope_reset_body_non_spool P_((struct folder_info 
						      *folder,
						      READ_STATE
						      read_state_ptr));
static int mbx_copy_envelope_reset_body_non_spool(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                                      : read_state_ptr=%p\n",
		read_state_ptr));

    status = mbx_reset_folder(folder,read_state_ptr);

    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_non_spool=%d\n",status));
    return status;
}

static int mbx_copy_envelope_reset_body_spool P_((struct folder_info 
						  *folder,
						  READ_STATE
						  read_state_ptr));
static int mbx_copy_envelope_reset_body_spool(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"                                  : read_state_ptr=%p\n",
		read_state_ptr));

    
    if (!folder->p->fh_temp) {
	dprint(11, (debugfile,
		    "mbx_copy_envelope_reset_body_spool -- NO temp file %s\n",
		    folder->cur_tempfolder));
    } else if (fseek(folder->p->fh_temp, read_state_ptr -> fbytes_body , 
	      SEEK_SET) == -1) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCouldntSeekBytesIntoFolder,
			  "\nCouldn't seek %ld bytes into folder.\n"),
		  read_state_ptr -> fbytes_body);		
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile,
		   "Error: Couldn't seek folder %s: (offset %ld) Errno %s (%s)\n",
		   folder->cur_tempfolder, read_state_ptr -> fbytes_body, 
		   error_description(err), 
		   "mbx_copy_envelope_reset_body_spool"));
	status = 0;
	goto clean;
    }

    status = mbx_reset_folder(folder,read_state_ptr);

 clean:
    dprint(11, (debugfile,
		"mbx_copy_envelope_reset_body_spool=%d\n",status));
    return status;
}

static FILE * mbx_no_name_to_fd P_((struct folder_info *folder,long offset));
static FILE * mbx_no_name_to_fd(folder,offset)
     struct folder_info *folder;
     long offset;
{
    dprint(11, (debugfile,
		"mbx_no_name_to_fd: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_no_name_to_fd=NULL\n"));
    return NULL;
}

static FILE * mbx_non_spool_to_fd P_((struct folder_info *folder,long offset));
static FILE * mbx_non_spool_to_fd(folder,offset)
     struct folder_info *folder;
     long offset;
{
    FILE * status = NULL;
    dprint(11, (debugfile,
		"mbx_non_spool_to_fd: folder=%p (%s)\n",
		folder,folder->cur_folder));
    if (offset != -1L &&
	fseek(folder->p->fh_folder, offset , 
	      SEEK_SET) == -1) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCouldntSeekBytesIntoFolder,
			  "\nCouldn't seek %ld bytes into folder.\n"),
		  offset);		
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile,
		   "Error: Couldn't seek folder %s: (offset %ld) Errno %s (%s)\n",
		   folder->cur_folder, offset, 
		   error_description(err), "mbx_non_spool_to_fd"));
	status = NULL;
	goto clean;
    }
    status = folder->p->fh_folder;
 clean:
    dprint(11, (debugfile,
		"mbx_non_spool_to_fd=%s (%p)\n",
		status ? "NON-NULL" : "NULL",
		status));
    return status;
}

static FILE * mbx_spool_to_fd P_((struct folder_info *folder,long offset));
static FILE * mbx_spool_to_fd(folder,offset)
     struct folder_info *folder;
     long offset;
{
    FILE * status = NULL;
    dprint(11, (debugfile,
		"mbx_spool_to_fd: folder=%p (%s)\n",
		folder,folder->cur_folder));
    if (offset != -1L &&
	fseek(folder->p->fh_temp, offset , 
	      SEEK_SET) == -1) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmCouldntSeekBytesIntoFolder,
			  "\nCouldn't seek %ld bytes into folder.\n"),
		  offset);		
	lib_error(FRM("** %s. **\n"), error_description(err));
	dprint(1, (debugfile,
		   "Error: Couldn't seek folder %s: (offset %ld) Errno %s (%s)\n",
		   folder->cur_tempfolder, offset, 
		   error_description(err), "mbx_spool_to_fd"));
	status = NULL;
	goto clean;
    }
    status = folder->p->fh_temp;
 clean:
    dprint(11, (debugfile,
		"mbx_spool_to_fd=%s (%p)\n",
		status ? "NON-NULL" : "NULL",
		status));
    return status;
}

static int mbx_new_mail_on_no_name P_((struct folder_info *folder,int *bytes));
static int mbx_new_mail_on_no_name(folder,bytes)
     struct folder_info *folder;
     int *bytes;
{
    dprint(11, (debugfile,
		"mbx_new_mail_on_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_new_mail_on_no_name=0\n"));
    return 0;
}

static int mbx_new_mail_on_normal P_((struct folder_info *folder,
					 int *bytes));
static int mbx_new_mail_on_normal(folder,bytes)
     struct folder_info *folder;
     int *bytes;
{
    int status = 0;
    long size = 0L;
    dprint(11, (debugfile,
		"mbx_new_mail_on_normal: folder=%p (%s)\n",
		folder,folder->cur_folder));

    size = file_bytes(folder->cur_folder);
    if (size > folder->mailfile_size)
	status = 1;
    *bytes = size - folder->mailfile_size;

    dprint(11, (debugfile,
		"mbx_new_mail_on_normal=%d (*bytes=%d)\n",
		status,*bytes));
    return status;
}

static int mbx_consider_remove_null P_((struct folder_info *folder));
static int mbx_consider_remove_null(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_consider_remove_null: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_consider_remove_null=0\n"));
    return 0;
}

static int mbx_consider_remove_non_spool P_((struct folder_info *folder));
static int mbx_consider_remove_non_spool(folder)
     struct folder_info *folder;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_consider_remove_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));
    if (0L == file_bytes(folder->cur_folder)) {
	folder->mailfile_size = 0;
	if (0 == unlink(folder->cur_folder))
	    status = 1;
	else {
	    int err = errno;
	    dprint(11, (debugfile,
			"mbx_consider_remove_non_spool: unlink(%s) failed: %s (errno=%d)\n",
			folder->cur_folder,
			error_description(err),err));
	}
    }
    dprint(11, (debugfile,
		"mbx_consider_remove_non_spool=%d\n",
		status));
    return status;
}

static void malloc_keep_folder_state P_((struct keep_folder_state **ptr));
static void malloc_keep_folder_state(ptr)
     struct keep_folder_state **ptr;
{
    *ptr = safe_malloc(sizeof (struct keep_folder_state));
    bzero((void *)*ptr,sizeof (struct keep_folder_state));
}

static void mbx_zero_ks_fields_file P_((struct keep_folder_state *rs));
static void mbx_zero_ks_fields_file(rs)
     struct keep_folder_state *rs;
{
    rs -> a.file.temp_keep_file       = NULL;
    rs -> a.file.keep_file            = NULL;
}


static void free_keep_folder_state P_((struct keep_folder_state **ptr));
static void free_keep_folder_state(ptr)
     struct keep_folder_state **ptr;
{
    free(*ptr);
    *ptr = NULL;
}

static void mbx_free_ks_fields_file P_((struct keep_folder_state *rs));
static void mbx_free_ks_fields_file(rs)
     struct keep_folder_state *rs;
{
    if (rs -> a.file.temp_keep_file) {
	free(rs -> a.file.temp_keep_file);
	rs -> a.file.temp_keep_file = NULL;
    }

    if (rs -> a.file.keep_file) {
	fclose(rs -> a.file.keep_file);
	rs -> a.file.keep_file = NULL;
    }
}

static int mbx_open_temp_file P_((struct folder_info *folder,
				  KEEP_STATE keep_state_ptr));
static int mbx_open_temp_file(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    int status = 0,err;
    keep_state_ptr -> a.file.temp_keep_file = 
	elm_message(FRM("%s%s%d"),temp_dir, temp_file, getpid());

    if ((err = can_open(keep_state_ptr -> a.file.temp_keep_file, "sw"))) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveTempFileDenied,
			  "Permission to create temp file %s for writing denied! Leaving folder intact."),
		  keep_state_ptr -> a.file.temp_keep_file);
	dprint(1, (debugfile,
		   "Error: Permission to create temp file %s denied!! (%s)\n",
		   keep_state_ptr -> a.file.temp_keep_file, 
		   "mbx_open_temp_file"));
	dprint(1, (debugfile, "** %s **\n", error_description(err)));
	status = 0;
	goto clean;
    }

    if ((keep_state_ptr -> a.file.keep_file = 
	 fopen(keep_state_ptr -> a.file.temp_keep_file,"w+")) == NULL) {
	err = errno;
	
	dprint(1, (debugfile, "Error: could not create file %s\n", 
		   keep_state_ptr ->a.file.temp_keep_file));
	dprint(1, (debugfile, "** %s **\n", error_description(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveCouldNotCreate,
			  "Could not create temp file %s!"), 
		  keep_state_ptr -> a.file.temp_keep_file);

	status = 0;
	goto clean;
    }
    status = 1;

 clean:	
    return status;
}

static int mbx_prepare_keep_no_name P_((struct folder_info *folder,
				       KEEP_STATE keep_state_ptr));
static int mbx_prepare_keep_no_name(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_prepare_keep_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_prepare_keep_no_name=0\n"));
    return 0;
}

static int mbx_prepare_keep_normal P_((struct folder_info *folder,
				       KEEP_STATE keep_state_ptr));
static int mbx_prepare_keep_normal(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_prepare_keep_normal: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!mbx_open_temp_file(folder,keep_state_ptr)) {
	status = 0;
	goto clean;
    }
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_prepare_keep_normal=%d\n",
		status));
    return status;
}

static int mbx_end_keep_no_name P_((struct folder_info *folder,
				    KEEP_STATE keep_state_ptr));
static int mbx_end_keep_no_name(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    dprint(11, (debugfile,
		"mbx_end_keep_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_end_keep_no_name=1\n"));
    return 1;
}

static int copy_to_folder P_((FILE * from, struct folder_info *to));
static int copy_to_folder(from, to) 
     FILE * from;
     struct folder_info *to;
{
  int need_reopen = 1;
  int truncate_fh = -1;
  int code;

  dprint(10, (debugfile,
	      "copy_to_folder: folder=%p (%s)\n",
	      to,to->cur_folder));

  rewind(to->p->fh_folder);
#ifdef FTRUNCATE  
  if (0 == ftruncate(fileno(to->p->fh_folder),0)) {
    dprint(10, (debugfile,
		"copy_to_folder: truncated (ftruncate)\n"));
    need_reopen = 0;
  }
  else {
    int err = errno;
    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantTruncateFolder,
		      "Sorry, can't truncate the folder %s [%s]!"),
	      to -> cur_folder, error_description(err));    
  }
#endif
  if (need_reopen) {
    truncate_fh = open(to->cur_folder,O_RDWR|O_TRUNC,0600);
    if (truncate_fh == -1) {
      int err = errno;

      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSorryCantTruncateFolder,
			"Sorry, can't truncate the folder %s [%s]!"),
		to -> cur_folder, error_description(err));    
      return 1;
    }
    dprint(10, (debugfile,
		"copy_to_folder- truncated (reopened)\n"));
  }
  code = copy_to_fh(from,to->p->fh_folder);

  /* In this point we may lost locks,
     but copy is already done */
  if (truncate_fh != -1)
      close(truncate_fh);

  dprint(10, (debugfile,
	      "copy_to_folder=%d\n",code));
  return code;
}

static int mbx_move_temp_file P_((struct folder_info *folder,
				  KEEP_STATE keep_state_ptr,
				  int need_to_copy));
static int mbx_move_temp_file(folder,keep_state_ptr,need_to_copy)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
     int need_to_copy;
{
    int status = 0;
    int was_symlink = FALSE;

    struct stat    buf;		/* stat command  */
#ifdef SYMLINK
    struct stat    lbuf;		/* lstat command  */
#endif
#if defined(BSD_TYPE) && !defined(UTIMBUF)
    time_t utime_buffer[2];		/* utime command */
#else
    struct utimbuf utime_buffer;	/* utime command */
#endif

    if (fflush(keep_state_ptr->a.file.keep_file) == EOF ) {
	int err = errno;
	
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveFlushFailedTemp,
			  "Flush failed in leavembox: %s: %s"),
		  keep_state_ptr->a.file.temp_keep_file,
		  error_description(err));
	dprint(2, (debugfile, 
		   "\n\rfflush err on temp_keep_file - mbx_move_temp_file\n\r"));
	unlink(keep_state_ptr->a.file.temp_keep_file);
	status = 0;
	goto clean;
    }

    if (ferror(folder->p->fh_folder)) {
	dprint(1,(debugfile,"Failed to read mailfile!"));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorReadMailfile,
			  "Error when reading mailfile %s"),
		  folder->cur_folder);
	clearerr(folder->p->fh_folder);
	unlink(keep_state_ptr->a.file.temp_keep_file);
	status = 0;
	goto clean;
    }
    
    dprint(10, (debugfile,
		"!! Need copy (or link) from %s to %s\n",
		keep_state_ptr->a.file.temp_keep_file, 
		folder->cur_folder));	    

    /* Get original permissions and access time of the original
     * mail folder before we remove it.
     */
    if(save_file_stats(folder->cur_folder) != 0) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveProblemsSavingPerms,
			  "Problems saving permissions of folder %s!"), 
		  folder->cur_folder);
	
    }

    if (stat(folder->cur_folder, &buf) != 0) {
	int err = errno;
	dprint(1, (debugfile, "Error: errno %s attempting to stat file %s\n", 
		   error_description(err), 
		   folder->cur_folder));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorOnStat,
			  "Error %s on stat(%s)."), 
		  error_description(err), 
		  folder->cur_folder);
    }

#ifdef SYMLINK
    if (lstat(folder->cur_folder, &lbuf) != 0) {
	int err = errno;
	dprint(1, (debugfile, "Error: errno %s attempting to stat file %s\n", 
		   error_description(err), 
		   folder->cur_folder));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveErrorOnStat,
			  "Error %s on stat(%s)."), 
		  error_description(err), 
		  folder->cur_folder);
    }
#endif

    flush_folder(folder);

    if (buf.st_nlink > 1) {
	dprint(10, (debugfile,"* several links\n"));
	need_to_copy = TRUE;
    }
    if (buf.st_mode & 07000) { /* copy if special modes set */
	dprint(10, (debugfile,"* special modes\n"));
	need_to_copy = TRUE;    /* such as enforcement lock */
    }
#ifdef SYMLINK
    if (
#ifdef S_ISLNK
	S_ISLNK(lbuf.st_mode)
#else
	(lbuf.st_mode & S_IFMT) == S_IFLNK
#endif
	) {
	dprint(10, (debugfile,"* symlink\n"));
	need_to_copy = TRUE;
	was_symlink = TRUE;
    }
#endif

#ifdef _PC_CHOWN_RESTRICTED
    if (!need_to_copy) {
/*
 * Chown may or may not be restricted to root in SVR4, if it is,
 *	then need to copy must be true, and no restore of permissions
 *	should be performed.
 */
	if (pathconf(folder->cur_folder, 
		     _PC_CHOWN_RESTRICTED)) {
	    dprint(10, (debugfile,"* _PC_CHOWN_RESTRICTED\n"));
	    
	    need_to_copy = TRUE;
	}
    }
#endif  /* _PC_CHOWN_RESTRICTED */

    dprint(10, (debugfile,
		"-- need_to_copy=%d\n",
		need_to_copy));

    if(!need_to_copy) {
	if (0 == unlink(folder->cur_folder)) {
	    dprint(10, (debugfile,
			"!! unlinked folder: %s\n",
			folder->cur_folder));
	}

	if (link(keep_state_ptr->a.file.temp_keep_file, 
		 folder->cur_folder) != 0) {
	    if(errno == EXDEV || errno == EEXIST) {
		int err = errno;
		/* oops - can't link across file systems - 
		   use copy instead */
		
		dprint(10, (debugfile, "link(%s, %s) failed: %s\n", 
			    keep_state_ptr->a.file.temp_keep_file, 
			    folder->cur_folder,
			    error_description(err)));		
		need_to_copy = TRUE;
	    } else {
		int err = errno;
		dprint(1, (debugfile, "link(%s, %s) failed (leavembox)\n", 
			   keep_state_ptr->a.file.temp_keep_file, 
			   folder->cur_folder));
		dprint(1, (debugfile, "** %s **\n", 
			   error_description(err)));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveLinkFailed,
				  "Link failed! %s.\n"), 
			  error_description(err));
		need_to_copy = TRUE;
	    }
	} else {
	    dprint(10, (debugfile, "!! linked %s to  %s\n", 
			keep_state_ptr->a.file.temp_keep_file, 
			folder->cur_folder));
	    status = 1;
	}

	if (folder->p->fh_folder)
	    fclose(folder->p->fh_folder);	
	folder->p->fh_folder = NULL;
	
	folder->p->fh_folder = open_or_create(folder->cur_folder);
	    
	if (NULL == folder->p->fh_folder) {
	    goto PANIC_copy_failed;
	}
	dprint(10, (debugfile,
		    "!! reopened (created) current folder after unlink: %s\n",
		    folder->cur_folder));	        
    }

    if(need_to_copy) {
	char new_cur_folder[100];

	rewind(keep_state_ptr->a.file.keep_file);
	if (copy_to_folder(keep_state_ptr->a.file.keep_file, folder) != 0) {
	    /* copy to cur_folder failed - try to copy to special file */
	    {
		int err = errno;
		dprint(1, (debugfile, "leavembox: copy(%s, %s) failed;",
			   keep_state_ptr->a.file.temp_keep_file, 
			   folder->cur_folder));
		dprint(1, (debugfile, "** %s **\n", error_description(err)));
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmLeaveCouldntModifyFolder,
				  "Couldn't modify folder!"));
	    }
	PANIC_copy_failed:

	    if (sleepmsg > 0) {
#if POLL_METHOD	  
		wait_for_timeout((sleepmsg + 1) / 2);
#else
		sleep((sleepmsg + 1) / 2);
#endif
	    }

	    elm_sfprintf(new_cur_folder,sizeof new_cur_folder,
			 FRM("%s/%s"), home, unedited_mail);

	    if (copy1(keep_state_ptr->a.file.keep_file, 
		      new_cur_folder,FALSE) != 0) {

		/* couldn't copy to special file either */
		int err = errno;
		dprint(1, (debugfile, 
			   "leavembox: couldn't copy to %s either!!  Help;", 
			   new_cur_folder));
		dprint(1, (debugfile, "** %s **\n", error_description(err)));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveCantCopyMailbox,
				  "Can't copy folder, system trouble : contents preserved in %s\n"),
			  keep_state_ptr->a.file.temp_keep_file);
		
		fclose(keep_state_ptr->a.file.keep_file);
		keep_state_ptr->a.file.keep_file = NULL;
		status = 0;
		goto clean;
	    } else {
		dprint(1, (debugfile,
			   "\nWoah! Confused - Saved mail in %s (leavembox)\n", 
			   new_cur_folder));
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveSavedMailIn,
				  "Saved mail in %s."), 
			  new_cur_folder);
	    }
	} else
	    status = 1;
    }

    /* link or copy complete - remove temp keep file */
    unlink(keep_state_ptr->a.file.temp_keep_file);
    fclose(keep_state_ptr->a.file.keep_file);
    keep_state_ptr->a.file.keep_file = NULL;

    /*
     * restore permissions and access times of folder only if not
     * a symlink, as symlinks have no permissions, and not worth
     * tracking down what it points to.
     */

    if (!was_symlink) {
	if(restore_file_stats(folder->cur_folder) != 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveProblemsRestoringPerms,
			      "Problems restoring permissions of folder %s!"),
		      folder->cur_folder);
	}
    }

#if defined(BSD_TYPE) && !defined(UTIMBUF)
    utime_buffer[0]     = buf.st_atime;
    utime_buffer[1]     = buf.st_mtime;
#else
    utime_buffer.actime = buf.st_atime;
    utime_buffer.modtime= buf.st_mtime;
#endif

    if (utime(folder->cur_folder,
#if defined(BSD_TYPE) && !defined(UTIMBUF)
	      utime_buffer
#else
	      &utime_buffer
#endif
	      ) != 0) {
	int err = errno;
	dprint(1, (debugfile, 
		   "Error: encountered error doing utime (leavmbox)\n"));
	dprint(1, (debugfile, "** %s **\n", error_description(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLeaveChangingAccessTime,
			  "Error %s trying to change file %s access time."), 
		  error_description(err), 
		  folder->cur_folder);
    }
    
 clean:
    dprint(12, (debugfile,
		"mbx_move_temp_file=%d\n",
		status));
    return status;
}

static int mbx_end_keep_non_spool P_((struct folder_info *folder,
				    KEEP_STATE keep_state_ptr));
static int mbx_end_keep_non_spool(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    int need_to_copy = FALSE;
    int status = 0;
    dprint(11, (debugfile,
		"mbx_end_keep_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (lockfolders) {
	dprint(10, (debugfile,"* lockfolders\n"));
	need_to_copy = TRUE;
    }

    if (!mbx_move_temp_file(folder,keep_state_ptr,need_to_copy)) {
	status = 0;
	goto clean;
    }
    folder->mailfile_size = file_bytes(folder->cur_folder);
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_end_keep_non_spool=%d\n",
		status));
    return status;
}

static int mbx_end_keep_spool P_((struct folder_info *folder,
				    KEEP_STATE keep_state_ptr));
static int mbx_end_keep_spool(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
{
    int status = 0;
    int need_to_copy = FALSE;

#ifdef SAVE_GROUP_MAILBOX_ID
    gid_t oldgid = getegid();
    setgid(mailgroupid);
#endif

    dprint(11, (debugfile,
		"mbx_end_keep_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

#ifdef SYSCALL_LOCKING
    dprint(10, (debugfile,"* SYSCALL_LOCKING (SPOOL)\n"));
    need_to_copy = TRUE;
#endif /* SYSCALL_LOCKING */

    if (!mbx_move_temp_file(folder,keep_state_ptr,need_to_copy)) {
	status = 0;
	goto clean;
    }
    folder->mailfile_size = file_bytes(folder->cur_folder);
    status = 1;

 clean:
#ifdef SAVE_GROUP_MAILBOX_ID
    setgid(oldgid);
#endif

    dprint(11, (debugfile,
		"mbx_end_keep_spool=%d\n",
		status));
    return status;
}

static void mbx_mark_keep_no_name P_((struct folder_info *folder,
				      KEEP_STATE keep_state_ptr,
				      struct header_rec *entry,
				      int keep));
static void mbx_mark_keep_no_name(folder,keep_state_ptr,entry,keep)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
     struct header_rec *entry;
     int keep;
{
    dprint(11, (debugfile,
		"mbx_mark_keep_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
}


static void mbx_mark_keep_normal P_((struct folder_info *folder,
				      KEEP_STATE keep_state_ptr,
				      struct header_rec *entry,
				      int keep));
static void mbx_mark_keep_normal(folder,keep_state_ptr,entry,keep)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
     struct header_rec *entry;
     int keep;
{
    int count;
    FILE * fh;
    int status_written = 0;
    dprint(11, (debugfile,
		"mbx_mark_keep_normal: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (keep) {
	fh = folder_to_fd(folder,entry->offset);
	if (!fh)
	    return;
    
	dprint(11, (debugfile,
		    "mbx_mark_keep_normal: envelope + header %ld - %ld\n",
		    entry->offset,entry->mime_rec.offset));
	count = 0;
	while (ftell(fh) < entry->mime_rec.offset) {
	    char * buffer = NULL,*val;
	    int len,i;

	    if (!read_line(fh,&buffer,&len,
			   entry->mime_rec.offset - ftell(fh)))
		return;
	    
	    /* End of headers ? */
	    if ((len == 1 && 0 == memcmp(buffer,"\n",len)) ||
		(len == 2 && 0 == memcmp(buffer,"\r\n",len))) {
		if (!status_written) {
		    count += fprintf(keep_state_ptr->a.file.keep_file,
				     "Status: ");

		    if (!ison(entry->status, UNREAD)) {
			putc('R',keep_state_ptr->a.file.keep_file);
			count ++;
		    }
		    if (ison(entry->status, NEW)) {
			putc('N',keep_state_ptr->a.file.keep_file);
			count ++;
		    } else {
			putc('O',keep_state_ptr->a.file.keep_file);
			count ++;
		    }
		    if (ison(entry->status, REPLIED)) {
			putc('r',keep_state_ptr->a.file.keep_file);
			count ++;
		    }
		    /* Write unknown flags */
		    for (i=0; entry->mailx_status[i] != '\0'; i++)
			switch(entry->mailx_status[i]) {
			case 'R':
			case 'O':
			case 'r':
			case 'N':
			    break;
			default:
			    putc(entry->mailx_status[i],
				 keep_state_ptr->a.file.keep_file);
			    count++;
			}
		    /* Write EOLN */
		    for (i = 0; i < len; i++) {
			putc(buffer[i],keep_state_ptr->a.file.keep_file);
			count++;
		    }
		    status_written = 1;
		}
	    }

	    if (NULL != (val = header_cmp(buffer, "Status", NULL))) {
		/* SKIP */
	    } else {
		for (i = 0; i < len; i++) {
		    putc(buffer[i],keep_state_ptr->a.file.keep_file);
		    count++;
		}
	    }
	    free(buffer);
	}
	if (!status_written) {
	    dprint(1, (debugfile,
		       "mbx_mark_keep_normal: ERROR : Status: not written!\n"));
	}

	dprint(11, (debugfile,
		    "mbx_mark_keep_normal: wrote %d bytes of header\n",count));
	dprint(11, (debugfile,
		    "mbx_mark_keep_normal: content length %ld\n",
		    entry->content_length));
	
	count = 0;
	while (count < entry->content_length) {
	    int ch = fgetc(fh);
	    
	    putc(ch,keep_state_ptr->a.file.keep_file);
	    if (ferror(fh))
		return;	
	    count++;
	}
	dprint(11, (debugfile,
		    "mbx_mark_keep_normal: wrote %d bytes of body\n",
		    count));
    }
}

CONST char * mbx_no_name_type P_((struct folder_info *folder));
CONST char * mbx_no_name_type(folder)
     struct folder_info *folder;
{
    char * status = "NULL";

    dprint(11, (debugfile,
		"mbx_no_name_type: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_no_name_type=%s\n",
		status));
    return status;
}

CONST char * mbx_non_spool_type P_((struct folder_info *folder));
CONST char * mbx_non_spool_type(folder)
     struct folder_info *folder;
{
    static char *FOLDER = NULL;
    if (!FOLDER)
	FOLDER = catgets(elm_msg_cat, ElmSet, ElmFolder, "Folder");

    dprint(11, (debugfile,
		"mbx_non_spool_type: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_non_spool_type=%s\n",
		FOLDER));
    return FOLDER;
}

CONST char * mbx_read_only_type P_((struct folder_info *folder));
CONST char * mbx_read_only_type(folder)
     struct folder_info *folder;
{
    static char *FOLDER = NULL;
    if (!FOLDER)
	FOLDER = catgets(elm_msg_cat, ElmSet, ElmReadOnlyFolder, 
			 "Read-only folder");

    dprint(11, (debugfile,
		"mbx_read_only_type: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_read_only_type=%s\n",
		FOLDER));
    return FOLDER;
}

CONST char * mbx_spool_type P_((struct folder_info *folder));
CONST char * mbx_spool_type(folder)
     struct folder_info *folder;
{
    static char *mailbox = NULL;
    if (!mailbox)
	mailbox = catgets(elm_msg_cat, ElmSet, ElmMailbox, "Mailbox");

    dprint(11, (debugfile,
		"mbx_spool_type: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_spool_type=%s\n",
		mailbox));
    return mailbox;
}

static int mbx_start_edit_no_name P_((struct folder_info *folder, 
				     CONST char **buffer));
static int mbx_start_edit_no_name(folder,buffer)
     struct folder_info *folder;
     CONST char **buffer;
{
    dprint(11, (debugfile,
		"mbx_start_edit_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_start_edit_no_name=0\n"));
    return 0;
}

static int mbx_start_edit_non_spool P_((struct folder_info *folder, 
				     CONST char **buffer));
static int mbx_start_edit_non_spool(folder,buffer)
     struct folder_info *folder;
     CONST char **buffer;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_start_edit_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    *buffer = folder->cur_folder;
    status = 1;

    dprint(50, (debugfile,
		"mbx_start_edit_non_spool=%d | *buffer=%s\n",
		status,
		*buffer ? *buffer : "NULL"));
    return status;
}

static int mbx_start_edit_spool P_((struct folder_info *folder, 
				     CONST char **buffer));
static int mbx_start_edit_spool(folder,buffer)
     struct folder_info *folder;
     CONST char **buffer;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_start_edit_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    *buffer = NULL;

    if(save_file_stats(folder->cur_folder) != 0) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmPermFolder,
			  "Problems saving permissions of folder %s!"), 
		  folder->cur_folder);
	status = 0;
	goto clean;
    }

    *buffer = folder->cur_tempfolder;
    status = 1;

 clean:
    dprint(50, (debugfile,
		"mbx_start_edit_spool=%d | *buffer=%s\n",
		status,
		*buffer ? *buffer : "NULL"));
    return status;
}

static int mbx_end_edit_no_name P_((struct folder_info *folder));
static int mbx_end_edit_no_name(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_end_edit_no_name: folder=%p (%s)\n",
		folder,folder->cur_folder));
    dprint(11, (debugfile,
		"mbx_end_edit_no_name=0\n"));
    return 0;
}

static int mbx_end_edit_non_spool P_((struct folder_info *folder));
static int mbx_end_edit_non_spool(folder)
     struct folder_info *folder;
{
    int status = 0;
    dprint(11, (debugfile,
		"mbx_end_edit_non_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!sessionlock_folder(folder,SESSIONLOCK_REOPEN)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntOpenFolder,
			  "Couldn't open %s for reading!  Edit LOST!"), 
		  folder->cur_folder);
	status = 0;
	goto clean;
    }

    lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmResyncingNewVersion,
		      "Resynchronizing with new version of folder..."));
    status = 1;

 clean:
    dprint(11, (debugfile,
		"mbx_end_edit_non_spool=%d\n",
		status));
    return status;
}

static int mbx_end_edit_spool P_((struct folder_info *folder));
static int mbx_end_edit_spool(folder)
     struct folder_info *folder;
{
    int status = 0;
#ifdef SAVE_GROUP_MAILBOX_ID
    gid_t oldgid = getegid();
#endif
    
    dprint(11, (debugfile,
		"mbx_end_edit_spool: folder=%p (%s)\n",
		folder,folder->cur_folder));

    if (!sessionlock_folder(folder,SESSIONLOCK_REOPEN)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntOpenFolder,
			  "Couldn't open %s for reading!  Edit LOST!"), 
		  folder->cur_folder);
	status = 0;
	goto clean;
    }
    
    if (!lock_folder(OUTGOING,folder)) {
	status = 0;
	goto clean;
    }
    
    if (file_bytes(folder->cur_folder) != 
	folder->mailfile_size) {
	char buffer[1000];
	int len;
	
	/* SIGH.  We've received mail since we invoked the editor
	   on the folder.  We'll have to do some strange stuff to
	   remedy the problem... */
	
	lib_transient(CATGETS(elm_msg_cat, ElmSet, 
			  ElmWarnNewMailRecv,
			  "Warning: new mail received..."));

	if (0 != fseek(folder->p->fh_temp,0,SEEK_END)) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntSeekTemp,
			      "Couldn't seek to end of temp file. Edit LOST!"));
	    status = 0;
	    goto clean;
	}
	
	if (fseek(folder->p->fh_folder, 
		  folder->mailfile_size, 0) == -1) {
	    dprint(1, (debugfile,
		       "Couldn't seek to end of cur_folder (offset %ld) (%s)\n",
		       folder->mailfile_size, "mbx_end_edit_non_spool"));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntSeekEnd,
			      "Couldn't seek to end of folder.  Edit LOST!"));
	    
	    unlock_folder(0,folder);
	    status = 0;
	    goto clean;
	}

	/** Now we can finally stream the new mail into the tempfile **/

	while ((len = mail_gets(buffer, SLEN, 
				folder ->p->fh_folder)) != 0)
	    if (fwrite(buffer, 1, len, folder ->p->fh_temp) != len) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntCopyMailfile,
				  "\nCouldn't copy %s to mail file %s!\n"),
			  folder->cur_folder, 
			  folder->cur_tempfolder);
		unlock_folder(0,folder);
		status = -1;
		goto clean;
	    }

	if (0 != fflush(folder->p->fh_temp)) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntCopyMailfile,
			      "\nCouldn't copy %s to mail file %s!\n"),
		      folder->cur_folder, 
		      folder->cur_tempfolder);
	    unlock_folder(0,folder);
	    status = -1;
	    goto clean;
	}

#ifdef SAVE_GROUP_MAILBOX_ID
	setgid(mailgroupid);
#endif
	/* remove real mail_file and then
	 * link or copy the edited mailfile to real mail_file */

	if (copy_to_folder(folder->p->fh_temp,folder) != 0) {
	    unlock_folder(0,folder);
	    status = -1;
	    goto clean;
	}

	/* restore file permissions before removing lock */
	
	if(restore_file_stats(folder->cur_folder) != 1) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmProblemsRestoringPerms,
			      "Problems restoring permissions of folder %s!"),
		      folder->cur_folder);
	    lib_error(FRM(""));
	}
	lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmChangesIncorporated,
			  "Changes incorporated into new mail..."));
    } else
	lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmResyncingNewVersion,
			  "Resynchronizing with new version of folder..."));

    unlock_folder(0,folder);
    status = 1;

 clean:
#ifdef SAVE_GROUP_MAILBOX_ID
    setgid(oldgid);
#endif

    dprint(11, (debugfile,
		"mbx_end_edit_spool=%d\n",
		status));
    return status;
}

static int mbx_get_no_name_mode P_((struct folder_info *folder));
static int mbx_get_no_name_mode(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_get_no_name_mode=0: folder=%p (%s)\n",
		folder,folder->cur_folder));
    return 0;
}

static int mbx_get_spool_mode P_((struct folder_info *folder));
static int mbx_get_spool_mode(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_get_spool_mode=FOLDER_MBOX|FOLDER_FILE: folder=%p (%s)\n",
		folder,folder->cur_folder));
    return FOLDER_MBOX | FOLDER_FILE;
}

static int mbx_get_non_spool_mode P_((struct folder_info *folder));
static int mbx_get_non_spool_mode(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_get_spool_mode=FOLDER_FILE: folder=%p (%s)\n",
		folder,folder->cur_folder));
    return FOLDER_FILE;
}

static int mbx_get_read_only_mode P_((struct folder_info *folder));
static int mbx_get_read_only_mode(folder)
     struct folder_info *folder;
{
    dprint(11, (debugfile,
		"mbx_get_read_only_mode=FOLDER_RDONLY|FOLDER_FILE: folder=%p (%s)\n",
		folder,folder->cur_folder));
    return FOLDER_RDONLY |FOLDER_FILE;
}


static struct folder_type no_name   = { mbx_close_no_name,   
					mbx_lock_no_name,
                                        mbx_init_null,
					mbx_sessionlock_no_name,
					mbx_unlock_no_name,
					mbx_flush_null,
					mbx_ferror_no_name,
					mbx_prepare_read_no_name,
					mbx_end_read_null,
					mbx_copy_envelope_no_name,
					mbx_is_forwarded_null,
					mbx_copy_header_no_name,
					mbx_copy_body_no_name,
					mbx_copy_envelope_end_no_name,
					mbx_copy_envelope_reset_body_no_name,
					mbx_no_name_to_fd,
					mbx_new_mail_on_no_name,
					mbx_consider_remove_null,
					mbx_prepare_keep_no_name,
					mbx_end_keep_no_name,
					mbx_mark_keep_no_name,
					mbx_no_name_type,
					mbx_start_edit_no_name,
					mbx_end_edit_no_name,
					mbx_free_null,
					mbx_zero_rs_fields_file,
					mbx_free_rs_fields_file,
					mbx_zero_ks_fields_file,
					mbx_free_ks_fields_file,
					mbx_get_no_name_mode};
static struct folder_type read_only = { mbx_close_non_spool,
                                        mbx_lock_non_spool,
                                        mbx_init_non_spool,
					mbx_sessionlock_read_only,
					mbx_unlock_non_spool,
					mbx_flush_null,
					mbx_ferror_non_spool,
					mbx_prepare_read_non_spool,
					mbx_end_read_non_spool,
					mbx_copy_envelope_non_spool,
					mbx_is_forwarded_null,
					mbx_copy_header_non_spool,
					mbx_copy_body_non_spool,
					mbx_copy_envelope_end_non_spool,
					mbx_copy_envelope_reset_body_non_spool,
					mbx_non_spool_to_fd,
					mbx_new_mail_on_normal,
					mbx_consider_remove_null,
					mbx_prepare_keep_no_name,
					mbx_end_keep_no_name,
					mbx_mark_keep_no_name,
					mbx_read_only_type,
					mbx_start_edit_no_name,
					mbx_end_edit_no_name,
					mbx_free_non_spool,
					mbx_zero_rs_fields_file,
					mbx_free_rs_fields_file,
					mbx_zero_ks_fields_file,
					mbx_free_ks_fields_file,
					mbx_get_read_only_mode };
static struct folder_type non_spool = { mbx_close_non_spool,
                                        mbx_lock_non_spool,
                                        mbx_init_non_spool,
					mbx_sessionlock_non_spool,
					mbx_unlock_non_spool,
					mbx_flush_non_spool,
					mbx_ferror_non_spool,
					mbx_prepare_read_non_spool,
					mbx_end_read_non_spool,
					mbx_copy_envelope_non_spool,
					mbx_is_forwarded_null,
					mbx_copy_header_non_spool,
					mbx_copy_body_non_spool,
					mbx_copy_envelope_end_non_spool,
					mbx_copy_envelope_reset_body_non_spool,
					mbx_non_spool_to_fd,
					mbx_new_mail_on_normal,
					mbx_consider_remove_non_spool,
					mbx_prepare_keep_normal,
					mbx_end_keep_non_spool,
					mbx_mark_keep_normal,
					mbx_non_spool_type,
					mbx_start_edit_non_spool,
					mbx_end_edit_non_spool,
					mbx_free_non_spool,
					mbx_zero_rs_fields_file,
					mbx_free_rs_fields_file,
					mbx_zero_ks_fields_file,
					mbx_free_ks_fields_file,
					mbx_get_non_spool_mode };
static struct folder_type spool     = { mbx_close_spool,
                                        mbx_lock_spool,
					mbx_init_spool,
					mbx_sessionlock_spool,
					mbx_unlock_spool,
					mbx_flush_spool,
					mbx_ferror_spool,
					mbx_prepare_read_spool,
					mbx_end_read_spool,
					mbx_copy_envelope_spool,
					mbx_is_forwarded_spool,
					mbx_copy_header_spool,
					mbx_copy_body_spool,
					mbx_copy_envelope_end_spool,
					mbx_copy_envelope_reset_body_spool,
					mbx_spool_to_fd,
					mbx_new_mail_on_normal,
					mbx_consider_remove_null,
					mbx_prepare_keep_normal,
					mbx_end_keep_spool,
					mbx_mark_keep_normal,
					mbx_spool_type,
					mbx_start_edit_spool,
					mbx_end_edit_spool,
					mbx_free_spool,
					mbx_zero_rs_fields_file,
					mbx_free_rs_fields_file,
					mbx_zero_ks_fields_file,
					mbx_free_ks_fields_file,
					mbx_get_spool_mode };

CONST folder_type_p NO_NAME   = &no_name;
CONST folder_type_p NON_SPOOL = &non_spool;	/* mailfile -- not mailbox */
CONST folder_type_p SPOOL     = &spool;		/* normal mailbox */
CONST folder_type_p READ_ONLY = &read_only;	/* read only mailfile */
#ifdef REMOTE_MBX
CONST folder_type_p POP_MBX   = &pop_mbx;	/* POP mailbox */
CONST folder_type_p IMAP_MBX  = &imap_mbx;	/* IMAP mailbox */
#endif


static int valid_magic P_((folder_type_p magic));
static int valid_magic(magic)
     folder_type_p magic;
{
    if (magic == &no_name || 
        magic == &non_spool ||
	magic == &spool ||
#ifdef REMOTE_MBX
	magic == &pop_mbx ||
	magic == &imap_mbx ||
#endif
	magic == &read_only)
	return 1;

    return 0;
}

/* General routines -------------------------------------------------------- */

static folder_type_p get_folder_type P_((CONST char *filename));
static folder_type_p get_folder_type(filename)
     CONST char *filename;
{
    /** returns the type of mailfile filename is
	NO_NAME = no name
	SPOOL = consisting only of mailhome plus base file name
	(no intervening directory name)
	NON_SPOOL = a name that is not SPOOL type above
	READ_ONLY = open this read-only
    **/

    struct stat    buf;		/* stat command  */
    char * p;

    /* if filename is null or is of zero length */
    if((filename == NULL) || (*filename == '\0')) {
	dprint(8, (debugfile, 
		   "get_folder_type=NO_NAME: no filename\n"));
	return(NO_NAME);
    }
    
    dprint(8, (debugfile, 
	       "get_folder_type: filename=%s\n",filename));
    
    /* if filename begins with mailhome,
     * and there is a slash in filename,
     * and there is a filename after it (i.e. last slash is not last char),
     * and the last character of mailhome is last slash in filename,
     * it's a spool file .
     */
    if(is_in_spool(filename)) {
	dprint(8, (debugfile, 
		   "get_folder_type=SPOOL\n"));
	return(SPOOL);
    }

    if (extra_mailbox_dir[0]) {
	CONST char * c = strrchr(filename,'/');
	char * temp;
	int ret;
	if (c)
	    c++;
	else
	    c = filename;
	temp = elm_message(FRM("%s/%s"),extra_mailbox_dir,c);
	ret = same_file(temp,filename);
	free(temp);
	if (ret) {
	    dprint(8, (debugfile, 
		       "get_folder_type=SPOOL (is on %s)\n",
		       extra_mailbox_dir));
	    return(SPOOL);
	}
    }
	
    /* if file name == default mailbox, its a spool file also
     * even if its not in the spool directory. (SVR4)
     */
    if (strcmp(filename, defaultfile) == 0) {
	dprint(8, (debugfile, 
		   "get_folder_type=SPOOL\n"));
	return(SPOOL);
    }
    if (stat(filename, &buf) != 0) {
	int err = errno;
	dprint(8, (debugfile, 
		   "get_folder_type: errno %s attempting to stat file %s\n", 
		   error_description(err), filename));
    } else {
	if (buf.st_mode & 07000) { 
	    /* is 'SPOOL' file if special modes set */
	    dprint(8, (debugfile, 
		       "get_folder_type=SPOOL; mode=%05o\n",buf.st_mode));
	    return(SPOOL);
	}
	/* If owner do not have write permission, open it read-only */
	if (0 == (buf.st_mode & 00200)) {
	    dprint(8, (debugfile, 
		       "get_folder_type=READ_ONLY; mode=%05o\n",buf.st_mode));
	    return(READ_ONLY);
	}
    }
    
    /* Open read only if we have no write access ... 
     * This also detects (at least on Linux) cases where filesystem is
     * mounted read-only
     */
    if (0 != access(filename,WRITE_ACCESS)) {
	int err = errno;

	dprint(8, (debugfile, 
		   "get_folder_type=READ_ONLY; no write access: %s\n",
		   error_description(err)));
	return(READ_ONLY);
    }

    dprint(8, (debugfile, 
	       "get_folder_type=NON_SPOOL\n"));
    return(NON_SPOOL);
}

int same_file(name1,name2)
     char *name1, *name2;
{
    struct stat buf1, buf2;
    
    if (0 == strcmp(name1,name2)) {
	dprint(8, (debugfile, 
		   "same_file(%s,%s)=TRUE\n",name1,name2)); 
	return TRUE;
    }
    
    if (stat(name1, &buf1) != 0) {
	int err = errno;
	dprint(8, (debugfile, 
		   "same_file(%s,%s)=FALSE: %s (errno=%d) attempting to stat file %s\n", 
		   name1,name2,error_description(err), err, name1));
	return FALSE;
    }
    
    if (stat(name2, &buf2) != 0) {
	int err = errno;
	dprint(8, (debugfile, 
		   "same_file(%s,%s)=FALSE: %s (errno=%d) attempting to stat file %s\n", 
		   name1,name2,error_description(err), err, name2));
	return FALSE;
    }
    
    if (buf1.st_ino == buf2.st_ino && buf1.st_dev == buf2.st_dev) {
	dprint(8, (debugfile, 
		   "same_file(%s,%s)=TRUE\n",name1,name2)); 
	return TRUE;
    }
    dprint(8, (debugfile, 
	       "same_file(%s,%s)=FALSE\n",name1,name2)); 
    return FALSE;
}

int is_in_spool(mbox)
     CONST char *mbox;
{
    char * work = safe_strdup(mbox);
    char * ptr = strrchr(work,'/');
    int in_spool = 0;
    
    if (ptr) {
	char temp = *(++ptr);
	*ptr = '\0';
	in_spool = 0 == strcmp(work,mailhome);
	
	dprint(8,(debugfile,"is_in_spool: base=%s\n",
		  work));
	
	*ptr = temp;
    } else
	ptr = work;
    
    dprint(8,(debugfile,"is_in_spool(%s): in_spool=%d\n",
	      work,in_spool));

    if (!in_spool) {
	char * work2 = elm_message(FRM("%s%s"),mailhome,ptr);	    
	if (same_file(work2,work))
	    in_spool = TRUE;
	free(work2);
    }
    dprint(8,(debugfile,"is_in_spool=%d\n",
	      in_spool));
    free(work);
    return in_spool;
}

void close_folder(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{
    dprint(10, (debugfile,
		"close_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));
    
    folder -> folder_type->close_it(folder,mode);

    if (folder -> p) {
	if (folder -> p->fh_folder != NULL)
	    fclose(folder -> p->fh_folder); 
	folder -> p->fh_folder = NULL;

	if (folder -> p->fh_temp != NULL)
	    fclose(folder -> p->fh_temp); 
	folder -> p->fh_temp = NULL;
    }
}


struct folder_info *enter_new_folder(new_file) 
     char *new_file;
{
    struct folder_info *new_folder = NULL;

    dprint(10, (debugfile,
		"enter_new_folder(%s)\n",
		new_file));

    new_folder = safe_malloc(sizeof (struct folder_info));
    new_folder -> p = NULL;
    new_folder -> cur_folder = safe_strdup(new_file);
    new_folder -> cur_tempfolder[0] = '\0';
    new_folder -> mailfile_size     = 0;

    /* determine type of new mailfile and calculate temp file name
       remote_folder_type calls folder's init_it routine if
       succees
    */
    if (!remote_folder_type(new_folder)) {
	new_folder -> folder_type = get_folder_type(new_file);
	new_folder->folder_type->init_it(new_folder);
    }

#if 0
    dprint(1, (debugfile,
	       "New folder %s type %s temp file %s (%s)\n", 
	       current_folder->cur_folder, 
	       (current_folder -> folder_type == SPOOL ? "spool" : "non-spool"),
	       (*current_folder-> cur_tempfolder ? 
		current_folder -> cur_tempfolder : "none"), "newmbox"));
#endif

    dprint(10, (debugfile,
		"enter_new_folder=%p (%s), type=%p\n",
		new_folder,new_folder->cur_folder,new_folder->folder_type));

    return new_folder;
}

void leave_old_folder(folder,mode) 
     struct folder_info *folder;
     enum close_mode mode;
{
    dprint(10, (debugfile,
		"leave_old_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder->folder_type));
    
    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"leave_old_folder",
	      "Bad magic number (folder type)",0);

    close_folder(folder, mode);    
    folder->folder_type->free_it(folder);

    if (folder -> cur_folder)
	free(folder -> cur_folder);
    free(folder);
}

int lock_folder(direction, folder)
     int direction;
     struct folder_info *folder;
{   
    int ret;

    dprint(10, (debugfile,
		"lock_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"lock_folder",
	      "Bad magic number (folder type)",0);


    ret = folder -> folder_type->lock_it(direction,folder);

    dprint(10, (debugfile,
		"lock_folder=%d\n",
		ret));

    return ret;
}

int unlock_folder(interrupt, folder)
     int interrupt;
     struct folder_info *folder;
{   
    int ret;

    /* If interrupt is set, avoid calling no-reentrant routines ... */
    
    /** Flush the mailfile buffer if necessary before removing the lock.
     **/

    dprint(10, (debugfile,
		"unlock_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"unlock_folder",
	      "Bad magic number (folder type)",interrupt);

    if (!interrupt)
	flush_folder(folder);

    ret = folder -> folder_type->unlock_it(interrupt,folder);

    dprint(10, (debugfile,
		"unlock_folder=%d\n",
		ret));
    return ret;
}


/* Return 1 on succeed */
int sessionlock_folder(folder,mode)
     struct folder_info *folder;
     enum sessionlock_mode mode;
{
    int ret;
    dprint(10, (debugfile,
		"sessionlock_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"sessionlock_folder",
	      "Bad magic number (folder type)",0);

    ret = folder->folder_type->sessionlock_it(folder,mode);

    dprint(10, (debugfile,
		"sessionlock_folder=%d\n",
		ret));
    return ret;
}

void flush_folder(folder)
     struct folder_info *folder;
{
    dprint(10, (debugfile,
		"flush_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"flush_folder",
	      "Bad magic number (folder type)",0);

    folder->folder_type->flush_it(folder);
}

int ferror_folder(folder,clean)
     struct folder_info *folder; 
     int clean;
{
    int ret;
    dprint(10, (debugfile,
		"ferror_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"ferror_folder",
	      "Bad magic number (folder type)",0);

    ret = folder->folder_type->ferror_it(folder,clean);
    dprint(10, (debugfile,
		"ferror_folder=%d\n",
		ret));
    return ret;
}

int prepare_read_folder(folder,mode,read_state_ptr)
     struct folder_info *folder;
     enum prepare_mode mode;
     READ_STATE * read_state_ptr;
{
    int status;
    READ_STATE ptr = NULL;
    
    dprint(10, (debugfile,
		"prepare_read_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));
    
    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"prepare_read_folder",
	      "Bad magic number (folder type)",0);

    malloc_read_folder_state(&ptr);
    folder->folder_type->zero_rs_fields_it(ptr);
    ptr->mode = mode;

    status = folder->folder_type->prepare_read_it(folder,mode,ptr);
    if (!status) {
	folder->folder_type->free_rs_fields_it(ptr);
	free_read_folder_state(&ptr);
    }
    *read_state_ptr = ptr;

    dprint(10, (debugfile,
		"prepare_read_folder=%d\n",
		status));

    return status;
}

int end_read_folder(folder,read_state_ptr,silent)
     struct folder_info *folder;
     READ_STATE * read_state_ptr;
     int silent;
{
    int status;
    READ_STATE ptr = *read_state_ptr;

    dprint(10, (debugfile,
		"end_read_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"end_read_folder",
	      "Bad magic number (folder type)",0);

    status = folder->folder_type->end_read_it(folder,ptr,silent);

    folder->folder_type->free_rs_fields_it(ptr);
    free_read_folder_state(&ptr);
    *read_state_ptr = ptr;

    dprint(10, (debugfile,
		"end_read_folder=%d\n",
		status));

    return status;
}

/* Returns:  0 = End of folder
             1 = OK
	    -1 = format error
*/
int copy_envelope_folder(folder,read_state_ptr,entry)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     struct header_rec *entry;
{
    int status;

    dprint(10, (debugfile,
		"copy_envelope_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));
    dprint(10, (debugfile,
		"                     : entry=%p",entry));
    if (!entry->mbx_info) {
	dprint(10, (debugfile,", { mbx_info=NULL }\n"));
    } else {
	dprint(10, (debugfile,", { mbx_info=%p, type=%p }\n",
		    entry->mbx_info,entry->mbx_info->type_code));
    }


    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_envelope_folder",
	      "Bad magic number (folder type)",0);

    status = folder->folder_type->copy_envelope_it(folder,read_state_ptr,
						   entry);
    read_state_ptr -> fbytes_body = read_state_ptr -> fbytes;

    dprint(10, (debugfile,
		"copy_envelope_folder=%d\n",
		status));
    return status;
}

CONST char * is_forwarded_folder(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    CONST char *ret;

    dprint(10, (debugfile,
		"is_forwarded_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"is_forwarded_folder",
	      "Bad magic number (folder type)",0);


    ret=folder->folder_type->is_forwarded_it(folder,read_state_ptr);

    dprint(10, (debugfile,
		"is_forwarded_folder=%s\n",
		ret ? ret : "<NULL>"));
    return ret;
}

long copy_fbytes_folder(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(10, (debugfile,
		"copy_fbytes_folder=%ld: folder=%p (%s), type=%p\n",
		read_state_ptr->fbytes,
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_fbytes_folder",
	      "Bad magic number (folder type)",0);

    return read_state_ptr->fbytes;
}

int copy_lines_folder(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    dprint(10, (debugfile,
		"copy_lines_folder=%d: folder=%p (%s), type=%p\n",
		read_state_ptr -> linecounter,
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_lines_folder",
	      "Bad magic number (folder type)",0);

    return read_state_ptr -> linecounter;
}


/* Gives header including continuation lines */
int copy_header_folder(folder,read_state_ptr,buffer,len)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
{
    int status;
    *buffer = NULL;
    *len = 0;

    dprint(10, (debugfile,
		"copy_header_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_header_folder",
	      "Bad magic number (folder type)",0);

    status = folder->folder_type->copy_header_it(folder,read_state_ptr,
						 buffer,len);

    read_state_ptr -> fbytes_body = read_state_ptr -> fbytes;

    dprint(10, (debugfile,
		"copy_header_folder=%d\n",
		status));
    return status;
}

int copy_body_folder(folder,read_state_ptr,buffer,len,content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     char **buffer; 
     int *len;
     long *content_remaining;
{
    int ret;
    *buffer = NULL;
    *len = 0;

    dprint(10, (debugfile,
		"copy_body_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_body_folder",
	      "Bad magic number (folder type)",0);

    if (*content_remaining == 0L) {
	/* If content remining has gone 0, let
	   let routine copy_envelope_end_folder() 
	   check end of message
	   -1 indicates no content-length given;
	*/
	dprint(10, (debugfile,
		    "copy_body_folder=0 (*content_remaining == 0)\n"));

	return 0;
    }

    ret=folder->folder_type->copy_body_it(folder,read_state_ptr,
					  buffer,len, 
					  content_remaining);
    dprint(10, (debugfile,
		"copy_body_folder=%d\n",
		ret));
    return ret;
}

/* Returns 0 if not end of message (needs resync without content-length) */
int copy_envelope_end_folder(folder,read_state_ptr)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
{
    int ret;
    dprint(10, (debugfile,
		"copy_envelope_end_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_envelope_end_folder",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->copy_envelope_end_it(folder,read_state_ptr);
    dprint(10, (debugfile,
		"copy_envelope_end_folder=%d\n",
		ret));
    return ret;
}

int copy_envelope_reset_body(folder,read_state_ptr,content_remaining)
     struct folder_info *folder;
     READ_STATE read_state_ptr;
     long *content_remaining;
{
    int ret;
    dprint(10, (debugfile,
		"copy_envelope_reset_body: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"copy_envelope_reset_body",
	      "Bad magic number (folder type)",0);

    if (*content_remaining < 0L) {
       	panic("MBX PANIC",__FILE__,__LINE__,"copy_envelope_reset_body",
	      "*content_remaining < 0",0);

    }
    *content_remaining = -1L;
    
    ret=folder->folder_type->copy_envelope_reset_it(folder,read_state_ptr);
    dprint(10, (debugfile,
		"copy_envelope_reset_body=%d\n",
		ret));
    return ret;
}

FILE * folder_to_fd(folder,offset)
     struct folder_info *folder;
     long offset;
{
    FILE *ret;
    dprint(10, (debugfile,
		"folder_to_fd: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"folder_to_fd",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->xxx_to_fd(folder,offset);
    dprint(10, (debugfile,
		"folder_to_fd=%s (%p)\n",
		ret ? "NON-NULL" : "NULL",
		ret));
    return ret;
}

int new_mail_on_folder(folder,bytes)
     struct folder_info *folder;
     int *bytes;
{
    int ret;
    dprint(10, (debugfile,
		"new_mail_on_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"new_mail_on_folder",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->new_mail_on_it(folder,bytes);
    dprint(10, (debugfile,
		"new_mail_on_folder=%d\n",
		ret));
    return ret;
}

int consider_remove_folder(folder)
     struct folder_info *folder; 
{    
    dprint(10, (debugfile,
		"consider_remove_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"consider_remove_folder",
	      "Bad magic number (folder type)",0);
    
    if (!keep_empty_files) {
	if(folder->folder_type->consider_remove_it(folder)) {
	    close_folder(folder,CLOSE_NORMAL);
	    dprint(10, (debugfile,
			"consider_remove_folder=1\n"));
	    return 1;
	}
    }
    dprint(10, (debugfile,
		"consider_remove_folder=0\n"));
    return 0;
}

int prepare_keep_folder(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE * keep_state_ptr;
{
    int status;
    KEEP_STATE ptr = NULL;

    dprint(10, (debugfile,
		"prepare_keep_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"prepare_keep_folder",
	      "Bad magic number (folder type)",0);

    malloc_keep_folder_state(&ptr);
    folder->folder_type->zero_ks_fields_it(ptr);


    status = folder->folder_type->prepare_keep_it(folder,ptr);
   
    if (!status) {
	folder->folder_type->free_ks_fields_it(ptr);
	free_keep_folder_state(&ptr);
    }

    *keep_state_ptr = ptr;

    dprint(10, (debugfile,
		"prepare_keep_folder=%d\n",
		status));
    return status;
}

int end_keep_folder(folder,keep_state_ptr)
     struct folder_info *folder;
     KEEP_STATE * keep_state_ptr;
{
    int status = 0;
    KEEP_STATE ptr = *keep_state_ptr;

    dprint(10, (debugfile,
		"end_keep_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"end_keep_folder",
	      "Bad magic number (folder type)",0);
    
    status = folder->folder_type->end_keep_it(folder,ptr);
    
    folder->folder_type->free_ks_fields_it(ptr);
    free_keep_folder_state(&ptr);
    *keep_state_ptr = ptr;

    dprint(10, (debugfile,
		"end_keep_folder=%d\n",
		status));
    return status;
}

void mark_keep_folder(folder,keep_state_ptr,entry,keep)
     struct folder_info *folder;
     KEEP_STATE keep_state_ptr;
     struct header_rec *entry;
     int keep;
{
    dprint(10, (debugfile,
		"mark_keep_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));
    dprint(10, (debugfile,
		"                : entry=%p",entry));
    if (!entry->mbx_info) {
	dprint(10, (debugfile,", { mbx_info=NULL }\n"));
    } else {
	dprint(10, (debugfile,", { mbx_info=%p, type=%p }\n",
		    entry->mbx_info,entry->mbx_info->type_code));
    }

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"mark_keep_folder",
	      "Bad magic number (folder type)",0);
    
    folder->folder_type->mark_keep_it(folder,keep_state_ptr,entry,keep);
}

CONST char * folder_type(folder)
     struct folder_info *folder;
{
    CONST char *ret;

    dprint(10, (debugfile,
		"folder_type: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"folder_type",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->type(folder);

    dprint(10, (debugfile,
		"folder_type=%s\n",
		ret ? ret : "<NULL>"));
    return ret;
}

int get_folder_mode(folder) 
     struct folder_info *folder;
{
    int ret = 0;
    dprint(10, (debugfile,
		"get_folder_mode: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"get_folder_mode",
	      "Bad magic number (folder type)",0);

    ret = folder->folder_type->get_it_mode(folder);
    
    dprint(10, (debugfile,
		"get_folder_mode=%d\n",ret));
    return ret;
}

int start_edit_folder(folder,buffer)
     struct folder_info *folder; 
     CONST char **buffer;
{
    int ret;
    dprint(10, (debugfile,
		"start_edit_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"start_edit_folder",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->start_edit_it(folder,buffer);
    dprint(10, (debugfile,
		"start_edit_folder=%d\n",
		ret));
    return ret;
}

int end_edit_folder(folder)
     struct folder_info *folder;
{
    int ret;
    dprint(10, (debugfile,
		"end_edit_folder: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"end_edit_folder",
	      "Bad magic number (folder type)",0);

    ret=folder->folder_type->end_edit_it(folder);
    dprint(10, (debugfile,
		"end_edit_folder=%d\n",
		ret));
    return ret;
}

void write_folder_info(F,folder)
     FILE *F; 
     struct folder_info *folder;
{
    dprint(10, (debugfile,
		"write_folder_info: folder=%p (%s), type=%p\n",
		folder,folder->cur_folder,folder -> folder_type));

    if (!valid_magic(folder->folder_type)) 
	panic("MBX PANIC",__FILE__,__LINE__,"write_folder_info",
	      "Bad magic number (folder type)",0);

    fprintf(F, "F%s\n",
	    (folder->folder_type == NON_SPOOL ? 
	     folder->cur_folder : 
	     folder->cur_tempfolder));
}

void change_rec_mbx_info(entry, t)
     struct header_rec *entry;
     info_type_t t;
{
    dprint(14, (debugfile,
		"change_rec_mbx_info: entry=%p, (new)t=%p\n",
		entry,t));
    /* Make new mbx_info */
    if (entry->mbx_info) {
	dprint(14, (debugfile,
		    "                 : mbx_info=%p, type=%p\n",
		    entry->mbx_info,entry->mbx_info->type_code));	
	dprint(14, (debugfile,
		    "change_rec_mbx_info: resetting current mbx_info...\n"));
	entry->mbx_info->type_code->free_it(entry->mbx_info);
    } else {
	dprint(14, (debugfile,
		    "                 : mbx_info=NULL\n"));
    }

    entry->mbx_info = safe_realloc(entry->mbx_info, 
				   sizeof (struct mbx_hdr_info));
    bzero((void *)entry->mbx_info,sizeof (struct mbx_hdr_info));
    entry->mbx_info->type_code = t;
    entry->mbx_info->type_code->zero_it(entry->mbx_info);
}

void free_rec_mbx_info(entry)
     struct header_rec *entry;
{
    dprint(10, (debugfile,
		"free_rec_mbx_info: entry=%p\n",
		entry));

    if (!entry->mbx_info) {
	dprint(10, (debugfile,
		    "                 : mbx_info=NULL\n"));
	return;
    }
    if (entry->mbx_info) {
	dprint(10, (debugfile,
		    "                 : mbx_info=%p, type=%p\n",
		    entry->mbx_info,entry->mbx_info->type_code));
	entry->mbx_info->type_code->free_it(entry->mbx_info);
	free(entry->mbx_info);
	entry->mbx_info = NULL;
    }
}

void unfold_header(buffer,len,current_header)
     char *buffer; 
     int *len; 
     struct header_rec *current_header;
{
    char *c, *p;
    enum tag { t_NORMAL, t_CR, t_LF } tag = t_NORMAL;
    
    dprint(95, (debugfile,
		"fold_header- len=%d,buffer=%.*s",*len,*len,buffer));
    if (*len < 1 || !(buffer) || (buffer)[*len -1] != '\n') {
	dprint(95, (debugfile,
		    "\nfold_header- NO NEWLINE\n"));
    } 

    /* Unfold and remove newline ... */
    for (c = buffer, p = buffer; c < buffer + *len; c++) {
	switch(*c) {
	case '\n':
	    if (tag != t_CR && current_header && current_header->binary) {
		current_header->binary = 0;
		dprint(12,(debugfile,"-- Not a binary message\n"));
	    }			
	    tag = t_LF;
	    break;
	case '\r':
	    tag = t_CR;
	    break;
	case ' ':
	case '\t':
	    if (tag == t_LF) {
		*(p++) = ' ';
		tag = t_NORMAL;
		break;
	    }
	default:
	    *(p++) = *c;
	case '\0':
	    tag = t_NORMAL;
	}
    }
    *p = '\0';
    
    *len = p-buffer;
    dprint(95, (debugfile,
		"fold_header: len=%d,buffer=%.*s\n",*len,*len,buffer));

}

void return_path_to_env_from(entry,value)
     struct header_rec *entry;
     CONST char *value;
{
    dprint(24, (debugfile,
		"return_path_to_env_from- value=%s\n",value));
    dprint(24, (debugfile, "env_from_source=%d\n",env_from_source));

    /* env_from_source:      
       0 == forward-from,
       1 == from,
       2 == return-path
    */
    
    if (env_from_source <3) {
	char ** tokens = rfc822_tokenize(value);

	remove_space_tokenized(tokens);

	if (tokens[0] && 0 == strcmp(tokens[0],"<")) {
	    int start = 1,i;
	    
	    /* Strip source path */
	    for (i = 1; tokens[i]; i++) {
		if (0 == strcmp(":",tokens[i]))
		    start = i+1;
		if (0 == strcmp(">",tokens[i]))
		    break;
	    }

	    if (!tokens[i]) {
		dprint(24, (debugfile, 
			   "Bad Return-Path: %s\n",value));
	    } else {
		char * temp = safe_strdup("");
		int j;
		
		if (0 != strcmp(tokens[1],"@"))
		    start = 1;
		else if (start > 1) {
		    dprint(24, (debugfile, "  (stripping source path)\n"));
		}

		for (j = start; j < i; j++)
		    temp = strmcat(temp,tokens[j]);
		
		strfcpy(entry->env_from, temp, sizeof(entry->env_from));
		dprint(24, (debugfile, "  user=%s\n", temp));
		free(temp);
	    }

	} else {
	    dprint(24, (debugfile, 
			"Bad Return-Path: %s\n",value));
	}
	free_rfc822tokenized(tokens);
    }    
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


