static char rcsid[] = "@(#)$Id: okay_addr.c,v 1.13 2000/06/14 15:40:28 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.13 $   $State: Exp $
 *
 *  Reimplemented by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  This code is reimplemented. Original code was following copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** 

**/

#include "headers.h"

int addr_is_user(addr)
     char *addr;
{
    struct addr_rec  *alternatives;
    char our_address[SLEN]; 
    char *host_part;
    
    if (0 == strcmp(username, addr)) {
	dprint(25,(debugfile,"addr_is_user(\"%s\")=TRUE: matches username\n",
		   addr));
	return TRUE;
    }

    /* UUCP hostnames are case sensitive */
    elm_sfprintf(our_address, sizeof our_address,
		 FRM("%s!%s"), hostname, username);
    if (0 == strcmp(our_address, addr)) {
	dprint(25,(debugfile,
		   "addr_is_user(\"%s\")=TRUE: matches UUCP address\n",
		   addr));
	return TRUE;
    }

    elm_sfprintf(our_address, sizeof our_address,
		 FRM("%s!%s"), hostfullname, username);
    if (0 == strcmp(our_address, addr)) {
	dprint(25,(debugfile,
		   "addr_is_user(\"%s\")=TRUE: matches UUCP address\n",
		   addr));
	return TRUE;
    }
    
    /* Internet hostnames are case insensitive */
    host_part = qstrpbrk(addr,"@");
    if (host_part) {
	char * buffer = safe_malloc(host_part - addr +1);
	strncpy(buffer,addr,host_part - addr);
	buffer[host_part - addr] = '\0';
	
	dprint(25,(debugfile,
		   "addr_is_user(\"%s\"): user part=%s, host part=%s\n",
		   addr,buffer,host_part+1));
    
	if (0 == strcmp(buffer,username)) {
	    if (0 == istrcmp(hostname,host_part+1)) {
		dprint(25,(debugfile,
			   "addr_is_user(\"%s\")=TRUE: matches internet address\n",
			   addr));
		free(buffer);
		return(TRUE);
	    }
      
	    if (0 == istrcmp(hostfullname,host_part+1)) {
		dprint(25,(debugfile,
			   "addr_is_user(\"%s\")=TRUE: matches internet address\n",
			   addr));		
		free(buffer);
		return(TRUE);
	    }
	}

	alternatives = alternative_addresses;  
	while (alternatives != NULL) {
	    char * host_part2 = qstrpbrk(alternatives->address,"@");
	    if (host_part2) {
		char * buffer2 = 
		    safe_malloc(host_part2 - alternatives->address +1);
		strncpy(buffer2,alternatives->address,
			host_part2 - alternatives->address);
		buffer2[host_part2 - alternatives->address] = '\0';
		
		dprint(27,(debugfile,
			   "addr_is_user: %s: user part=%s, host part=%s\n",
			   alternatives->address,buffer2,host_part2+1));
		
		if (0 == strcmp(buffer,buffer2) &&
		    0 == istrcmp(host_part2+1,host_part+1)) {
		    dprint(25,(debugfile,
			       "addr_is_user(\"%s\")=TRUE: matches alternative\n",
			       addr));		
		    free(buffer2);
		    free(buffer);		    
		    return(TRUE);		    
		}
		free(buffer2);
	    }
	    alternatives = alternatives->next;
	}
	free(buffer);
    }
    
    alternatives = alternative_addresses;  
    while (alternatives != NULL) {
	if (0 == strcmp(addr,alternatives->address)) {
	    dprint(25,(debugfile,
		       "addr_is_user(\"%s\")=TRUE: matches alternative address\n",
		       addr));		
	    return(TRUE);
	}
	alternatives = alternatives->next;
    }
    dprint(25,(debugfile,
	       "addr_is_user(\"%s\")=FALSE\n",addr));
    return FALSE;
}

int okay_address_l(address, return_address)
     CONST struct addr_item *address, *return_address;
{
    
    /** This routine checks to ensure that the address we just got
	from the "To:" or "Cc:" line isn't us AND isn't the person	
	who sent the message.  Returns true iff neither is the case **/
    
    CONST struct addr_item *ptr;
    
    for (ptr = return_address; ptr && ptr -> addr && ptr->fullname; ptr++) {
	if (0 == strcmp(ptr -> addr, address -> addr)) {
	    dprint(25,(debugfile,"okay_address_l=FALSE:\n"));
	    if (debug >= 25 && debugfile)
		elm_fprintf(debugfile,
			    FRM("     address{addr=%s,fullname=%S} same as return_address {addr=%s,fullname=%S}\n"),
			    address -> addr, address -> fullname,
			    ptr -> addr, ptr -> fullname);
	    return(FALSE);
	}
    }

    if (addr_is_user(address -> addr)) {
	dprint(25,(debugfile,"okay_address_l=FALSE:\n"));
	if (debug >= 25 && debugfile)
	    elm_fprintf(debugfile,
			FRM("     address{addr=%s,fullname=%S} is user's address\n"),
			address -> addr, address -> fullname);	
	return(FALSE);
    }
    
    dprint(25,(debugfile,"okay_address_l=TRUE:\n"));
    if (debug >= 25 && debugfile)
	elm_fprintf(debugfile,
		    FRM("     address{addr=%s,fullname=%S} is not our address.\n"),
		    address -> addr, address -> fullname); 
    return(TRUE);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
