static char rcsid[] = "@(#)$Id: output.c,v 1.13 2000/07/30 15:12:11 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.13 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}
static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}



static int def_err_handler P_((CONST char *str));
static int def_err_handler (str) 
     CONST char *str;
{
  int ret;
  int l = strlen(str);
  ret = fprintf(stderr,"%s",str);
  if (ret > 0 && l && str[l-1] != '\n')
    ret += fprintf(stderr,"\n");
  return ret;
}

static err_handler *H = def_err_handler;
extern void set_error_handler (h)
     err_handler *h;
{
    H = h;
}

static err_handler *T = def_err_handler;
void set_transient_handler(h)
     err_handler *h;
{
    T = h;
}

static char * def_prompt_handler P_((const char *str, int pass));
static char * def_prompt_handler (str,pass) 
     CONST char *str;
     int pass;
{
  int l = strlen(str);
  char buffer[200], *x;
  
  FILE * f = fopen("/dev/tty","r+");
  if (!f)
      return NULL;
  if (pass)
      x = getpass(str);
  else {
      int ret = fprintf(f,"%s",str);
      char *y;
      fflush(f);
      x = fgets(buffer,sizeof buffer,f);
      if (x && (y = strchr(x,'\n')))
	  *y = '\0';
      if (ret > 0 && l && str[l-1] != '\n')
	  ret += fprintf(stderr,"\n");
  }
  fclose(f);
  if (!x)
      return NULL;
  return safe_strdup(x);
}

static prompt_handler *P = def_prompt_handler;
void set_prompt_handler(h)
     prompt_handler *h;
{
    P = h;
}

char *elm_message P_((const char * format, const char *msg, ...));
int elm_sfprintf P_((char *buffer, int size,
		    const char * format, const char *msg, ...));

struct string * elm_smessage P_((int max_alloc,
				 const char *format, const char *msg,
				 va_list args));

struct string * elm_smessage(max_alloc,format,msg,args)
     int max_alloc;
     CONST char *format;
     CONST char *msg;
     va_list args;
{
    CONST char * s;
    struct string * store1 = NULL, * msg1 = NULL;
    charset_t msg_charset  = display_charset;

    struct string ** store2 = NULL;
    int store2_len = 0,i,pos;

  
    dprint(82,(debugfile,"elm_smessage: max_alloc=%d, format='%s', msg='%s'\n",
	       max_alloc,format,msg));
    
#define INC(p) { (p)++; if (!*(p)) break; }

    for (s = format; *s; s++) {
	int fill = ' ';
	int left = 0;
	int plus = 0;
	int val1 = 0, val2 = 0;
	int long_f = 0;
	struct string * param_value = NULL;
	
	if (*s != '%') {
	    continue;
	}
	INC(s);
	
	if ('%' == *s) {
	    continue;
	}
	
	
	if ('0' == *s)      { fill = '0'; INC(s); }
	else if ('-' == *s) { left = 1;   INC(s); }
	else if ('+' == *s) { plus = 1;   INC(s); }
	
	if (*s >= '1' && *s <= '9') {
	    while (*s >= '0' && *s <= '9') {
		val1 = val1 * 10 + *s - '0';
		INC(s);
	    }
	    
	    if (*s == '$') {
		val1 = 0;
		INC(s);
		dprint(82,(debugfile,
			   " -- Positional args are not allowed in: %s\n",
			   format));
		goto error1;
	    }
	    
	} else if ('*' == *s) { 
	    val1 = va_arg(args,int);
	    INC(s);
	}
	
	if ('.' == *s) {
	    INC(s);
	    if (*s >= '1' && *s <= '9') {
		while (*s >= '0' && *s <= '9') {
		    val2 = val2 * 10 + *s - '0';
		    INC(s);
		}
	    } else if ('*' == *s) { 
		val2 = va_arg(args,int);
		INC(s);
	    }      
	}
	
	if ('l' == *s) { 
	    long_f = 1;
	    INC(s);
	}
	
	
#define MAX_WIDTH 64
	switch(*s) {
	    CONST char *str, *a;
	    int base;
	    char *seq;
	    
	    int l,c;
	    long val,valz;
	    unsigned long r;
	    unsigned long r1;      
	    int quote,X;
	    struct string * S, *S2, *S3;
	    unsigned char width[MAX_WIDTH+1];
	    char * stemp, *sptr;
	    struct charset_state * ch;
	    int I;
	    
	case 'c':
	    c = va_arg(args, int);    
	    param_value = new_string(display_charset);
	    
	    dprint(83,(debugfile,
		       "-- [c] val1=%d, c=%c, param_value=%p\n",
		       val1,c,param_value));
	    l = 1;
	    if (l < val1 && !left) { 
		fill_ascii_to_string(param_value,val1-l,fill);
		l += val1-l; 
	    }
	    add_streambyte_to_string(param_value,c);
	    if (l < val1 && left) { 
		fill_ascii_to_string(param_value,val1-l,' ');
		l += val1-l; 
	    }
	    break;
	case 'C':
	    ch =  va_arg(args, struct charset_state *);
	    param_value = new_string(ch->charset);

	    dprint(83,(debugfile,
		       "-- [C] val1=%d, param_value=%p\n",
		       val1,param_value));

	    l = 1;
	    if (l < val1 && !left) { 
		fill_ascii_to_string(param_value,val1-l,fill);
		l += val1-l; 
	    }
	    add_state_to_string(param_value,ch);
	    if (l < val1 && left) { 
		fill_ascii_to_string(param_value,val1-l,' ');
		l += val1-l; 
	    }
	    break;

	case 'S':
	    S = va_arg(args, struct string *);
	    if (!verify_string(S)) {
	        dprint(1,(debugfile,"elm_smessage: BAD %%S argument\n"));
		goto error1;
	    }

	    l = string_len(S);
	    if (val2 <= 0) val2 = l;
	    
	    X = 0;
	    if (max_alloc > 0 && val2 > max_alloc) {
		dprint(82,(debugfile,
			   "--  adjusting val2 =%d -> %d (max_alloc)\n",
			   val2,max_alloc));
		val2 = max_alloc;
	    }
	    S2  = clip_from_string(S,&X,val2);
	    
	    dprint(83,(debugfile,
		       "-- [%c] val1=%d, S=%p, val2=%d,(clip)S2=%p\n",
		       *s,val1,S,val2,S2));
	    S3 = new_string(S->string_type);
	    
	    if (l < val1 && !left) { 
		fill_ascii_to_string(S3,val1-l,fill);
		l += val1-l; 
	    }
	    
	    param_value = cat_strings(S3,S2,0);
	    dprint(83,(debugfile,
		       "--   param_value=%p\n",param_value)); 
	    
	    free_string(&S2);
	    free_string(&S3);
	    
	    if (l < val1 && left) { 
		fill_ascii_to_string(param_value,val1-l,' ');
		l += val1-l; 
	    }
	    break;
	    
	case 's': case 'Q':
	    str = va_arg(args, char *);
	    if (!str)
		goto error1;

	    quote = ('Q' == *s);
	    
	    l = strlen(str);
	    if (quote) {
		l += 2;
		for (a=str; *a; a++) {
		    if (*a == '\\' || *a == '"')
			l++;
		}
	    }

	    if (val2 <= 0) val2 = l;
	    if (quote)
		val2 -= 2;
	    if (max_alloc > 0 && val2 > max_alloc) {
		dprint(82,(debugfile,
			   "--  adjusting val2 =%d -> %d (max_alloc)\n",
			   val2,max_alloc));
		val2 = max_alloc;
	    }

	    stemp = safe_malloc(l+2);

	    sptr = stemp;
	    for (a = str; 
		 *a && a - str < val2 && sptr - stemp < l; a++) { 
		if (quote && (*a == '\\' || *a == '"')) {
		    *sptr++ = '\\'; 
		}
		*sptr++ = *a;
	    }
	    *sptr = '\0';
	    S2 = new_string2(display_charset,s2us(stemp));
  
	    dprint(83,(debugfile,
		       "-- [%c] val1=%d, val2=%d, str=%s, ",
		       *s,val1,val2,str));
	    dprint(83,(debugfile,
		       "S2=%p, stemp=%s\n",S2,stemp));
	    free(stemp); stemp = NULL;

	    S3 = new_string(display_charset);	    

	    if (l < val1 && !left) { 
		fill_ascii_to_string(S3,val1-l,fill);
		l += val1-l; 
	    }

	    if (quote) {
		add_ascii_to_string(S3,s2us("\"")); 
	    }

	    param_value = cat_strings(S3,S2,0);
	    dprint(83,(debugfile,
		       "--   param_value=%p\n",param_value)); 

	    free_string(&S2);
	    free_string(&S3);

	    if (quote) {
		add_ascii_to_string(param_value,s2us("\"")); 
	    }

	    if (l < val1 && left) { 
		fill_ascii_to_string(param_value,val1-l,' ');
		l += val1-l; 
	    }
	    break;
	    
	case 'd': case 'x': case 'X':
	    switch (*s) {
	    case 'd': seq = "0123456789"; base = 10; break;
	    case 'x': seq = "0123456789abcdef"; base = 16; break;
	    case 'X': seq = "0123456789ABCDEF"; base = 16; break;
	    }
	    if (long_f)
		val = va_arg(args, long);
	    else
		val = va_arg(args, int);
	    c = 0;
	    if (val < 0) {
		c = '-'; 
		val = -val;
		val1--;
	    } else if (plus) {
		c = '+'; 
		val1--;
	    }
	    param_value = new_string(display_charset);
	    
	    dprint(83,(debugfile,
		       "-- [d] val1=%d, val2=%d, (sign)c=%c, val=%ld, base=%d, seq=%s, param_value=%p\n",
		       val1,val2,c, val,base,seq,param_value));
	    l = 0;
	    valz = val;
	    r1 = 1;
	    do {
		r = (valz / base);
		valz = r;
		l++;
		r1 *= (long) base;
	    } while(r && l <MAX_WIDTH);
	    
	    if (l < val1 && !left &&  fill != '0') { 
		fill_ascii_to_string(param_value,val1-l,fill);
		l += val1-l; 
	    }	
	    if (c) { /* sign */
		fill_ascii_to_string(param_value,1,c);
	    }
	    if (l < val1 && !left &&  fill == '0') { 
		fill_ascii_to_string(param_value,val1-l,fill);
		l += val1-l; 
	    }	
	    
	    I = 0;
	    valz = val;
	    while(r1 >= (long) base && I < MAX_WIDTH) {
		r1 /= (long) base;
		r = (valz / r1);
		width[I++] = seq[r];
		valz -= r * r1;
	    } 
	    width[I] = '\0';
	    add_ascii_to_string(param_value,width);
	    if (l < val1 && left) { 
		fill_ascii_to_string(param_value,val1-l,' ');
		l += val1-l; 
	    }
	    break;
	default:
	    dprint(1,(debugfile,"elm_smessage: BAD [%c] ????\n",*s));
	error1:
	    param_value = new_string(display_charset);
	    add_ascii_to_string(param_value,s2us("[?-ERROR-?]"));
	    dprint(83,(debugfile,
		       " --- error, param_value=%p\n",
		       param_value));
	}
	store2 = safe_realloc(store2, (store2_len+1) * 
			      sizeof (struct string *)); 
	store2[store2_len++] = param_value; 
	dprint(83,(debugfile," -- adding param[%d] = %p\n", 
		   store2_len-1,store2[store2_len-1])); 
    }

    for (i = 0; i < store2_len; i++) {
	dprint(83,(debugfile,"  - %d: %p\n",
		   i,store2[i]));
    }

    pos = -1;
    /* Actual printing */
    for (s = msg; *s; s++) {
	int val1 = 0;
	if (*s != '%') {
	add_character:                 /* Jump position for %% */
	    
	    if (!msg1) {
		msg1 = new_string(msg_charset);
		dprint(83,(debugfile,
			   "** msg_charset={'%s'; type=%p }, msg1=%p\n",
			   msg_charset->MIME_name ? msg_charset->MIME_name : "<none>",
			   msg_charset->charset_type,
			   msg1));		
	    }
	    
	    add_streambyte_to_string(msg1,*s);
	    continue;
	}
	INC(s);
	
	if ('%' == *s) 
	    goto add_character;         /* Back jump */
	
	pos++;
	
	if (msg1) {
	    if (!store1)
		store1 = msg1;
	    else {
		struct string * temp = cat_strings(store1,msg1,0);
		
		free_string(&store1);
		free_string(&msg1);
		store1 = temp;
	    }
	    msg1 = NULL;
	}

    again2:
	if ('0' == *s)      { INC(s); }
	else if ('-' == *s) { INC(s); }
	else if ('+' == *s) { INC(s); }
	
	if (*s >= '1' && *s <= '9') {
	    while (*s >= '0' && *s <= '9') {
		val1 = val1 * 10 + *s - '0';
		INC(s);
	    }
	    if ('$' == *s) {
		pos = val1;
		val1 = 0;
		goto again2;
	    }
	} else if ('*' == *s) { 
	    INC(s);
	}
	
	if ('.' == *s) {
	    INC(s);
	    if (*s >= '1' && *s <= '9') {
		while (*s >= '0' && *s <= '9') {
		    INC(s);
		}
	    } else if ('*' == *s) { 
		INC(s);
	    }      
	}
	
	if ('l' == *s) { 
	    INC(s);
	}
	
	switch(*s) {
	    CONST struct string *str;
	case 's': case 'Q': case 'd': case 'c': case 'C':
	case 'x': case 'X': case 'S':
	    if (pos < 0 || pos >= store2_len)
		goto error2;
	    str = store2[pos];
	    if (!str)
		goto error2;
	    dprint(83,(debugfile,"-- [%c] str=%p\n",
		       *s,str));
	    
	    if (!store1)
		store1 = dup_string(str);
	    else {
		struct string * temp = cat_strings(store1,str,0);
	    
		free_string(&store1);
		store1 = temp;
	    }
	    break;	
	default:
	    dprint(82,(debugfile,"-- [%c] ????\n",*s));
	error2:
	    dprint(82,(debugfile,"--- error\n"));
	    
	    if (!store1)
		store1 = new_string(display_charset);
	    add_ascii_to_string(store1,s2us("[???]"));
	}
    }

    for (i = 0; i < store2_len; i++) {
	if (store2[i])
	    free_string(&store2[i]);
    }
    if (store2)
	free(store2);
    
    if (msg1) {
	if (!store1)
	    store1 = msg1;
	else {
	    struct string * temp = cat_strings(store1,msg1,0);
	    
	    free_string(&store1);
	    free_string(&msg1);
	    store1 = temp;
	}
	msg1 = NULL;
    }
    if (!store1)
	store1 = new_string(display_charset);

#undef INC

    return store1;
}


char *elm_vmessage P_((int max_alloc,
		       const char *format, const char *msg, 
		       va_list args));


char *elm_vmessage (max_alloc,format,msg,vl)
     int max_alloc;
     CONST char *format;
     CONST char *msg;
     va_list vl;
{
    char * store1 = NULL;

    if (display_charset &&
	0   /* Alternate implementation used for testing... */
	) {

	struct string * R = elm_smessage(max_alloc,format,msg,vl);
	struct string * R1 = convert_string(display_charset,R,1);
	
	store1 = us2s(stream_from_string(R1,0,NULL));
	
	free_string(&R1);
	free_string(&R);
    } else {

	CONST char * s;
	int ptr1 = 0, alloced1 = 0;
	
	struct store2 {
	    char * store;
	    int ptr;
	    int alloced;
	} * store2 = NULL;
	int store2_len = 0;
	int i,pos;
	
	
	dprint(82,(debugfile,"elm_vmessage: max_alloc=%d, format=%s\n",
		   max_alloc,format));
	dprint(82,(debugfile,"              msg=%s\n",msg));
	
	
#define ENLARGE(ptr,alloced,S) do { \
    if (ptr >= alloced) { \
    dprint(99,(debugfile,"ENLARGE(%d,%d,..) -- reallocing\n",ptr,alloced)); \
    S = safe_realloc(S,alloced+10); alloced += 10; } } while(0)
#define PUTC1(c) do { ENLARGE(ptr1,alloced1,store1); store1[ptr1++] = c; } \
  while(0)
#define PUTC2(c) do { ENLARGE(store2[store2_len-1].ptr,\
   store2[store2_len-1].alloced,store2[store2_len-1].store); \
   store2[store2_len-1].store[store2[store2_len-1].ptr++] = c; } \
while(0)
#define NEW_STORE do { \
  dprint(99,(debugfile,"NEW_STORE -- %d\n",store2_len)); \
  store2 = safe_realloc(store2, \
  (store2_len+1) * sizeof (struct store2)); store2_len++; \
  store2[store2_len-1].store = NULL; store2[store2_len-1].alloced = 0; \
  store2[store2_len-1].ptr = 0; } while(0);
#define INC(p) { (p)++; if (!*(p)) break; }
	
	for (s = format; *s; s++) {
	    int fill = ' ';
	    int left = 0;
	    int plus = 0;
	    int val1 = 0, val2 = 0;
	    int long_f = 0;
	    if (*s != '%') {
		continue;
	    }
	    INC(s);
	    
	    if ('%' == *s) {
		continue;
	    }
	    NEW_STORE;
	    
	    
	    if ('0' == *s)      { fill = '0'; INC(s); }
	    else if ('-' == *s) { left = 1;   INC(s); }
	    else if ('+' == *s) { plus = 1;   INC(s); }
	    
	    if (*s >= '1' && *s <= '9') {
		while (*s >= '0' && *s <= '9') {
		    val1 = val1 * 10 + *s - '0';
		    INC(s);
		}
		
		if (*s == '$') {
		    val1 = 0;
		    INC(s);
		    dprint(82,(debugfile,
			       "-- Positional args are not allowed in: %s\n",
			       format));
		    goto error1;
		}
		
	    } else if ('*' == *s) { 
		val1 = va_arg(vl,int);
		INC(s);
	    }
	    
	    if ('.' == *s) {
		INC(s);
		if (*s >= '1' && *s <= '9') {
		    while (*s >= '0' && *s <= '9') {
			val2 = val2 * 10 + *s - '0';
			INC(s);
		    }
		} else if ('*' == *s) { 
		    val2 = va_arg(vl,int);
		    INC(s);
		}      
	    }
	    
	    if ('l' == *s) { 
		long_f = 1;
		INC(s);
	    }
	    
	    switch(*s) {
		CONST char *str, *a;
		int base;
		char *seq;
		
		int l,c;
		long val,valz;
		unsigned long r;
		unsigned long r1;      
		int quote,X;
		struct string * S, *S1;
		char * temp;
		struct charset_state *ch;
		
	    case 'c':
		c = va_arg(vl, int);      
		dprint(83,(debugfile,"-- [c] val1=%d, c=%c\n",
			   val1,c));
		l = 1;
		while (l < val1 && !left) { PUTC2(fill); l++; }
		PUTC2(c);
		while (l < val1 && left) { PUTC2(' '); l++; }
		break;
	    case 'C':
		ch = va_arg(vl, struct charset_state *);
		S = new_string(display_charset);
		add_state_to_string(S,ch);
		
		temp = us2s(stream_from_string(S,1,NULL));
		dprint(83,(debugfile,"-- [%c] val1=%d, str=%s\n",
			   *s,val1,temp));

		l = 1;
		while (l < val1 && !left) { PUTC2(fill); l++; }
		for (a=temp; *a; a++) {
		    PUTC2(*a); 
		}
		while (l < val1 && left) { PUTC2(' '); l++; }
		free(temp);
		free_string(&S);
		break;

	    case 'S':
		S = va_arg(vl, struct string *);
		if (!verify_string(S)) {
		    dprint(1,(debugfile,"elm_vmessage: BAD %%S argument\n"));
		    goto error1;
	        }

		S1 = convert_string(display_charset,S,1);
		
		l = string_len(S1);
		if (val2 <= 0) val2 = l;
		
		X = 0;
		temp = us2s(streamclip_from_string(S1,&X,val2,NULL));
				
		dprint(83,(debugfile,"-- [%c] val1=%d, val2=%d,(clip)str=%s\n",
			   *s,val1,val2,temp));
		while (l < val1 && !left) { PUTC2(fill); l++; }

		for (a=temp; *a; a++) {

		    PUTC2(*a); 
		    if (max_alloc > 0 && a - temp > max_alloc) {
			dprint(1,(debugfile,
				  " --- too long string (max=%d): %.30s...\n",
				  max_alloc,temp));
			break;
		    }
		}
		
		free(temp);
		free_string(&S1);
		
		while (l < val1 && left) { PUTC2(' '); l++; }
		
		break;

	    case 's': case 'Q':
		str = va_arg(vl, char *);
		if (!str)
		    goto error1;
		
		quote = ('Q' == *s);
		
		l = strlen(str);
		if (quote) {
		    l += 2;
		    for (a=str; *a; a++) {
			if (*a == '\\' || *a == '"')
			    l++;
		    }
		}
		if (val2 <= 0) val2 = l;
		if (quote)
		    val2 -= 2;
		dprint(83,(debugfile,"-- [%c] val1=%d, val2=%d, str=%s\n",
			   *s,val1,val2,str));
		while (l < val1 && !left) { PUTC2(fill); l++; }
		
		if (quote) {
		    PUTC2('"'); 
		}
		for (a = str; *a && a - str < val2; a++) { 
		    if (quote && (*a == '\\' || *a == '"')) {
			PUTC2('\\'); 
		    }
		    PUTC2(*a); 
		    if (max_alloc > 0 && a - str > max_alloc) {
			dprint(1,(debugfile,
				  "--- too long string (max=%d): %.30s...\n",
				  max_alloc,str));
			break;
		    }
		}
		if (quote) {
		    PUTC2('"'); 
		}
		while (l < val1 && left) { PUTC2(' '); l++; }
		break;
	    case 'd': case 'x': case 'X':
		switch (*s) {
		case 'd': seq = "0123456789"; base = 10; break;
		case 'x': seq = "0123456789abcdef"; base = 16; break;
		case 'X': seq = "0123456789ABCDEF"; base = 16; break;
		}
		if (long_f)
		    val = va_arg(vl, long);
		else
		    val = va_arg(vl, int);
		c = 0;
		if (val < 0) {
		    c = '-'; val1--;
		    val = -val;
		} else if (plus) {
		    c = '+'; val1--;
		}
		dprint(83,(debugfile,"-- [d] val1=%d, val2=%d, c=%c, val=%ld, base=%d, seq=%s\n",
			   val1,val2,c, val,base,seq));
		l = 0;
		valz = val;
		r1 = 1;
		do {
		    r = (valz / base);
		    valz = r;
		    l++;
		    r1 *= (long) base;
		} while(r);
		while (l < val1 && !left && fill != '0') { PUTC2(fill); l++; }
		if (c) { PUTC2(c); }
		while (l < val1 && !left && fill == '0') { PUTC2('0');  l++; }
		valz = val;
		while(r1 >= (long) base) {
		    r1 /= (long) base;
		    r = (valz / r1);
		    PUTC2( seq[r]);
		    valz -= r * r1;
		} 
		while (l < val1 && left) { PUTC2(' '); l++; }
		break;
	    default:
		dprint(1,(debugfile,"elm_vmessage: BAD [%c] ????\n",*s));
	    error1:
		dprint(83,(debugfile,"-- error\n"));
	    }
	    PUTC2('\0');
	}
	
	for (i = 0; i < store2_len; i++) {
	    dprint(83,(debugfile," -- %d: %.*s\n",i,
		       store2[i].ptr,NONULL(store2[i].store)));
	}
	
	pos = -1;
	/* Actual printing */
	for (s = msg; *s; s++) {
	    int val1 = 0;
	    if (*s != '%') {
		PUTC1(*s);
		continue;
	    }
	    INC(s);
	    
	    if ('%' == *s) {
		PUTC1(*s);
		continue;
	    }
	    pos++;
	    
	again2:
	    if ('0' == *s)      { INC(s); }
	    else if ('-' == *s) { INC(s); }
	    else if ('+' == *s) { INC(s); }
	    
	    if (*s >= '1' && *s <= '9') {
		while (*s >= '0' && *s <= '9') {
		    val1 = val1 * 10 + *s - '0';
		    INC(s);
		}
		if ('$' == *s) {
		    pos = val1;
		    val1 = 0;
		    goto again2;
		}
	    } else if ('*' == *s) { 
		INC(s);
	    }
	    
	    if ('.' == *s) {
		INC(s);
		if (*s >= '1' && *s <= '9') {
		    while (*s >= '0' && *s <= '9') {
			INC(s);
		    }
		} else if ('*' == *s) { 
		    INC(s);
		}      
	    }
	    
	    if ('l' == *s) { 
		INC(s);
	    }
	    
	    switch(*s) {
		CONST char *str,*a;
	    case 's': case 'Q': case 'd': case 'c': case 'C': case 'x': 
	    case 'X': case 'S':
		if (pos < 0 || pos >= store2_len)
		    goto error2;
		str = store2[pos].store;
		if (!str)
		    goto error2;
		dprint(83,(debugfile,"-- [%c] str=%s\n",*s,str));
		for (a = str; *a; a++) { 
		    PUTC1(*a); 
		}
		break;
	    default:
		dprint(82,(debugfile,"-- [%c] ????\n",*s));
	    error2:
		dprint(83,(debugfile,"--- error\n"));
		PUTC1('[');
		PUTC1('?');
		PUTC1(']');
	    }
	}
	
	PUTC1('\0');
#undef PUTC1
#undef INC
#undef PUTC2
#undef NEW_STORE

	for (i = 0; i < store2_len; i++) {
	    if (store2[i].store)
		free(store2[i].store);
	}
	if (store2)
	    free(store2);

    }

    return store1;
}

int lib_error P_((const char * format, const char *msg, ...));

int lib_error (
#if ANSI_C
			      const char * format, 
			      const char *msg, ...
#else
			      format, msg, va_alist
#endif
			      )
#if !ANSI_C
     CONST char * format;
     CONST char *msg;
     va_dcl
#endif
{
    int ret = 0;
    va_list vl;
    char * store1;
    
    dprint(77,(debugfile,"lib_error: format=%s\n",format));
    dprint(77,(debugfile,"              msg=%s\n",msg));
    
    Va_start(vl, msg);           /* defined in defs.h */
    
    store1 = elm_vmessage(0,format,msg,vl);
    
    va_end(vl);
    
    ret = H(store1);
    free(store1);
    return ret;
}


int lib_transient P_((CONST char * format, CONST char *msg, ...));

int lib_transient (
#if ANSI_C
			      CONST char * format, 
			      CONST char *msg, ...
#else
			      format, msg, va_alist
#endif
			      )
#if !ANSI_C
     CONST char * format;
     CONST char *msg;
     va_dcl
#endif
{
  int ret = 0;
  va_list vl;
  char * store1;

  dprint(79,(debugfile,"lib_transient: format=%s\n",format));
  dprint(79,(debugfile,"              msg=%s\n",msg));
  	      
  Va_start(vl, msg);           /* defined in defs.h */
  
  store1 = elm_vmessage(0,format,msg,vl);

  va_end(vl);

  if (def_err_handler == T)
      ret = H(store1);
  else
      ret = T(store1);

  free(store1);
  return ret;
}
 
char * lib_prompt P_((int prompt, CONST char * format, CONST char *msg, ...));
char * lib_prompt (
#if ANSI_C
		int prompt, CONST char * format, CONST char *msg, ...
#else
		prompt,format,msg,va_alist
#endif
		)
#if !ANSI_C
     int prompt;
     CONST char * format;
     CONST char *msg;
     va_dcl
#endif
{
  va_list vl;
  char * store1, *ret;

  dprint(79,(debugfile,"lib_prompt: prompt=%d, format=%s\n",prompt,format));
  dprint(79,(debugfile,"               msg=%s\n",msg));
  	      
  if (getppid() == 1) 	{
      /* we've lost our shell! */
      dprint(4,(debugfile,"lib_prompt=NULL: Lost our shell (getppid=1)\n"));
      return NULL;
  }      
      
  Va_start(vl, msg);           /* defined in defs.h */

  store1 = elm_vmessage(0,format,msg,vl);

  va_end(vl);

  ret = P(store1,prompt);
  free(store1);
  return ret;
}

char *elm_message (
#if ANSI_C
		   CONST char * format, 
		   CONST char *msg, ...
#else
		   format, msg, va_alist
#endif
		)   
#if !ANSI_C
     CONST char * format; 
     CONST char *msg;
     va_dcl
#endif
{
  va_list vl;
  char * store1;

  dprint(79,(debugfile,"elm_message: format=%s\n",format));
  dprint(79,(debugfile,"             msg=%s\n",msg));
  	            
  Va_start(vl, msg);           /* defined in defs.h */
  
  store1 = elm_vmessage(0,format,msg,vl);

  va_end(vl);

  return store1;
}

struct string * format_string (
#if ANSI_C
			       CONST char * format, 
			       CONST char *msg, ...
#else
			       format, msg, va_alist
#endif
		)   
#if !ANSI_C
     CONST char * format; 
     CONST char *msg;
     va_dcl
#endif
{
  va_list vl;
  struct string * store1;

  dprint(79,(debugfile,"format_string: format=%s\n",format));
  dprint(79,(debugfile,"               msg=%s\n",msg));
  	            
  Va_start(vl, msg);           /* defined in defs.h */
  
  store1 = elm_smessage(0,format,msg,vl);

  va_end(vl);

  return store1;
}


int elm_sfprintf (
#if ANSI_C
		 char * buffer, int size,
		 CONST char * format, 
		 CONST char *msg, ...
#else
		 buffer, size, format, msg, va_alist
#endif
		)   
#if !ANSI_C
     char * buffer;
     int size;
     CONST char * format; 
     CONST char *msg;
     va_dcl
#endif
{
  va_list vl;
  char * store1;

  dprint(79,(debugfile,"elm_sfprintf: size=%d, format=%s\n",size, format));
  dprint(79,(debugfile,"             msg=%s\n",msg));
  	            
  Va_start(vl, msg);           /* defined in defs.h */
  
  store1 = elm_vmessage(size-1,format,msg,vl);

  va_end(vl);

  strfcpy(buffer,store1,size);
  free(store1);
  return strlen(buffer);
}

int elm_fprintf (
#if ANSI_C
		 FILE *f, CONST char * format, CONST char *msg, ...
#else
		 f, format,msg, va_alist
#endif
		 )
#if !ANSI_C
     FILE *f;
     CONST char * format;
     CONST char *msg;
     va_dcl     
#endif
{
    int ret;

    va_list vl;
    char * store1;
    
    dprint(79,(debugfile,"elm_fprintf: format=%s\n",format));
    dprint(79,(debugfile,"             msg=%s\n",msg));
    
    Va_start(vl, msg);           /* defined in defs.h */
    
    store1 = elm_vmessage(0,format,msg,vl);
    
    va_end(vl);
    
    ret = fputs(store1,f);    
    if (ret != EOF)
	ret = strlen(store1);
    free(store1);
    return ret;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
