static char rcsid[] = "@(#)$Id: remote_mbx.c,v 1.3 2000/07/30 15:12:11 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "mbx_imp.h"
#include "s_me.h"
#include "s_elm.h"

#ifdef REMOTE_MBX

/* Seems that h_errno is macro on AIX */
#ifndef h_errno
extern int h_errno;
#endif

#include <errno.h>
extern int errno;

void zero_remote_account(ra)
     struct remote_account *ra;
{
    /* bzero is defined hdrs/defs.h */
    bzero (ra, sizeof (struct remote_account));

    ra->hostaddr.sa.sa_family = AF_UNSPEC;
    ra->socket                = -1;
    ra->username              = NULL;
    ra->host                  = NULL;
}

extern void free_remote_account(ra)
     struct remote_account *ra;
{
    ra->hostaddr.sa.sa_family = AF_UNSPEC;

    if (-1 != ra->socket) {
	close(ra->socket);
	ra->socket = -1;
    }
    if (ra->username) {
	free(ra->username);
	ra->username = NULL;
    }
    if (ra->host) {
	free(ra->host);
	ra->host = NULL;
    }
}

int connect_remote_account P_((struct remote_account *ra,
			       PORTS portlist[],
			       PORTS *got,
			       struct hostent *he));

int connect_remote_account(ra,portlist,got,he)
     struct remote_account *ra;
     PORTS portlist[];
     PORTS *got;
     struct hostent *he;	    
{
    int ok = 0;
    int last_error = 0;
    
#ifdef I_NETINET_IN
    if (he->h_addrtype == AF_INET) {
	int idx;

	ra->hostaddr.sin.sin_family  = he->h_addrtype;

	if (sizeof(ra->hostaddr.sin.sin_addr)  != he->h_length) {
	    lib_error(FRM("%s: Bad IP-addr length %d (should be %d)"),
		      he->h_name,he->h_length,
		      sizeof(ra->hostaddr.sin.sin_addr));
	    ok = 0;
	    goto clean;
	}

	ra->socket = socket(he->h_addrtype,SOCK_STREAM,IPPROTO_TCP);
	if (-1 == ra->socket) {
	    int err =errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeFailedToCreate,
			      "Failed to create socket: %s"),
		      error_description(err));
	    ok = 0;
	    goto clean;
	}

	for (idx = 0; !ok  && he->h_addr_list[idx]; idx++) {
	    int idx2;
	    memcpy(&(ra->hostaddr.sin.sin_addr),
		   he->h_addr_list[idx],he->h_length);

	    lib_transient(CATGETS(elm_msg_cat, MeSet,MeConnecting,
				  "Connecting to %s [%s]... (%d)"),
			  he->h_name,
			  inet_ntoa(ra->hostaddr.sin.sin_addr),
			  idx);


	    for (idx2 = 0; portlist[idx2] != PORT_end && !ok; idx2++) {
		ra->hostaddr.sin.sin_port = htons(portlist[idx2]); 
		
#ifdef NEED_REOPEN_AFTER_FAILED_CONNECT
		if (idx || idx2) {
		    dprint(12, (debugfile,
				"connect_remote_account: NEED_REOPEN_AFTER_FAILED_CONNECT -- reopening...\n"));
		    close(ra->socket);
		    ra->socket = socket(he->h_addrtype,
					SOCK_STREAM,IPPROTO_TCP);
		    if (-1 == ra->socket) {
			dprint(12, (debugfile,
				    "connect_remote_account: socket failed: %s\n",
				    error_description(errno)));
			goto fail_connect;
		    }			
		}
#endif

		if (-1 == connect(ra->socket,& (ra->hostaddr.sa),
				  sizeof (ra->hostaddr.sin))) {
		    last_error= errno;
		    dprint(12, (debugfile,
				"connect_remote_account: connect failed: %s\n",
				error_description(last_error)));
		    
		    if (errno == EINVAL) {
			/* Seems that it is not not allowed several connect 
			 * attempts with same socket ...
			 */
			last_error= errno;
			dprint(12, (debugfile,
				    "connect_remote_account: -- reopening socket and retrying...\n"));
			
			close(ra->socket);
			ra->socket = socket(he->h_addrtype,SOCK_STREAM,
					    IPPROTO_TCP);
			if (-1 == ra->socket ||
			    -1 == connect(ra->socket,&(ra->hostaddr.sa),
					  sizeof (ra->hostaddr.sin))) {
			    last_error= errno;
			    dprint(12, (debugfile,
					"connect_remote_account: connect failed: %s\n",
					error_description(last_error)));
			} else
			    ok = 1;
		    }
		} else
		    ok = 1;
		
		/* If got connection to host and and it refused
		   connection, try another ports
		*/
		if (!ok && last_error != ECONNREFUSED) 
		    break;			   
		if (ok) {
		    *got = portlist[idx2];
		    
		    dprint(12, (debugfile,
				"connect_remote_account: Connection succeed %s [%s], port %d\n",
				he->h_name,
				inet_ntoa(ra->hostaddr.sin.sin_addr),
				portlist[idx2]));
		}
	    }	    
	}
    } else 
#endif /* I_NETINET_IN */	
	/* Should AF_INET6 also be supported on here? ... */
	{
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnsupportedAddrType,
			      "Name %s have odd type address"),
		      he->h_name);
	    ok = 0;
	    goto clean;
	}

    if (!ok) {
    fail_connect:
	lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailed,
			  "Failed to connect %s: %s"),
		  he->h_name,error_description(last_error));
	ok = 0;
	goto clean;
    }

    /* We set folder to non-blocking after connect, because
     * non-blocking connect is little complicate to use
     */
    
    if (-1 == fcntl(ra->socket,F_SETFL,O_NONBLOCK)) {
	int err = errno;
	dprint(12, (debugfile,
		    "connect_remote_account: fcntl [O_NONBLOCK] failed: %s\n",
		    error_description(err)));
    }


 clean:
    dprint(12, (debugfile,
		"connect_remote_account=%d\n",ok)); 
    return ok;
}
#endif /* REMOTE_MBX */

int remote_folder_type(fh)
     struct folder_info *fh;
{
    char * rest, *sep,*p;
    int ret = 0;

    /* Explicite disallow $MAIL to be interpreted as remote mailbox */

    if ('/' == fh->cur_folder[0] ||
	((p = getenv("MAIL")) && strcmp(fh->cur_folder, p) == 0) ||
	0 == access(fh->cur_folder,ACCESS_EXISTS)) {
	dprint(8, (debugfile,
		   "remote_folder_type=0 (file: %s)\n",
		   fh->cur_folder));
	return 0;
    }
    
    rest = strpbrk(fh->cur_folder,"/:");
    sep  = strchr(fh->cur_folder,'@');
    if (sep && (!rest || rest > sep)) {
#ifdef REMOTE_MBX
	struct hostent *he;	    
	int status = 1;
	struct remote_account X;
	PORTS ports[] = { PORT_imap4, PORT_pop3, PORT_end };
	PORTS ports_imaponly[] = { PORT_imap4, PORT_end };
	char * imap_only = NULL;  /* Flag and name of imap folder ... */
	PORTS got;

	zero_remote_account(&X);
	
	if (sep == fh->cur_folder || sep[1] == '\0' ||
	    rest == sep+1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeBadRemoteMailbox,
			      "Bad remote mailbox: %s"),
		      fh->cur_folder);
	    status = 0;
	    goto clean;
	}

	if (rest) {
	    CONST char * START = sep+1;
	    X.host = safe_malloc(rest - START+1);
	    memcpy(X.host,START,rest - START);
	    X.host[rest - START] = '\0';
	} else
	    X.host = safe_strdup(sep+1);	   

	X.username = safe_malloc(sep - fh->cur_folder+1);
	memcpy(X.username,fh->cur_folder,sep - fh->cur_folder);
	X.username[sep - fh->cur_folder] = '\0';


	he = gethostbyname(X.host);
	if (!he) {
	    dprint(12, (debugfile,
			"remote_folder_type: %s: h_errno=%d\n",
			X.host,h_errno));
	    switch(h_errno) {
	    case HOST_NOT_FOUND:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoRemoteHost,
				  "No remote mailbox server host: %s"),
			  X.host);
		status = 0;
		break;
	    case NO_ADDRESS:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNoAddress,
				  "Name %s have not IP-address"),
			  X.host);
		status = 0;
	    case TRY_AGAIN:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeTemporary,
				  "Address for %s is not yet found..."),
			  X.host);
		status = 0;
		break;			  
	    default:
		lib_error(CATGETS(elm_msg_cat, MeSet,MeNameError,
				  "Failed to get address for %s"),
			  X.host);
		status = 0;
		break;
	    }
	    goto clean;
	}
	
	/* canonify folder name .... */
	if (rest) {
	    /* Note that  'rest' is pointing to same are than fh->cur_folder
	     */
	    imap_only = safe_strdup(rest);
	    free(fh->cur_folder);
	    fh->cur_folder = elm_message(FRM("%s@%s%s"),
					 X.username,he->h_name,imap_only); 
	    rest = NULL; /* is no longer valid pointer */
	} else {
	    free(fh->cur_folder);
	    fh->cur_folder = elm_message(FRM("%s@%s"),
					 X.username,he->h_name);
	}

	/* Canonify host name ... */
	free(X.host);
	X.host = safe_strdup(he->h_name);

	if (!connect_remote_account(&X,
				    imap_only ? ports_imaponly : ports,
				    &got,he))
	    goto clean;

	switch (got) {
	case PORT_pop3:
	    fh -> folder_type = POP_MBX;
	    status = 1;
	    dprint(8, (debugfile,
		       "remote_folder_type: POP mailbox (succeed)\n"));

	    fh->folder_type->init_it(fh);
	    free_remote_account(&(fh->p->a.pop_mbx.C));
	    fh->p->a.pop_mbx.C = X;	    
	    ret = 1;
	    
	    break;
	case PORT_imap4:
	    if (!imap_only)
		imap_only = safe_strdup(":INBOX");
	    fh -> folder_type = IMAP_MBX;
	    status = 1;
	    dprint(8, (debugfile,
		       "remote_folder_type: IMAP mailbox (succeed)\n"));

	    fh->folder_type->init_it(fh);
	    free_remote_account(&(fh->p->a.imap_mbx.C));
	    fh->p->a.pop_mbx.C = X;	    
	    if (':' == imap_only[0]) 
		fh->p->a.imap_mbx.folder = safe_strdup(imap_only+1);
	    else
		fh->p->a.imap_mbx.folder = safe_strdup(imap_only);
	    ret = 1;
	    break;
	}
	
	if (status) {
	    /* In here we can use temp_dir instead of default_temp because
	     * that file does not act as session lock ...
	     */
	    if (imap_only) {
		char *Y = strrchr(imap_only,'/');
		if (!Y)
		    Y = imap_only;
		elm_sfprintf(fh-> cur_tempfolder,
			     sizeof fh -> cur_tempfolder,
			     FRM("%s%s%s@%s:%s"),
			     temp_dir,temp_mbox,X.username,X.host,Y+1);	    
	    } else
		elm_sfprintf(fh-> cur_tempfolder,
			     sizeof fh -> cur_tempfolder,
			     FRM("%s%s%s@%s"),
			     temp_dir,temp_mbox,X.username,X.host);
	} else {
	clean:
	    free_remote_account(&X);
	    
	    fh -> folder_type = NO_NAME;
	    fh->folder_type->init_it(fh);
	    ret = 1;      /* Is 'remote mailbox' ... */	    
	}
	if (imap_only)
	    free(imap_only);
#else
	dprint(8, (debugfile,
		   "remote_folder_type: Remote mailboxes are not supported.\n"));
#endif /* REMOTE_MBX */
    }
    dprint(8, (debugfile,
	       "remote_folder_type=%d\n",ret));
    return ret;
}

void zero_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    buffer->read_buffer = NULL;
    buffer->read_len    = 0;
}

void free_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    if (buffer->read_buffer) {
	free(buffer->read_buffer);
	buffer->read_buffer = NULL;
    }
    buffer->read_len = 0;
}

#define READ_BLOCK   256

int ReadFromSocket(fd,buffer,wanted)
     int fd; 
     struct Read_Buffer *buffer;
     int wanted;
{
    int n;

    if (wanted > 0) {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + wanted);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, wanted);
    } else {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + READ_BLOCK);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, READ_BLOCK);
    }

    if (n > 0) {
	buffer -> read_len += n;
	dprint(13, (debugfile,
                    "ReadFromSocket: Readed %d bytes (%d total)\n",
		    n,buffer->read_len));
    }

    return n;
}

int find_crlf(buffer, add_null)
     struct Read_Buffer *buffer;
     int add_null;
{
    char * p = buffer->read_buffer;
    int i;

    for (i = 0; i < buffer->read_len -1; i++)
	if ('\r' == p[i] && '\n' == p[i+1]) {
	    if (add_null) {
		p[i] = '\0';
		p[i+1] = '\0';
	    }
	    dprint(49, (debugfile,"find_crlf=%d (%d total)\n",
			i+2,buffer->read_len));
	    return i+2;
	}
    dprint(49, (debugfile,"find_crlf=0 (%d total)\n",
		buffer->read_len));
    return 0;
}

void cut_line(buffer, len)
     struct Read_Buffer *buffer;
     int len;
{
    char * p = buffer->read_buffer;
    
    buffer->read_len -= len;
    dprint(49, (debugfile,"cut_line: %d chars consumed, %d left\n",
		len,buffer->read_len));
    if (buffer->read_len)
	memmove(p, p+len, buffer->read_len);
}     


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


