static char rcsid[] = "@(#)$Id: schedule.c,v 1.6 2000/07/30 15:12:11 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.6 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"

#include <errno.h>
extern int errno;

int no_action_routine(fd,data)
     int fd; 
     void * data;
{
    return 0;
}

static struct actions {
    int    fd; 
    int    timeout_sec;
    action_routine * read_act; 
    action_routine * write_act; 		       
    action_routine * timeout_act; 		       
    time_t next_timeout;
    void  * data;
    int    busy;                     /* Prevent nested calls --
				      * when read_act(), write_act() or
				      * timeout_act() calls wait_for_timeout()
				      * or wait_for_action() directly or
				      * indirectly (for example with lib_error)
				     */
}         * actions       = NULL;
static int  actions_count = 0; 

static void remove_action P_((int ptr));
static void remove_action(ptr)
     int ptr;
{
    int i;

    dprint(4,(debugfile,"** Removing action %d  (fd=%d)\n", ptr,
	      actions[ptr].fd)); 
    for (i = ptr+1; i < actions_count; i++) {
	actions[i-1] = actions[i];
    }
    actions_count--;
}

void change_action (fd,timeout_sec,read_act,write_act,timeout_act,data)
     int fd; 
     int timeout_sec;
     action_routine * read_act; 
     action_routine * write_act; 	
     action_routine * timeout_act;
     void *data;
{
    int ptr;

    for (ptr = 0; ptr < actions_count; ptr++)
	if (actions[ptr].fd == fd)
	    break;

    if (ptr == actions_count) {	
	actions = safe_realloc(actions, (++actions_count) * 
			       sizeof (struct actions));	
	actions[ptr].timeout_sec  = 0;
	actions[ptr].next_timeout = 0;
	dprint(4,(debugfile,"** Adding action %d  (fd=%d)\n", ptr,fd)); 
    }

    if (actions[ptr].timeout_sec != timeout_sec) {
	actions[ptr].next_timeout = time(NULL) + timeout_sec;
    }

    actions[ptr].fd          = fd;
    actions[ptr].timeout_sec = timeout_sec;
    actions[ptr].read_act    = read_act;
    actions[ptr].write_act   = write_act;
    actions[ptr].timeout_act = timeout_act;
    actions[ptr].data        = data;
    actions[ptr].busy        = 0;
    
    if (no_action_routine == read_act &&
	no_action_routine == write_act &&
	no_action_routine == timeout_act) {
	remove_action(ptr);
    }
}

void clear_action(fd)
     int fd;
{
    change_action(fd,0,
		  no_action_routine,no_action_routine,no_action_routine,
		  NULL);
}

int have_actions()
{
    return actions_count;
}

#if POLL_METHOD == 1
/* --------------------- select() ------------------------------------------ */

static int real_wait P_((action_routine * action,
			 time_t wait_timeout,time_t next_timeout));

static int real_wait(action,wait_timeout,next_timeout)
     action_routine * action;
     time_t wait_timeout, next_timeout;
{
    int result = 0;
    int done = 0;
    int err = 0;

    dprint(25,(debugfile,
	       "real_wait: [select] %s %p, wait_timeout=%ld, next_timeout=%ld (actions_count=%d)\n", 	       
	       action == no_action_routine ? "no action" : "action",
	       action,
	       (long) wait_timeout, (long) next_timeout, actions_count)); 

    do {
	int i, n = 0;
	fd_set read_set;
	fd_set write_set;
	struct timeval timeout, *timeout_val = NULL;
	int ret;
	time_t current;

	FD_ZERO(&read_set);
	FD_ZERO(&write_set);

	dprint(25,(debugfile,
		   "real_wait: actions:"));
	for (i = 0; i < actions_count; i++) {
	    if (n <= actions[i].fd)
		n = actions[i].fd+1;

	    dprint(25,(debugfile, " (fd=%d",actions[i].fd));
	    if (actions[i].busy) {
		dprint(25,(debugfile, " busy!"));
	    } else {
		if (actions[i].read_act != no_action_routine) {
		    FD_SET(actions[i].fd,&read_set);
		    dprint(25,(debugfile, " READ"));
		}
	    
		if (actions[i].write_act != no_action_routine) {
		    FD_SET(actions[i].fd,&write_set);
		    dprint(25,(debugfile, " WRITE"));
		}	
		if (actions[i].timeout_act != no_action_routine) {
		    dprint(25,(debugfile, " TIMEOUT"));
		}
	    }
	    dprint(25,(debugfile, ")"));
	}
	dprint(25,(debugfile, "\n"));	

	if (next_timeout > 0) {
	    int diff = next_timeout - time(NULL);

	    dprint(25,(debugfile, "real_wait: diff=%d\n", diff));
	    
	    if (diff <= 0) 
		timeout.tv_sec = 0;
	    else
		timeout.tv_sec = diff;
	    timeout.tv_usec    = 0;
	    timeout_val        = &timeout;
	}

	dprint(25,(debugfile, "real_wait: n = %d, read set=", n));
	for (i = 0; i < n; i++)
	    if (FD_ISSET(i,&read_set)) {
		dprint(25,(debugfile, " %d", i));
	    }
	dprint(25,(debugfile, ", write set="));
	for (i = 0; i < n; i++)
	    if (FD_ISSET(i,&write_set)) {
		dprint(25,(debugfile, " %d", i));
	    }
	dprint(25,(debugfile, ", timeout="));
	if (timeout_val) {
	    dprint(25,(debugfile, " sec=%ld usec=%ld\n", 
		       (long) timeout_val->tv_sec,
		       (long) timeout_val->tv_usec));
	} else {
	    dprint(25,(debugfile, "NULL\n")); 
	}
#ifdef _HPUX_SOURCE
	ret = select(n,(int *)&read_set,(int *)&write_set,NULL,timeout_val);
#else
	ret = select(n,&read_set,&write_set,NULL,timeout_val);
#endif
	if (ret < 0) {
	    err = errno;
	    dprint(25, (debugfile,
			"select: %s (errno %d)\n",
			error_description(err),err));
	    result = 0;
	    break;
	}

	current = time(NULL);
	dprint(25,(debugfile, "real_wait: select=%d, current=%ld\n",
		   ret,(long)current));	

	if (ret > 0) {
	    for (i = 0; i < actions_count; i++) {
		if (!actions[i].busy) {
		    actions[i].busy = 1;
		    if (FD_ISSET(actions[i].fd,&read_set)) {
			int val; 
			dprint(25,(debugfile, 
				   "real_wait: running read_act for %d (action %d)\n",
				   actions[i].fd,i));
			val  = actions[i].read_act(actions[i].fd,
						   actions[i].data);
			dprint(25,(debugfile, 
				   "real_wait: [%d].read_act=%d\n",i,val));
			if (action == actions[i].read_act) {
			    done = 1;
			    result = 1;
			}
			if (!val)
			    actions[i].read_act = no_action_routine;
		    }
		    
		    if (FD_ISSET(actions[i].fd,&write_set)) {
			int val;
			dprint(25,(debugfile, 
				   "real_wait: running write_act for %d (action %d)\n",
				   actions[i].fd,i));
			val = actions[i].write_act(actions[i].fd,
						   actions[i].data);
			dprint(25,(debugfile, 
				   "real_wait: [%d].write_act=%d\n",i,val));
			if (action == actions[i].write_act) {
			    done = 1;
			    result = 1;
			}
			if (!val)
			    actions[i].write_act = no_action_routine;
		    }
		    actions[i].busy = 0;
		} 
	    }
	}

	for (i = 0; i < actions_count; i++) {
	    if (actions[i].next_timeout &&
		actions[i].next_timeout <= current) {

		if (actions[i].busy) {
		    dprint(25,(debugfile, 
			       "real_wait: %d busy (action %d) -- timout skipped\n",
			       actions[i].fd,i));
		} else {		
		    int val;
		    actions[i].busy = 1;

		    dprint(25,(debugfile, 
			       "real_wait: running timeout_act for %d (action %d)\n",
			       actions[i].fd,i));
		    val = actions[i].timeout_act(actions[i].fd,
						 actions[i].data);
		    dprint(25,(debugfile, 
			       "real_wait: [%d].timeout_act=%d\n",i,val));
		    if (action == actions[i].timeout_act) {
			done = 1;
			result = 1;
		    }
		    if (!val) {
			actions[i].timeout_act  = no_action_routine;
			actions[i].next_timeout = 0;
		    } else {
			actions[i].next_timeout = 
			    current + actions[i].timeout_sec;
		    }		
		    actions[i].busy = 0;
		}
	    }
	}

	if (next_timeout > 0 && next_timeout < current) {
	    done = 1;
	    if (result == 0)
		result = -1;
	}

	if (wait_timeout > 0 && wait_timeout <= current) {
	    done = 1;
	    result = 1;
	}
	dprint(25,(debugfile, 
		   "real_wait: done=%d, result=%d\n",done,result));

    } while(!done);

    
    dprint(25,(debugfile, 
	       "real_wait=%d (errno=%d)\n",result,err));
    errno = err;
    return result;
}

#endif

#if POLL_METHOD == 2
/* --------------------- poll() -------------------------------------------- */

/* For MAXINT */
#include <values.h>

static int real_wait P_((action_routine * action,
			 time_t wait_timeout,time_t next_timeout));

static int real_wait(action,wait_timeout,next_timeout)
     action_routine * action;
     time_t wait_timeout, next_timeout;
{
    int result = 0;
    int done = 0;
    int err = 0;
    struct pollfd  * polls = NULL;
    int polls_count = 0;

    dprint(25,(debugfile,
	       "real_wait: [poll] %s %p, wait_timeout=%ld, next_timeout=%ld (actions_count=%d)\n", 	       
	       action == no_action_routine ? "no action" : "action",
	       action,
	       (long) wait_timeout, (long) next_timeout, actions_count)); 

    do {
	int i;
	int ret;
	time_t current;
	int timeout = -1;

	if (polls_count != actions_count) {
	    polls = safe_realloc(polls,
				 actions_count * sizeof(struct pollfd));
	    polls_count = actions_count;
	    bzero(polls,actions_count * sizeof(struct pollfd));
	    dprint(25,(debugfile,
		       "real_wait: polls=%p, polls_count=%d\n",
		       polls,polls_count));	    
	}


	dprint(25,(debugfile,
		   "real_wait: actions:"));
	for (i = 0; i < actions_count; i++) {

	    dprint(25,(debugfile, " (fd=%d",actions[i].fd));
	    polls[i].fd     = actions[i].fd;
	    polls[i].events = 0;
	    polls[i].revents = 0;

	    if (actions[i].busy) {
		dprint(25,(debugfile, " busy!"));
	    } else {
		if (actions[i].read_act != no_action_routine) {
		    polls[i].events |= POLLIN | POLLHUP;
		    dprint(25,(debugfile, " READ"));
		}
	    
		if (actions[i].write_act != no_action_routine) {
		    polls[i].events |= POLLOUT;
		    dprint(25,(debugfile, " WRITE"));
		}	
		if (actions[i].timeout_act != no_action_routine) {
		    dprint(25,(debugfile, " TIMEOUT"));
		}
	    }
	    dprint(25,(debugfile, ")"));
	}
	dprint(25,(debugfile, "\n"));	

	if (next_timeout > 0) {
	    int diff = next_timeout - time(NULL);

	    dprint(25,(debugfile, "real_wait: diff=%d\n", diff));
	    
	    if (diff <= 0) 
		timeout = -1;
	    else if (diff < MAXINT / 1000)
		timeout = diff * 1000;
	    else
		timeout = MAXINT;
	}

	dprint(25,(debugfile, "real_wait: polls_count = %d, polls=", 
		   polls_count));
	for (i = 0; i < polls_count; i++)
	    dprint(25,(debugfile, " {fd=%d events=%d}", 
		       polls[i].fd, polls[i].events));


	dprint(25,(debugfile, ", timeout=%d\n", 
		   timeout));

	ret = poll(polls,polls_count,timeout);

	if (ret < 0) {
	    err = errno;
	    dprint(25, (debugfile,
			"select: %s (errno %d)\n",
			error_description(err),err));
	    result = 0;
	    break;
	}

	current = time(NULL);
	dprint(25,(debugfile, "real_wait: poll=%d, current=%ld\n",
		   ret,(long)current));	

	if (ret > 0) {
	    for (i = 0; i < polls_count && i < actions_count; i++) {
		if (polls[i].revents) {
		    dprint(25,(debugfile, 
			       "real_wait: polls[%d]: fd=%d revents=%d\n",
			       i,polls[i].fd,polls[i].revents));
		}
		if (!actions[i].busy) {
		    actions[i].busy = 1;
		    if (polls[i].revents && (POLLIN|POLLHUP|POLLERR)) {
			int val; 
			dprint(25,(debugfile, 
				   "real_wait: running read_act for %d (action %d)\n",
				   actions[i].fd,i));
			val  = actions[i].read_act(actions[i].fd,
						   actions[i].data);
			dprint(25,(debugfile, 
				   "real_wait: [%d].read_act=%d\n",i,val));
			if (action == actions[i].read_act) {
			    done = 1;
			    result = 1;
			}
			if (!val)
			    actions[i].read_act = no_action_routine;
		    }
		    
		    if (polls[i].revents && (POLLOUT|POLLERR)) {
			int val;
			dprint(25,(debugfile, 
				   "real_wait: running write_act for %d (action %d)\n",
				   actions[i].fd,i));
			val = actions[i].write_act(actions[i].fd,
						   actions[i].data);
			dprint(25,(debugfile, 
				   "real_wait: [%d].write_act=%d\n",i,val));
			if (action == actions[i].write_act) {
			    done = 1;
			    result = 1;
			}
			if (!val)
			    actions[i].write_act = no_action_routine;
		    }
		    actions[i].busy = 0;
		}
		if (polls[i].revents & POLLERR) {
		    dprint(25,(debugfile, 
			       "real_wait: [%d] POLLERR\n",i));
		    actions[i].write_act = no_action_routine;
		    actions[i].read_act = no_action_routine;
		    done = 1;
		    result = 1;
		}
	    }
	}

	for (i = 0; i < actions_count; i++) {
	    if (actions[i].next_timeout &&
		actions[i].next_timeout <= current) {

		if (actions[i].busy) {
		    dprint(25,(debugfile, 
			       "real_wait: %d busy (action %d) -- timout skipped\n",
			       actions[i].fd,i));
		} else {		
		    int val;
		    actions[i].busy = 1;

		    dprint(25,(debugfile, 
			       "real_wait: running timeout_act for %d (action %d)\n",
			       actions[i].fd,i));
		    val = actions[i].timeout_act(actions[i].fd,
						 actions[i].data);
		    dprint(25,(debugfile, 
			       "real_wait: [%d].timeout_act=%d\n",i,val));
		    if (action == actions[i].timeout_act) {
			done = 1;
			result = 1;
		    }
		    if (!val) {
			actions[i].timeout_act  = no_action_routine;
			actions[i].next_timeout = 0;
		    } else {
			actions[i].next_timeout = 
			    current + actions[i].timeout_sec;
		    }		
		    actions[i].busy = 0;
		}
	    }
	}

	if (next_timeout > 0 && next_timeout < current) {
	    done = 1;
	    if (result == 0)
		result = -1;
	}

	if (wait_timeout > 0 && wait_timeout <= current) {
	    done = 1;
	    result = 1;
	}
	dprint(25,(debugfile, 
		   "real_wait: done=%d, result=%d\n",done,result));

    } while(!done);

    
    if (polls)
	free(polls);

    dprint(25,(debugfile, 
	       "real_wait=%d (errno=%d)\n",result,err));
    errno = err;
    return result;
}

#endif


#if POLL_METHOD
/* ----------- GENERIC ---------------------------------------------------- */

static int wait_for_something P_((action_routine * action,int seconds));
static int wait_for_something(action,seconds)
     action_routine * action;
     int seconds; 
{
    time_t next_timeout = 0;
    time_t wait_timeout = 0;
    int result,err = 0;
    int i;

    if (seconds >= 0) 
	wait_timeout = time(NULL) + seconds;

    do {
	next_timeout = wait_timeout;
	
	for (i = 0; i < actions_count; i++) {
	    if (!actions[i].busy &&
		actions[i].timeout_act != no_action_routine)
		if (!next_timeout || next_timeout > actions[i].next_timeout)
		    next_timeout =  actions[i].next_timeout;
	}
    
	result = real_wait(action,wait_timeout,next_timeout);
	err = errno;

	for (i = 0; i < actions_count; i++) {
	    if (no_action_routine == actions[i].read_act &&
		no_action_routine == actions[i].write_act &&
		no_action_routine == actions[i].timeout_act) {
		remove_action(i);
		break;
	    }
	}
	
    } while (-1 == result);

    errno = err;
    return result;
}

int wait_for_action(action)
     action_routine * action;
{
    int ret;
    dprint(4,(debugfile,"wait_for_action:    START | action=%p\n",action)); 
    ret=wait_for_something(action,-1);
    dprint(4,(debugfile,"wait_for_action=%d  END (errno=%d)\n",
	      ret,errno)); 
    return ret;
}

int wait_for_timeout(seconds)
     int seconds;
{
    int ret;
    dprint(4,(debugfile,"wait_for_timeout:    START | seconds=%d\n",seconds)); 
    ret = wait_for_something(no_action_routine,seconds);
    dprint(4,(debugfile,"wait_for_timeout=%d  END (errno=%d) \n",
	      ret,errno)); 
    return ret;
}

int wait_for_action_or_timeout(action,seconds)
     action_routine * action;
     int seconds;
{
    int ret;
    dprint(4,(debugfile,"wait_for_action_or_timeout:    START | action=%p seconds=%d\n",
	      action,seconds)); 
    ret = wait_for_something(action,seconds);
    dprint(4,(debugfile,"wait_for_action_or_timeout=%d  END (errno=%d)\n",
	      ret,errno)); 
    return ret;
}
#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
