static char rcsid[] = "@(#)$Id: file.c,v 1.10 2000/07/30 15:12:19 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.10 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** File I/O routines, mostly the save to file command...

**/

#include "headers.h"
#ifdef PWDINSYS
#  include <sys/pwd.h>
#else
#  include <pwd.h>
#endif

#include "s_elm.h"
#include <errno.h>
#include "me.h"

extern int errno;

char *nameof();

static void save_message P_((int number, char *filename, 
			     FILE *fd, int pause, int appending, 
			     int silently, int delete, int text_only,
			     int *redraw));

static char copyhelp[VERY_LONG_STRING];
static char savehelp[VERY_LONG_STRING];
static char prev_fold[SLEN];           /* name of previous folder */

void init_helpmsg( msg, action_word, allow_dot, size_msg )
     char	*msg , *action_word ;
     int      allow_dot ;
     int size_msg;
{
	strfcpy( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg1,
			      "\n\r\n\rYou must specify a file or folder to "),
		 size_msg);

	strfcat( msg, action_word, size_msg );

	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg2,
		" to.  Several options\n\rare available:\n\r\
\n\r  '!'  will use your incoming mailbox ("),
		 size_msg);

	strfcat( msg, defaultfile, size_msg ) ;

	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg3,
		")\n\r  '>'  will use your \"received\" folder ("),
		 size_msg);

	strfcat( msg, nameof( recvd_mail ), size_msg ) ;

	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg4,
		")\n\r  '<'  will use your \"sent\" folder ("),
		 size_msg);

	strfcat( msg, nameof( sent_mail ), size_msg ) ;

	if( allow_dot ) {
	    strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg5,
				  ")\n\r  '.' will use the previous folder ("),
		     size_msg);

	    strfcat( msg, nameof(prev_fold), size_msg);
	}


	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg6,
			      ")\n\r  '@alias' will use the default folder for \"alias\"\n\r\n\r\
If you enter a filename elm will use that file.  If the file name begins\n\r\
with a '=', elm will look for the file in your folder directory\n\r\
(your folder directory is "),
		 size_msg);

	strfcat( msg, raw_folders, size_msg ) ;

	strfcat( msg, ").\n\r\n\r", size_msg ) ;

	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg7,
			      "You may use wildcards in the name, exactly as you do in the shell.  If you\n\r\
do this, a list of all matching files or folders will be printed and you\n\r\
will be prompted for another name.  Thus typing '=a*' will print the names\n\r\
of all folders in your folder directory starting with 'a'.  Typing 'b*.c'\n\r\
will print all files in your local directory starting with 'b' and ending\n\r"),
		 size_msg);
	strfcat( msg, catgets(elm_msg_cat, ElmSet, ElmFileHelpmsg8,
"in '.c'.  A star (*) will usually list everything.  See your local shell\n\r\
manual for more information on wildcards.\n\r\
\n\r\
Sometimes elm will help you out by suggesting a name.  If you want to use\n\r\
the suggested name, simply hit return.  If you don't like it, hit your\n\r\
erase or line erase keys.\n\r\
\n\r\
\n\r\
\n\r"),
		 size_msg);
}

int save(redraw, silently, delete, text_only)
     int *redraw, silently, delete, text_only;
{
	/** Save all tagged messages + current in a folder.  If no messages
	    are tagged, save the current message instead!  This routine
	    will return ZERO if the operation failed.
	    'redraw' is set to TRUE iff we use the '?' and mess up
	    the screen.  Pretty reasonable, eh?  If "silently" is set,
	    then don't output the "D" character upon marking for
	    deletion...
	    If delete is set, then delete the saved messages, else
	    we are just copying the messages without deletion.
	    If text_only is text, message is decoded.
	**/

    int	tagged = 0, i, oldstat, appending = 0,
	is_ordinary_file, scount = 0;
    int mesgnum;	/* message whose address is used for save-by-name fn */
    char	filename[SLEN], buffer[LONG_STRING],
	msg_buffer[SLEN], answer;
    FILE *save_file;
    int save_fd;

	oldstat = headers[current-1]->status;	/* remember */
	*redraw = FALSE;

	for (i=0; i < message_count; i++) {
	  if (ison(headers[i]->status, TAGGED)) {
	    if(!tagged)
	      mesgnum = i;	/* first tagged msg -  use this one for
				 * save-by-name folder name */
	    tagged++;
	  }
	}

	if (tagged == 0) {
	    mesgnum = current-1; /* use this one for save-by-name 
				    folder name */
	    tagged = 1;
	    setit(headers[current-1]->status, TAGGED);
	    /* 2 == text_only -- Not envelope information */
	    if (text_only)
		text_only = 2;
	}

	dprint(4, (debugfile, "%d message%s tagged for saving (save)\n", tagged,
		plural(tagged)));

Retry_Save_If_Not_Confirmed:
	while (1) {
	  int code;
	  int X,Y;

	  /* 2 == text_only -- Not envelope information */
	  if (text_only > 0) {
	      if (tagged == 1)
		  PutLineX(elm_LINES-2, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmSaveTextMesTo,
				   "%s text of message to: "), 
			   (delete ? cap_save_word : cap_copy_word));
	      else
		  PutLineX(elm_LINES-2, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmSaveTextMessagesTo,
				   "%s text messages to: "), 
			   (delete ? cap_save_word : cap_copy_word));
	  } else {
	      if (tagged == 1)
		  PutLineX(elm_LINES-2, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmSaveMessageTo,
				   "%s message to: "), 
			   (delete ? cap_save_word : cap_copy_word));
	      else
		  PutLineX(elm_LINES-2, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmSaveMessagesTo,
				   "%s messages to: "), 
			   (delete ? cap_save_word : cap_copy_word));
	  }
	  GetXYLocation(&X,&Y);

	  if (save_by_name &&
	      headers[current-1]->from &&
	      headers[current-1]->from[0].addr) {
	    /** build default filename to save to **/
	    get_return_name(headers[current-1]->from[0].addr, 
			    buffer, TRUE, sizeof buffer);
	    elm_sfprintf(filename, sizeof filename,
			 FRM("=%s"), buffer);
	  }
	  else
	    filename[0] = '\0';

	  code = optionally_enter(filename, X, Y, OE_REDRAW_MARK,
				  sizeof filename);

	  if (REDRAW_MARK == code) {
	    *redraw = TRUE;
	    continue;
	  }
  

	  if (0 != code || 
	      strlen(filename) == 0) {  /** <return> means 'cancel', right? **/
	    headers[current-1]->status = oldstat;	/* BACK! */
	    return(0);
	  }
	 
	  if (strcmp(filename,"?") == 0) {	/* user asked for help */
	    *redraw = TRUE;	/* set the flag so we know what to do later */
	    if ( delete )
	    {
	      if ( !*savehelp )	/* format helpmsg if not yet done */
	        init_helpmsg( savehelp, save_word, TRUE, sizeof savehelp ) ;
	      ClearScreen();
	      (void) printf( savehelp ) ;
	    }
	    else
	    {
	      if ( !*copyhelp )	/* format helpmsg if not yet done */
		init_helpmsg( copyhelp, copy_word, TRUE, sizeof copyhelp ) ;
	      ClearScreen();
	      (void) printf( copyhelp ) ;
	    }
	    continue;
	  }

	  /* was this a wildcard request? */
	  if ( has_wildcards( filename ) )
	  {
	    *redraw = TRUE ;
	    list_folders( 4, NULL, filename ) ;
	    continue ;
	  }
	  /* else - got a folder name - check it out */
	  if (! expand_filename(filename, TRUE, sizeof filename)) {
	    dprint(2, (debugfile,
		  "Error: Failed on expansion of filename %s (save)\n", 
		  filename));
	    continue;
	  }

	  is_ordinary_file = 
	      0 != strncmp (filename, folders, strlen(folders)) &&
	      !is_in_spool(filename);

	  /* 2 == text_only -- Not envelope information */
	  if (2 == text_only && !is_ordinary_file)
	      text_only = 1;

	  
	  if ((errno = can_open(filename, "a"))) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCannotSaveMessage,
			      "Cannot %s message to folder %s!"),
		      delete ? save_word:copy_word, filename);
	    continue;
	  }
	  break;	/* got a valid filename */
	}

	save_file_stats(filename);

	if (access(filename,ACCESS_EXISTS)== 0) {	/* already there!! */
	    appending = 1;
	    if (confirm_append || (confirm_files && is_ordinary_file)) {

	        if (is_ordinary_file)
	          elm_sfprintf(msg_buffer, sizeof msg_buffer,
			       CATGETS(elm_msg_cat, ElmSet, 
				       ElmConfirmFilesAppend,
				       "Append to an existing file `%s'? (%c/%c) "),
			       filename, *def_ans_yes, *def_ans_no);
	        else
	          elm_sfprintf(msg_buffer, sizeof msg_buffer,
			       CATGETS(elm_msg_cat, ElmSet, 
				       ElmConfirmFolderAppend,
				       "Append to mail folder `%s'? (%c/%c) "),
			       filename, *def_ans_yes, *def_ans_no);

	        answer = want_to(msg_buffer, *def_ans_no, elm_LINES-1, 1);

		if (answer != *def_ans_yes) {
		    goto Retry_Save_If_Not_Confirmed;
		}
	    }
	}
	else {
            if (confirm_create || (confirm_folders && !is_ordinary_file)) {

	        if (is_ordinary_file)
	          elm_sfprintf(msg_buffer, sizeof msg_buffer,
			       CATGETS(elm_msg_cat, ElmSet, 
				       ElmConfirmFilesCreate,
				       "Create a new file `%s'? (%c/%c) "),
			       filename, *def_ans_yes, *def_ans_no);
	        else
	          elm_sfprintf(msg_buffer, sizeof msg_buffer,
			       CATGETS(elm_msg_cat, ElmSet, 
				       ElmConfirmFolderCreate,
				       "Create a new mail folder `%s'? (%c/%c) "),
			       filename, *def_ans_yes, *def_ans_no);

	        answer = want_to(msg_buffer, *def_ans_no, elm_LINES-1, 1);

		if (answer != *def_ans_yes) {
		    goto Retry_Save_If_Not_Confirmed;
		}
	    }
	}
	  
	/* 2 == text_only -- Not envelope information */
	if (2 == text_only && appending)
	    text_only = 1;

	dprint(4,(debugfile, "Saving mail to folder '%s'...\n", filename));

	/* 2 == text_only -- Not envelope information */

	/* save this filename for the next time */
	if (2 != text_only)
	    strfcpy(prev_fold, filename, SLEN);
      
	if ((save_fd = open(filename,O_RDWR|O_CREAT,0600)) == -1 ||
	    (save_file = fdopen(save_fd,"r+")) == NULL) {
	  close(save_fd);
	  dprint(2, (debugfile,
		"Error: couldn't append to specified folder %s (save)\n", 
		filename));
	  lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntAppendFolder,
			    "Couldn't append to folder %s!"), 
		    filename);
	  headers[current-1]->status = oldstat;	/* BACK! */
	  return(0); 
	}

	if (lock_in_copy) {
	  /*  lock the file to be modified to serialize access with other */
	  /* programs like procmail. */

	  if (Grab_the_file(save_fd) != FLOCKING_OK){
	    fclose(save_file);
	    dprint(2, (debugfile,
		       "Error: couldn't lock specified folder %s (save)\n", 
		       filename));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntAppendFolder,
			      "Couldn't lock folder %s!"), 
		      filename);
	    headers[current-1]->status = oldstat; /* BACK! */
	    return(0); 
	  }
	}

	/* copy_message want now seek backwards ... */
	if (0 != fseek(save_file,0,SEEK_END)) {
	  if (lock_in_copy)
	    Release_the_file(save_fd); 
	  dprint(2, (debugfile,
		"Error: couldn't seek to end of specified folder %s (save)\n", 
		filename));
	  lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldntAppendFolder,
			    "Couldn't append to folder %s!"), 
		    filename);
	  headers[current-1]->status = oldstat;	/* BACK! */
	  fclose(save_file);
	  return (0);
	}

	/* if we need a redraw that means index screen no longer present
	 * so whatever silently was, now it's true - we can't show those
	 * delete markings.
	 */
	if(*redraw) silently = TRUE;

	for (i=0; i < message_count; i++) 	/* save each tagged msg */
	  if (headers[i]->status & TAGGED) {	      
	      save_message(i, filename, save_file, (tagged > 1), appending++, 
			   silently, delete, text_only, redraw);      
	      if(*redraw) silently = TRUE;
	      scount++;
	  }

	if (lock_in_copy)
	  Release_the_file(save_fd); /* XXX LOCKING */
	fclose(save_file);

	restore_file_stats(filename);

	if (text_only) {
	    if (tagged == 1 && !appending)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTextMessageSaved,
				  "Text of message %s to %s."), 
			  delete ? saved_word: copied_word,
			  filename);
	    else if (tagged > 1)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTextMessagesSaved,
				  "Text of %d messages %s to %s."), scount, 
			  delete ? saved_word: copied_word,
			  filename);
	} else {
	    if (tagged == 1 && !appending)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMessageSaved,
				  "Message %s to %s."), 
			  delete ? saved_word: copied_word,
			  filename);
	    else if (tagged > 1)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMessagesSaved,
				  "%d messages %s to %s."), scount, 
			  delete ? saved_word: copied_word,
			  filename);
	}
	return(1);
}

static void save_message(number, filename, fd, pause, appending, silently, 
		  delete, text_only, redraw)
    int number, pause, appending, silently, delete;
    char *filename;
    FILE *fd;
    int text_only;
    int *redraw;
{
	/** Save an actual message to a folder.  This is called by 
	    "save()" only!  The parameters are the message number,
	    and the name and file descriptor of the folder to save to.
	    If 'pause' is true, a sleep(sleepmsg) will be done after the
	    saved message appears on the screen...
	    'appending' is only true if the folder already exists 
	    If 'delete' is true, mark the message for deletion.
	    If text_only is set then decode_message, if
	       text_only == 2 no folder envelope will be saved
	**/

	int save_current, is_new;
	
	dprint(4, (debugfile, "\tSaving message %d to folder...\n", number));

	if (!current_folder)
	  return;

	save_current = current;
	current = number+1;

	/* change status from NEW before copy and reset to what it was
	 * so that copy doesn't look new, but we can preserve new status
	 * of message in this mailfile. This is important because if
	 * user does a resync, we don't want NEW status to be lost.
	 * I.e. NEW becomes UNREAD when we "really" leave a mailfile.
	 */
	if(0 != (is_new = 
		 ison(headers[number]->status, NEW)))
	  clearit(headers[number]->status, NEW);
	switch(text_only) {
	case 2:
	    copy_message(current_folder,headers[number],
			 "", fd, CM_DECODE | CM_FILT_HDR | CM_REMOVE_ENVELOPE);
	    break;
	case 1:
	    copy_message(current_folder,headers[number],
			 "", fd, CM_DECODE | CM_FILT_HDR);
	    break;
	case 0: default:
	    copy_message(current_folder,headers[number],
			 "", fd, CM_UPDATE_STATUS);
	    break;
	}
	/* Need redraw -- PGP output ?*/
	if (raw_off_called()) {
	    *redraw = TRUE;
	    silently = TRUE;
	}

	if(is_new)
	  setit(headers[number]->status, NEW);
	current = save_current;

	if (delete)
	  setit(headers[number]->status, DELETED); /* deleted, but ...   */
	clearit(headers[number]->status, TAGGED);  /* not tagged anymore */

	if (text_only) {
	    if (appending)
		lib_transient(CATGETS(elm_msg_cat, ElmSet, 
				      ElmTextMessageAppendedFolder,
				      "Text of message %d appended to folder %s."), 
			      number+1, filename);
	    else if (2 == text_only)
		lib_transient(CATGETS(elm_msg_cat, ElmSet, 
				      ElmTextMessageSavedFile,
				      "Text of message %d %s to file %s."),
			      number+1, delete ? saved_word : copied_word, 
			      filename);
	    else
		lib_transient(CATGETS(elm_msg_cat, ElmSet, 
				      ElmTextMessageSavedFolder,
				      "Text of message %d %s to folder %s."),
			      number+1, delete ? saved_word : copied_word, 
			      filename);
	} else {
	    if (appending)
		lib_transient(CATGETS(elm_msg_cat, ElmSet, 
				      ElmMessageAppendedFolder,
				      "Message %d appended to folder %s."), 
			      number+1, filename);
	    else
		lib_transient(CATGETS(elm_msg_cat, ElmSet, 
				      ElmMessageSavedFolder,
				      "Message %d %s to folder %s."),
			      number+1, delete ? saved_word : copied_word, 
			      filename);
	}

	if (! silently)
	  show_new_status(number);	/* update screen, if needed */

	if (pause && (!silently) && (!appending)) 
	    sleep_message();
}

int expand_filename(filename, use_cursor_control, size_filename)
     char *filename;
     int use_cursor_control;
     int size_filename;
{
	/** Expands	~/	to the current user's home directory
			~user/	to the home directory of "user"
			=,+,%	to the user's folder's directory
			!	to the user's incoming mailbox
			>	to the user's received folder
			<	to the user's sent folder
			.	to the last folder used 
			@alias	to the default folder directory for "alias"
			shell variables (begun with $)

	    Side effect: strips off trailing blanks

	    Returns 	1	upon proper expansions
			0	upon failed expansions
	 **/

	char temp_filename[SLEN], varname[SLEN], env_value[SLEN],
	     logname[SLEN], *ptr, 
	     buffer[LONG_STRING];
	register int iindex;
	int too_long = FALSE;
	struct passwd *pass, *getpwnam();
	char *getenv();

	ptr = filename;
	while (*ptr == ' ') ptr++;	/* leading spaces GONE! */
	strfcpy(temp_filename, ptr, sizeof temp_filename);

	/** New stuff - make sure no illegal char as last **/
	/** Strip off any trailing backslashes or blanks **/

	ptr = temp_filename + strlen(temp_filename) - 1;
	while (*ptr == '\n' || *ptr == '\r'
	       || *ptr == '\\' || *ptr == ' ' || *ptr == '\t') {
	    *ptr-- = '\0';
	}

	if (temp_filename[0] == '~') {
	  if(temp_filename[1] == '/')
	    elm_sfprintf(filename, size_filename,
			 FRM("%s%s%s"),
			 home, (lastch(home) != '/' ? "/" : ""), 
			 &temp_filename[2]);
	  else {
	    for(ptr = &temp_filename[1], iindex = 0; *ptr && *ptr != '/'; ptr++, iindex++)
	      logname[iindex] = *ptr;
	    logname[iindex] = '\0';
	    if((pass = getpwnam(logname)) == NULL) {
	      dprint(3,(debugfile, 
		      "Error: Can't get home directory for %s (%s)\n",
		      logname, "expand_filename"));
	      if(use_cursor_control)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmDontKnowHomeCursor,
				  "Don't know what the home directory of \"%s\" is!"),
			  logname);
	      else
		printf(catgets(elm_msg_cat, ElmSet, ElmDontKnowHomePrintf,
		    "\n\rDon't know what the home directory of \"%s\" is!\n\r"),
		    logname);
	      return(0);
	    }
	    elm_sfprintf(filename, size_filename,
			 FRM("%s%s"), pass->pw_dir, ptr);
	  }

	}
	else if (temp_filename[0] == '=' || temp_filename[0] == '+' || 
	 	 temp_filename[0] == '%') {
	  elm_sfprintf(filename, size_filename,
		       FRM("%s%s%s"), folders, 
		       (temp_filename[1] != '/' && 
			lastch(folders) != '/')? "/" : "",
		       &temp_filename[1]);
	}
	else if  ( temp_filename[0] == '@' ) {

	  struct addr_item * address = 
	    get_alias_address_l(&(temp_filename[1]),
				FALSE,&too_long);
	    
	  if (address && address[0].addr ) {

	    /* get filename from address */
	    get_return_name(address[0].addr, buffer, TRUE, sizeof buffer);

	    elm_sfprintf(filename, size_filename,
			 FRM("%s%s%s"),
			 folders, 
			 (temp_filename[1] != '/' && 
			  lastch(folders) != '/') ? "/" : "", 
			 buffer);
	  }
          else {
	      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCannotExpand,
				"Cannot expand alias '%s'!"), 
			&(temp_filename[1]) );
	      if (address)
		free_addr_items(address);

	    return(0);
	  }
	}
	else if (temp_filename[0] == '$') {	/* env variable! */
	  for(ptr = &temp_filename[1], iindex = 0; isalnum(*ptr); ptr++, iindex++)
	    varname[iindex] = *ptr;
	  varname[iindex] = '\0';

	  env_value[0] = '\0';			/* null string for strlen! */
	  if (getenv(varname) != NULL)
	    strfcpy(env_value, getenv(varname), sizeof env_value);

	  if (strlen(env_value) == 0) {
	    dprint(3,(debugfile, 
		    "Error: Can't expand environment variable $%s (%s)\n",
		    varname, "expand_filename"));
	    if(use_cursor_control)
	      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmDontKnowValueCursor,
				"Don't know what the value of $%s is!"), 
			varname);
	    else
	      printf(catgets(elm_msg_cat, ElmSet, ElmDontKnowValuePrintf,
		    "\n\rDon't know what the value of $%s is!\n\r"), varname);
	    return(0);
	  }

	  elm_sfprintf(filename, size_filename,
		       FRM("%s%s%s"), env_value, 
		       (*ptr && *ptr != '/' && 
			lastch(env_value) != '/')? "/" : "", ptr);

	} else if (strcmp(temp_filename, "!") == 0) {
	  strfcpy(filename, defaultfile, size_filename);
	} else if (strcmp(temp_filename, ">") == 0) {
	  strfcpy(filename, recvd_mail, size_filename);
	} else if (strcmp(temp_filename, "<") == 0) {
	    strfcpy(filename, sent_mail, size_filename);
	} else if (strcmp(temp_filename, ".") == 0) {
	  strfcpy(filename, prev_fold, size_filename);
	} else
	  strfcpy(filename, temp_filename, size_filename);
	  

	dprint(9, (debugfile, 
		   "expand_filename: %s -> filename=%s (size=%d)\n",
		   temp_filename,filename,size_filename));

	return(1);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
