static char rcsid[] = "@(#)$Id: fileio.c,v 1.12 2000/07/30 15:12:20 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.12 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** File I/O routines, including deletion from the folder! 

**/

#include "headers.h"
#include "s_elm.h"
#include "s_me.h"
#include <sys/stat.h>
#include <errno.h>
#include "me.h"

extern int errno;

char *error_description();

static void copy_write_error_exit()
{
	MoveCursor(elm_LINES, 0);
	Raw(OFF);
	printf(catgets(elm_msg_cat, ElmSet, ElmWriteCopyMessageFailed,
		"\nWrite in copy_message failed\n"));
	dprint(1, (debugfile,"\n*** Fprint failed on copy_message;\n"));
	rm_temps_exit();
}

#ifdef USE_PGP
/* Prototype */
static void copy_pgp P_((char *,out_state_t *,int, struct header_rec *, 
			 FILE *));

static void copy_pgp(prefix,dest_file,cm_options,current_header, infile)
     char *prefix;
     out_state_t *dest_file;
     int cm_options;
     struct header_rec *current_header;
     FILE * infile;
{
    int body_bytes = 0;
    FILE *fpin = NULL, *fpout = NULL;
    char buffer[VERY_LONG_STRING];
    int buf_len=0,err,code;
    struct run_state RS;
    int stat;
    int pgp_seen = !pgp_noarmor && current_header->pgp != PGP_PUBLIC_KEY;
    enum pgp_version v = pgp2;
    enum pgp_version version;

    int raw = RawState ();

    dprint(5,(debugfile,
	      "copy_pgp called: Need read %d bytes\n",
	      current_header->content_length));

    if (current_header->pgp & PGP_PUBLIC_KEY) {
	state_printf(dest_file,CATGETS(elm_msg_cat, MeSet, MePgpPublicKeys,
				       "(** This message contains PGP public key(s) **)\n"));
	state_putc('\n',dest_file);
    }
    
    if(pgp_seen) {
	buf_len = mail_gets(buffer, VERY_LONG_STRING, infile);
    } else {
	while (body_bytes < current_header->content_length) {      
	    if (! (buf_len = mail_gets(buffer, VERY_LONG_STRING, infile)))
		break;
	    
	    if (strncmp(buffer, "-----BEGIN PGP", 14) == 0) {
		pgp_seen = 1;
		break;
	    }
	    /* text before PGP section */
	    if (0 == body_bytes)
		state_printf(dest_file,
			     CATGETS(elm_msg_cat, MeSet, MePgpBefore,
				     "[ There is text before PGP section. ]\n"));
	    
	    body_bytes += buf_len;
	    /* Take care of CRLF => LF conversion */
	    if (buf_len > 1 &&
		buffer[buf_len - 1] == '\n' &&
		buffer[buf_len - 2] == '\r') {
		buffer[buf_len - 2] = '\n';
		buffer[buf_len - 1] = '\0';
		buf_len--;
	    }

	    err = state_printf(dest_file, FRM("%s"), prefix);
	    if (err != EOF)
		err = state_put(buffer, buf_len, dest_file);
	    if (err != buf_len) {
		copy_write_error_exit();
	    }	
	}
    }
    
    if (buf_len <= 0 || body_bytes >= current_header->content_length)
	return;
    else {
	int l = ftell(infile);
	int len1;
	char buf[STRING];
	
	/* 
	 * On PGP 2 messages these is empty line immediately after
	 * -----BEGIN PGP SIGNED MESSAGE----
	 * 
	 * If there is something other such as 
	 * Hash: SHA1
	 * PGP 2 does not understand message.
	 */
	
	while (0 < (len1 = 
		    mail_gets (buf, sizeof (buf), infile))) {
	    if ((len1 == 1 && buf[0] == '\n') ||
		(len1 == 2 && buf[0] == '\r' && buf[1] == '\n'))
		break;
	    if (current_header->pgp == PGP_SIGNED_MESSAGE) {
		dprint (4, (debugfile, 
			    "copy_pgp: peek: Header on armor -- requires PGP 5 or GnuPG\n" ));
		v = gpg;
		break;
	    }
	    if (0 == strncmp("Version: ",buf,9)) {
		char *c = buf+9;

		if (0 == strncmp(c,"2.",2)) {
		    dprint (4, (debugfile,
				"copy_pgp: peek: Version -- PGP 2\n"));
		    v = pgp2;
		    break;
		} else if (0 == strncmp(c,"GnuPG",5)) {
		    dprint (4, (debugfile,
				"pgp_decode: peek: Version -- GnuPG\n"));
		    v = gpg;
		    break;
		} else {
		    dprint (4, (debugfile,
				"copy_pgp: peek: Version -- PGP 5 ?\n"));
		    if ('/' == pgp5_dir[0])
			v = pgp5;
		    else
			v = send_pgp_version;
		    break;
		}
	    }
	}
	fseek(infile,l,SEEK_SET);
    }
    
    version = have_pgp(v);
    
    if (!version) {
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, MeSet, MePgpNotAvailRawdata,
			     "[ PGP not available, raw data follows ]\n"));
	goto raw;
    }

    if ((current_header->pgp & PGP_MESSAGE) && pgp_keeppass) {
	if (!pgp_goodPassphrase(version)) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmDecryptBadPGP,
			      "Decrypting message... Bad PGP passphrase."));
	    state_puts("[ ",dest_file);
	    state_printf(dest_file,CATGETS(elm_msg_cat, ElmSet, 
					   ElmDecryptBadPGP,
					   "Decrypting message... Bad PGP passphrase."));
	    state_puts(" ]\n",dest_file);
	    return;
	}
    }
    
    if (!(code=pgp_decrypt_init(&fpin, &fpout, current_header->pgp,
				version,&RS))) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmDecryptFailInitPGP,
			  "Decrypting message... Failed to init PGP."));
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, ElmSet, ElmDecryptFailInitPGPRaw,
			     "[ Decrypting message... Failed to init PGP. Raw data follows. ]\n"));
				raw:

	while (body_bytes < current_header->content_length) {      
	    err = state_printf(dest_file, FRM("%s"), prefix);
	    if (err != EOF)
		err = state_put(buffer, buf_len, dest_file);
	    if (err != buf_len) {
		copy_write_error_exit();
	    }	
	    body_bytes += buf_len;
	    if (! (buf_len = mail_gets(buffer, VERY_LONG_STRING, infile)))
		break;
	}

	state_printf(dest_file,
		     CATGETS(elm_msg_cat, ElmSet, ElmDecryptEndRaw,
			     "[ Decrypting message... End of raw data. ]\n"));
	return;
    }

    /* Pass PGP section to pgp */   
    while (body_bytes < current_header->content_length) {      
	if (EOF == fputs(buffer, fpout))
	    copy_write_error_exit();	
	body_bytes += buf_len;
	if (! (buf_len = mail_gets(buffer, VERY_LONG_STRING, infile)))
	    break;
    }
    
    if (fclose(fpout) == EOF)
	copy_write_error_exit();
  
    dprint(5,(debugfile,
	      "copy_pgp: Passed %d bytes to PGP.\n",
	      body_bytes));
    body_bytes = 0;

    if (current_header->pgp & PGP_MESSAGE)
	state_printf(dest_file,CATGETS(elm_msg_cat, MeSet, MePgpStartEncoded,
				       "-- Start of PGP encoded section.\n"));
    else if (current_header->pgp & PGP_SIGNED_MESSAGE)
	state_printf(dest_file,CATGETS(elm_msg_cat, MeSet, MePgpStartSigned,
				       "-- Start of PGP signed section.\n"));
    else if (current_header->pgp & PGP_PUBLIC_KEY)
	state_printf(dest_file,CATGETS(elm_msg_cat, MeSet, MePgpStartOutput,
				       "-- Start of PGP output.\n"));
    else 
	state_printf(dest_file,CATGETS(elm_msg_cat, MeSet, MePgpStart,
				       "-- Start of PGP section.\n"));
  
    while (0 < (buf_len = mail_gets(buffer, VERY_LONG_STRING, fpin))) {
	body_bytes += buf_len;
	
	err = state_printf(dest_file, FRM("%s"), prefix);
	if (err != EOF)
	    err = state_put(buffer, buf_len, dest_file);
	if (err != buf_len) {
	    copy_write_error_exit();
	}
    }
    fclose(fpin);

    if (0 == code)
	code = wait_end(&RS,&stat);

    PressAnyKeyToContinue();
    if (raw)
	Raw (ON);

    if (current_header->pgp & PGP_MESSAGE)
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, MeSet, MePgpEndEncoded,
			     "-- End of PGP encoded section%s\n"),
		     code < 0 || stat ? catgets(elm_msg_cat, MeSet, MePgpFail,
						", PGP failed!") : ".");
    else if (current_header->pgp & PGP_SIGNED_MESSAGE)
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, MeSet, MePgpEndSigned,
			     "-- End of PGP signed section%s\n"),
		     code < 0 || stat ?  catgets(elm_msg_cat, MeSet, MePgpFail,
						 ", PGP failed!") : ".");
    else if (current_header->pgp & PGP_PUBLIC_KEY)
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, MeSet, MePgpEndOutput,
			     "-- End of PGP output%s\n"),
		     code < 0 || stat ?  catgets(elm_msg_cat, MeSet, MePgpFail,
						 ", PGP failed!") : ".");
    else
	state_printf(dest_file,
		     CATGETS(elm_msg_cat, MeSet, MePgpEnd,
			     "-- End of PGP section%s\n"),
		     code < 0 || stat ? catgets(elm_msg_cat, MeSet, MePgpFail,
						", PGP failed!") : ".");
  
    dprint(5,(debugfile,
	      "copy_pgp: Readed %d bytes from PGP.\n",
	      body_bytes));
    
    if (body_bytes != current_header->content_length) {
	dprint(1,(debugfile,
		  "copy_pgp: ERROR: readed bytes %d != content-length %d\n",
		  body_bytes,current_header->content_length));
    }
}
#endif /* USE_PGP */

/* Prototype */
void copy_mime P_((char *,out_state_t *,int, struct header_rec *, FILE *));

void copy_mime(prefix,dest_file,cm_options,current_header, infile)
     char *prefix;
     out_state_t *dest_file;
     int cm_options;
     struct header_rec *current_header;
     FILE *infile;
{
    in_state_t  state_in;
    in_state_clear(&state_in,  STATE_in_file);

    dprint(5,(debugfile,
	      "copy_mime called: Need read %d bytes\n",
	      current_header->content_length));

    attach_parse(current_header, infile);
    
    set_in_state_file(infile,&state_in);

    dest_file->prefix = prefix;
    dest_file->displaying      = (cm_options & CM_DISPLAYING) ? 1 : 0;
    mime_decode(&(current_header->mime_rec), &state_in, dest_file);
  
    dprint(5,(debugfile,"copy_mime: mail decoded\n"));


    in_state_destroy(&state_in); 
    dest_file->prefix = NULL;
}

/* Prototype */
static void copy_encrypted P_((char *,out_state_t *,int, struct header_rec *, 
			       FILE *));

static void copy_encrypted(prefix,dest_file,cm_options,current_header,infile)
     char *prefix;
     out_state_t *dest_file;
     int cm_options;
     struct header_rec *current_header;
     FILE *infile;
{
    char buffer[VERY_LONG_STRING];
    long body_bytes = 0;
    int crypted = OFF;
    int buf_len;
    
    dprint(5,(debugfile,
	      "copy_encrypted called: Need read %d bytes\n",
	      current_header->content_length));
    
    getkey(OFF);
  
    while (body_bytes < current_header->content_length) {      
	if (! (buf_len = mail_gets(buffer, sizeof buffer, infile)))
	    break;
    
	body_bytes += buf_len;
	/* Take care of CRLF => LF conversion */
	if (buf_len > 1 &&
	    buffer[buf_len - 1] == '\n' &&
	    buffer[buf_len - 2] == '\r') {
	    buffer[buf_len - 2] = '\n';
	    buffer[buf_len - 1] = '\0';
	    buf_len--;
	}
    
	if (mime_body_keywords && !strncmp(buffer, START_ENCODE, strlen(START_ENCODE))) {
	    crypted = ON;
	    state_puts(catgets(elm_msg_cat, MeSet, MeDecodeStartElmEncoded,
			       "-- Start of (Elm) encoded section.\n"),
		       dest_file);
	    continue;
	} else if (mime_body_keywords && !strncmp(buffer, END_ENCODE, strlen(END_ENCODE))) {
	    crypted = OFF;
	    state_puts(catgets(elm_msg_cat, MeSet, MeDecodeEndElmEncoded,
			       "-- End of (Elm) encoded section.\n"),
		       dest_file);
	    continue;
	} else if (crypted) {
	    no_ret(buffer);
	    encode(buffer);      
	    strfcat(buffer, "\n", sizeof buffer);
	}
	if (state_printf(dest_file, FRM("%s%s"), prefix, buffer) == EOF) {
	    copy_write_error_exit();
	}
    }
    dprint(5,(debugfile,
	      "copy_encrypted: Readed %d bytes from body\n",body_bytes));

    if (body_bytes != current_header->content_length) {
	dprint(1,(debugfile,
		  "copy_encrypted: ERROR: readed bytes %d != content-length %d\n",
		  body_bytes,current_header->content_length));
    }
}

/* Prototype */
void copy_plain P_((char *,out_state_t *,int, struct header_rec *, FILE *));

void copy_plain(prefix,dest_file,cm_options,current_header, infile) 
     char *prefix;
     out_state_t *dest_file;
     int cm_options;
     struct header_rec *current_header;
     FILE *infile;
{
    char buffer[VERY_LONG_STRING];
    int body_bytes = 0;
    int buf_len, err;
    
    dprint(5,(debugfile,
	      "copy_plain called: Need read %d bytes\n",
	      current_header->content_length));;
    
    
    while (body_bytes < current_header->content_length) {      
	if (! (buf_len = mail_gets(buffer, VERY_LONG_STRING, infile)))
	    break;
	
	body_bytes += buf_len;
	/* Take care of CRLF => LF conversion */
	if (buf_len > 1 &&
	    buffer[buf_len - 1] == '\n' &&
	    buffer[buf_len - 2] == '\r') {
	    buffer[buf_len - 2] = '\n';
	    buffer[buf_len - 1] = '\0';
	    buf_len--;
	}
	
	err = state_printf(dest_file, FRM("%s"), prefix);
	if (err != EOF)
	    err = state_put(buffer, buf_len, dest_file);
	if (err != buf_len) {
	    copy_write_error_exit();
	}
    }
    dprint(5,(debugfile,
	      "copy_plain: Readed %d bytes from body\n",body_bytes));
    
    if (body_bytes != current_header->content_length) {
	dprint(1,(debugfile,
		  "copy_plain: ERROR: readed bytes %d != content-length %d\n",
		  body_bytes,current_header->content_length));
    }
}

/* Prototype */
void copy_binary P_((char *,out_state_t *,int, struct header_rec *, FILE *));

void copy_binary(prefix,dest_file,cm_options,current_header, infile) 
     char *prefix;
     out_state_t *dest_file;
     int cm_options;
     struct header_rec *current_header;
     FILE *infile;
{
    char buffer[VERY_LONG_STRING];
    int body_bytes = 0;
    int buf_len, err;
    
    dprint(5,(debugfile,
	      "copy_binary called: Need read %d bytes\n",
	      current_header->content_length));;

    while (body_bytes < current_header->content_length) {      
	if (! (buf_len = mail_gets(buffer, VERY_LONG_STRING, infile)))
	    break;

	body_bytes += buf_len;
    
	err = state_printf(dest_file, FRM("%s"), prefix);
	if (err != EOF)
	    err = state_put(buffer, buf_len, dest_file);
	if (err != buf_len) {
	    copy_write_error_exit();
	}
    }
    dprint(5,(debugfile,
	      "copy_binary: Readed %d bytes from body\n",body_bytes));

    if (body_bytes != current_header->content_length) {
	dprint(1,(debugfile,
		  "copy_binary: ERROR: readed bytes %d != content-length %d\n",
		  body_bytes,current_header->content_length));
    }
}

copy_decoder_t select_copy_decoder (current_header) 
     struct header_rec * current_header;
{
    if (current_header->status & MIME_MESSAGE) 
	return copy_mime; 
#ifdef USE_PGP
    else if (current_header->pgp & (PGP_MESSAGE|PGP_SIGNED_MESSAGE|PGP_PUBLIC_KEY))
	return copy_pgp;
#endif
    else if (current_header -> encrypted)
	return copy_encrypted;
    else
	return copy_plain;
}

void
copy_message(infolder,
	     current_header,
	     prefix, 
	     dest_file, 
	     cm_options)
     struct folder_info *infolder;
     struct header_rec *current_header;
     char *prefix;
     FILE *dest_file;
     int cm_options;
{
    FILE * infile = folder_to_fd(infolder,current_header->offset);
    if (!infile) {
	dprint(1, (debugfile, 
		   "ERROR: Attempt to seek %d bytes into file failed (%s)",
		   current_header->offset, "copy_message"));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSeekFailed,
			  "ELM [seek] failed trying to read %d bytes into file."),
		  current_header->offset);
	return;
    }
 
    copy_message_1(infile,current_header,prefix,			
		   dest_file, cm_options);
    return;
}

/* Assumes that headers are already copied ------------------- */
void copy_body(infile,current_header,prefix,dest_file,cm_options)
     FILE *infile;
     struct header_rec *current_header;
     char *prefix; 
     out_state_t *dest_file; 
     int cm_options;
{
    int decode = cm_options & CM_DECODE;
    int j;

    dprint(9, (debugfile,
	       "copy_body:  decode=%d\n",decode));
    for (j = 0; dest_file->display_charset[j]; j++) {
	dprint(9, (debugfile,
		   "         : 'display charset[%d]'=%s\n",
		   j,
		   dest_file->display_charset[j]->MIME_name ?
		   dest_file->display_charset[j]->MIME_name :
		   "<no MIME name>"));
    }

    if (decode) {   
	copy_decoder_t decoder = select_copy_decoder(current_header);
	decoder(prefix,dest_file,cm_options,current_header,infile);
    } else    
	copy_binary(prefix,dest_file,cm_options,current_header,infile);

}


void copy_message_2(infile,current_header,prefix,dest_file,cm_options)
     FILE *infile;
     struct header_rec *current_header;
     char *prefix; 
     out_state_t *dest_file; 
     int cm_options;
{
    /** Copy current message to destination file, with optional 'prefix' 
	as the prefix for each line.  If remove_header is true, it will 
	skip lines in the message until it finds the end of header line...
	then it will start copying into the file... If remote is true
	then it will append "remote from <hostname>" at the end of the
	very first line of the file (for remailing) 
	
	If "filter_header" is true then it'll do some nice things to
	ensure that the forwarded message looks pleasant; e.g. remove
	stuff like ">From " lines and "Received:" lines.
	
	If "update_status" is true then it will write a new Status:
	line at the end of the headers.  It never copies an existing one.
	
	If "decode" decode MIME, PGP and elm's (unsafe) own decoding.
    **/
    

    char buffer[32*1024+1]; /* Allow one header line to be 32 KB
			     * after unfolding 
			     */
    int in_header = 1, first_line = TRUE;
    int remove_header = cm_options & CM_REMOVE_HEADER;
    int remove_envelope = cm_options & CM_REMOVE_ENVELOPE;
    int remote = cm_options & CM_REMOTE;
    int update_status = cm_options & CM_UPDATE_STATUS;
    int remail = cm_options & CM_REMAIL;
    int decode = cm_options & CM_DECODE;
    int filter_headers = cm_options & CM_FILT_HDR;
    int bytes_seen = 0;
    int content_length_seen = FALSE;
    long CL_pos = -1L, BODY_pos = -1L, END_pos = -1L;
    int was_binary = current_header -> binary && !decode;
    
    int buf_len;
    int i;
  

    dprint(5, (debugfile,
	       "copy_message_2(cm_options=(%d)%s%s%s%s%s%s%s%s)\n",
	       cm_options,
	       remove_header   ? " CM_REMOVE_HEADER"   : "",
	       remove_envelope ? " CM_REMOVE_ENVELOPE" : "",
	       remote          ? " CM_REMOTE"          : "",
	       update_status   ? " CM_UPDATE_STATUS"   : "",
	       remail          ? " CM_REMAIL"          : "",
	       decode          ? " CM_DECODE"          : "",
	       filter_headers  ? " CM_FILT_HDR"        : "",
	       (cm_options & CM_DISPLAYING) ? " CM_DISPLAYING" : ""));
    
    /** get to the first line of the message desired **/
    

    /* No filter on here ... */
    dest_file->filter = NULL;

    /* now while not EOF & still in message... copy it! */
    
    dprint(5,(debugfile,
	      "copy_message_2: [%ld] start mailbox separator section\n",
	      ftell(infile)));
  
#ifdef MMDF
    if (!(CM_MMDF_HEAD & cm_options) && !remove_header && !remove_envelope) {
	dprint(5,(debugfile,"copy_message: Write MMDF message separator\n"));
	if (fprintf(dest_file, "%s", MSG_SEPARATOR) == EOF)
	    copy_write_error_exit();
    }   
#endif
  
    while (in_header) {      
	long last_pos = ftell(infile);
	if (last_pos < 0) {
	    dprint(5,(debugfile,"copy_message: ftell(infile) failed!\n"));
	    break;
	}
	
	if (! (buf_len = mail_gets(buffer, sizeof(buffer), infile)))
	    break;
    
#ifdef MMDF
	if (strcmp(buffer, MSG_SEPARATOR) == 0)
	    continue; /* MSG SEPRATOR is already written */
#endif /* MMDF */
    
	if(buffer[buf_len - 1] == '\n') {
	    no_ret(buffer);
	    if (first_word(buffer, "From ")) {
		/* From header needs regenerated for POP messages ... */
		char * buffer1 = ctime(& current_header->received_time);
		
		if (first_line && remote && !remove_header && 
		    !remove_envelope) {
		  if (state_printf(dest_file, 
				   FRM("%sFrom %s %.24s remote from %s\n"),
				   prefix, 
				   current_header->env_from, 
				   buffer1,
				   hostname) == EOF) {
		      copy_write_error_exit();
		  }
		  dprint(5,(debugfile,
			    "copy_message- added: %sFrom %s %.24s remote from %s\n",
			    prefix, 
			    current_header->env_from, 
			    buffer1,
			    hostname));
		  first_line = FALSE;
		  continue;
	      }
	      if (!remove_header && !remove_envelope) {
		  if (state_printf(dest_file, 
				   FRM("%sFrom %s %.24s\n"), 
				   prefix, current_header->env_from,
				   buffer1) == EOF) {
		      copy_write_error_exit();
		  }
		  dprint(5,(debugfile,
			    "copy_message- added: %sFrom %s %.24s\n", 
			    prefix, current_header->env_from,
			    buffer1));
	      }
	      first_line = FALSE;
	      continue;
	    }
	    
	    if (!first_line && first_word_nc(buffer, ">From")) {
		if (!filter_headers && !remove_header && !remove_envelope) {
		    if (state_printf(dest_file, 
				     FRM("%s%s\n"), 
				     prefix, buffer) == EOF) {
			copy_write_error_exit();
		    }
		}
		continue;	
	    }
	    /* fall thru */
	}
	dprint(5,(debugfile,
		  "copy_message: Not a mailbox line -- seeking back!\n"));
	dprint(5,(debugfile,
		  "copy_message- Line was: %s\n",buffer));
	
	if (0 != fseek(infile,last_pos,SEEK_SET)) {
	    dprint(5,(debugfile,
		      "copy_message: seek failed!\n"));
	    copy_write_error_exit();
	}
	break; /* Go out of loop */
    }
    if (first_line) {
	dprint(5,(debugfile,
		  "copy_message: No 'From ' line seen...\n"));
	if (!remove_header && !remove_envelope) {
	    time_t thetime = time((time_t *) 0);
	    dprint(5,(debugfile,
		      "copy_message: Create it!\n"));
	    if (state_printf(dest_file,
			     FRM("%sFrom %s %s"), 
			     prefix,username, ctime(&thetime)) == EOF)
		copy_write_error_exit();
	}
    }
  
    dprint(5,(debugfile,"copy_message: [%ld] start header section\n",
	      ftell(infile)));
  
    while (in_header) {      
	char * colptr,*hdr,*body,*ptr;
	int header_class;
	
	if ( (buf_len = read_header_line(infile,buffer,sizeof(buffer),
					 RHL_CHECK_HEADER|RHL_MARK_FOLDING))
	     < 2) {
	    dprint(5,(debugfile,"copy_message: End of headers: buf_len=%d\n",
		      buf_len));
	    break;
	}
	
	bytes_seen += buf_len;
	
	colptr = strchr(buffer, ':');
	if (NULL == colptr) {
	    in_header = 0;
	    dprint(1,(debugfile,"copy_message: Software error; No ':' !!\n"));
	    break;
	}
	*colptr = '\0';
	hdr = buffer;
	body = colptr +1;
	
	if (remove_header)
	    continue;
    
	header_class = classify_header(hdr);
	dprint(12,(debugfile,"copy_message: header=%s,class=%d,rest=%s\n",
		   hdr,header_class,body));;
    
	if (decode && !(current_header -> status & NOHDRENCODING)) {
	    
	    if (header_class & HDR_TEXT) {
		if (is_rfc1522 (body))
		    rfc1522_decode (body, sizeof (buffer) - (body - buffer) -1);
	    } else if (header_class & HDR_STRUCTURED) {
		rfc1522_decode_structured(header_class,
					  body, 
					  sizeof (buffer) -(body - buffer) -1);
	    }
      
	    dprint(12,(debugfile,"copy_message: decoded rest=%s\n",body));
	}
	if ((remote || remail) && 0 == istrcmp("Sender",hdr)) {
	    continue;
	}
	if (0 == istrcmp("X-UIDL",hdr))    /* Generated by some POP deamons */
	    continue;
	/* These headers are incorrect after decoding ... */
	if (decode && (current_header->status & MIME_MESSAGE) &&
	    (0 == istrcmp(buffer, "MIME-Version") ||
	     0 == istrcmp(buffer, "Content-Type") ||
	     0 == istrcmp(buffer, "Content-Transfer-Encoding")))
	    continue;
	
	if(0 == istrcmp(buffer, "Content-Length")) {
	    /* make correct Content-Length later */
	    content_length_seen = TRUE;
	    continue;
	}
	
	if (!filter_headers) {
	    if (0 == istrcmp(buffer,"Status"))
		continue;   /* we will output a new Status: line later, if desired. */
	}
	else { /* filter_headers */
	    if (0 == istrcmp(buffer, "Received") ||
		0 == istrcmp(buffer, "Status") ||
		0 == istrcmp(buffer, "Return-Path"))
		continue;
	    if (remail && 0 == istrcmp(buffer, "To"))
		hdr = "Orig-To";
	}
	dprint(5,(debugfile,"copy_message- hdr=%s body=%s\n",hdr,body));
    
	for (ptr = strtok(body,"\n"); ptr; ptr = strtok(NULL,"\n")) { 
	    if (state_printf(dest_file,
			     FRM("%s"),
			     prefix) == EOF)
		copy_write_error_exit ();
      
	    if (ptr > body) { /* Do folding */
		--ptr;
		if (*(ptr+1) == ' ')
		    *ptr = ' ';
		else
		    *ptr = '\t';
	    } else {
		if (state_printf(dest_file,
				 FRM("%s:"),hdr) == EOF)
		    copy_write_error_exit ();
	    }
	    if (state_printf(dest_file,
			     FRM("%s%s"),
			     ptr,was_binary ? "\r\n" : "\n") == EOF)
		copy_write_error_exit ();
	}       
    }
  
    dprint(5,(debugfile,
	      "copy_message: [%ld] end of headers. Readed ~ %d bytes.\n",
	      ftell(infile), bytes_seen));
    
    if (!remove_header) {  
	if (out_state_seekable(dest_file) && 
	    ((!filter_headers && content_length_seen) ||
	     (!remove_envelope && (current_header->have_from || decode)))) {
	    
	    if (state_printf (dest_file, 
			      FRM("Content-Length: ")) == EOF)
		copy_write_error_exit ();
	    CL_pos = out_state_ftell(dest_file);
	    if (state_printf (dest_file, 
			      FRM("%5d%s"),
			      current_header->content_length,
			      was_binary ? "\r\n" : "\n") == EOF)
		copy_write_error_exit ();
	}
	if (remail) {
	    if (state_printf (dest_file, 
			      FRM("%sSender: %s%s"),
			      prefix,username,
			      was_binary ? "\r\n" : "\n") == EOF)
		copy_write_error_exit ();
	}
	if (update_status) {
	    if (state_printf (dest_file, 
			      FRM("%sStatus: "), 
			      prefix) == EOF)
		copy_write_error_exit ();
	
	    if (ison(current_header->status, NEW)) {
		if (state_printf(dest_file, FRM("N")) == EOF)
		    copy_write_error_exit();
	    } else {
		if (state_printf(dest_file, FRM("O")) == EOF)
		    copy_write_error_exit();
	    }
	    if (isoff(current_header->status, UNREAD)) {
	    	/* read */
		if (state_printf(dest_file, FRM("R")) == EOF)
		    copy_write_error_exit();
	    }
	    if (ison(current_header->status, REPLIED)) {
		if (state_printf (dest_file, FRM("r")) == EOF)
		    copy_write_error_exit ();
	    }
	    /* save all the status flags that ELM doesn't understand */
	    for (i=0; current_header->mailx_status[i] != '\0'; i++)
		switch (current_header->mailx_status[i]) {
		case 'N':
		case 'R':
		case 'O':
		case 'r':
		    break;
		default:
		    if (state_printf (dest_file, FRM("%c"), 
				      current_header->mailx_status[i]) == EOF)
			copy_write_error_exit ();
		}
	    if (state_printf(dest_file, 
			     FRM("%s"),
			     was_binary ? "\r\n" : "\n") == EOF) {
		copy_write_error_exit();
	    }      
	}	

    
	/*
	 * Add empty line between headers and body (that was not copied
	 *  in above)
	 */
	
	if (state_printf(dest_file, 
			 FRM("%s"),
			 was_binary ? "\r\n" : "\n") == EOF) {
	    copy_write_error_exit();
	}

    }
   

    dprint(5,(debugfile,
	      "copy_message: [%ld] Starting reading of body: %d bytes expected\n",
	      ftell(infile),
	      current_header->content_length));

    
    if (out_state_seekable(dest_file)) {
	BODY_pos = out_state_ftell(dest_file);
	dprint(5,(debugfile,
		  "copy_message> output offset=[%ld]\n",BODY_pos));
    }
     
    copy_body(infile,current_header,prefix,dest_file,cm_options);

  dprint(5,(debugfile,
	    "copy_message: [%ld] Body readed.\n",
	    ftell(infile)));

  if (out_state_seekable(dest_file)) {
      FILE * f1 = out_state_FILE(dest_file);
      END_pos = out_state_ftell(dest_file);

      dprint(5,(debugfile,
		"copy_message> output offset=[%ld]\n",END_pos));

      if (CL_pos > 0 && BODY_pos > 0 && END_pos >= BODY_pos) {
	  /* Actually written content length is good if conversions are
	   * not done and there was no errors ... 
	   */
	  if (f1)
	      clearerr(f1);

	  /* Notice that these tests indicates failure */
	  if (0   != out_state_fseek(dest_file,CL_pos) ||
	      out_state_ftell(dest_file) != CL_pos ||
	      EOF == state_printf (dest_file, 
				   FRM("%5d"), (int) (END_pos - BODY_pos)) ||
	      0   != out_state_fseek(dest_file,END_pos)) {
	      dprint(5,(debugfile,
			"copy_message: Writing Content-length -- writing or seeking failed.\n"));
	      copy_write_error_exit();
	  } else {
	      dprint(5,(debugfile,
			"copy_message: Content-length fixed: %d bytes.\n",
			(int) (END_pos - BODY_pos)));
	  }
      
      }

      if (NULL != f1) {
	  char buffer[2];
	  int i;
	  long NEW_pos = END_pos - 2L;
	  buffer[0] = '\0';
	  buffer[1] = '\0';
	  if (NEW_pos < 0L)
	      NEW_pos = 0L;
	  
	  if (0 != fseek(f1,NEW_pos,SEEK_SET)) {
	      dprint(5,(debugfile,
			"copy_message: Failed to seek to %ld (output)\n",
			NEW_pos)); 
	  }

	  i = fread(buffer,sizeof (char),2,f1);
	  if (i < 0) {
	      dprint(5,(debugfile,
			"copy_message: Read error! (output)\n"));
	      i = 0;
	  } else {
	      dprint(5,(debugfile,
			"copy_message: readed %d bytes (output). \n",i));
	  }
    
	  for (; i < 2; i++) {
	      buffer[i] = '\0';
	  }

	  dprint(5,(debugfile,
		    "copy_message: [%ld] last two bytes %d %d (output)\n",
		    ftell(infile), buffer[0], buffer[1])); 
    
	  if (0 != fseek(f1,0L,SEEK_CUR)) {
	      dprint(5,(debugfile,
			"copy_message: Failed to seek 0 bytes (output)\n"));
	  }

	  if (buffer[1] != '\n') {
	      /* No \n in end ? */ 	
	      dprint(5,(debugfile,
			"copy_message: NL missing from end of mail\n")); 
	      if (fprintf(f1, "\n") == EOF) {
		  dprint(5,(debugfile,
			    "copy_message: Failed to add NL to end of mail\n"));
		  copy_write_error_exit();
	      }
	  }

	  /* NOTE:
	   * Ending empty line or MSG_SEPARATOR is not part of 
	   * mail, but instead part of folder format, so it
	   * should NOT be included on calculation of content-length
	   */
#ifndef MMDF
	  if (buffer[0] != '\n') {
	      dprint(5,(debugfile,
			"copy_message: NL NL missing from end of mail\n")); 
	      /* blank line to keep mailx happy *sigh* */      
	      if (fprintf(f1, "\n") == EOF) {
		  dprint(5,(debugfile,
			    "copy_message: Failed to add second NL to end of mail\n"));
		  copy_write_error_exit();
	      }
	  }
#else
	  if (!(CM_MMDF_HEAD & cm_options) && !remove_header && !remove_envelope) {
	      dprint(5,(debugfile,"copy_message: Write MMDF message separator (end)\n"));
	      if (fprintf(f1, "%s", MSG_SEPARATOR) == EOF)
		  copy_write_error_exit();
	  }   
#endif
      }
  }
}

void copy_message_1(infile,
		   current_header,
		   prefix, 
		   dest_file, 
		   cm_options)
     FILE * infile;
     struct header_rec *current_header;
     char *prefix;
     FILE *dest_file;
     int cm_options;
{
    out_state_t buffer;
    
    out_state_clear(&buffer,STATE_out_file);
    set_out_state_file(dest_file,&buffer);
    copy_message_2(infile,current_header,prefix,&buffer,cm_options);
    out_state_destroy(&buffer);

    /* Since fprintf is buffered, its return value is only useful for
     * writes which exceed the blocksize.  Do a fflush to ensure that
     * the message has, in fact, been written.
     */
    
    if (fflush(dest_file) == EOF) {
	dprint(5,(debugfile,
		  "copy_message: Final fflush failed!\n")); 
	copy_write_error_exit();
    }
}

/** and finally, here's something for that evil trick: site hiding **/

#ifdef SITE_HIDING

int
is_a_hidden_user(specific_username)
char *specific_username;
{
	/** Returns true iff the username is present in the list of
	   'hidden users' on the system.
	**/
	
    FILE *hidden_users;
    char  buffer[VERY_LONG_STRING];

    /* 
	this line is deliberately inserted to ensure that you THINK
	about what you're doing, and perhaps even contact the author
	of Elm before you USE this option...
     */

	if ((hidden_users = fopen (HIDDEN_SITE_USERS,"r")) == NULL) {
	  dprint(1, (debugfile,
		  "Couldn't open hidden site file %s [%s]\n",
		  HIDDEN_SITE_USERS, error_description(errno)));
	  return(FALSE);
	}

	while (fscanf(hidden_users, "%s", buffer) != EOF)
	  if (strcmp(buffer, specific_username) == 0) {
	    dprint(3, (debugfile, "** Found user '%s' in hidden site file!\n",
		    specific_username));
	    fclose(hidden_users);
	    return(TRUE);
	  }

	fclose(hidden_users);
	dprint(3, (debugfile, 
		"** Couldn't find user '%s' in hidden site file!\n",
		specific_username));

	return(FALSE);
}

#endif


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
